/* Generated by Cython 0.29.21 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/home/vincent/miniconda3/envs/sage-flatsurf/lib/python3.7/site-packages/cysignals/macros.h",
            "/home/vincent/miniconda3/envs/sage-flatsurf/lib/python3.7/site-packages/cysignals/struct_signals.h",
            "/home/vincent/miniconda3/envs/sage-flatsurf/lib/python3.7/site-packages/ppl/ppl_shim.hh",
            "/home/vincent/miniconda3/envs/sage-flatsurf/lib/python3.7/site-packages/sage/cpython/cython_metaclass.h",
            "/home/vincent/miniconda3/envs/sage-flatsurf/lib/python3.7/site-packages/sage/libs/ntl/ntlwrap.h"
        ],
        "extra_compile_args": [
            "-std=c++11"
        ],
        "extra_link_args": [
            "-lpari"
        ],
        "include_dirs": [
            "/home/vincent/miniconda3/envs/sage-flatsurf/lib/python3.7/site-packages/ppl",
            "/home/vincent/miniconda3/envs/sage-flatsurf/lib/python3.7/site-packages/sage/cpython",
            "/home/vincent/miniconda3/envs/sage-flatsurf/lib/python3.7/site-packages/sage/libs/ntl",
            "/home/vincent/miniconda3/envs/sage-flatsurf/lib/python3.7/site-packages/cysignals"
        ],
        "language": "c++",
        "libraries": [
            "ppl",
            "gmp"
        ],
        "name": "surface_dynamics.interval_exchanges.iet_family",
        "sources": [
            "surface_dynamics/interval_exchanges/iet_family.pyx"
        ]
    },
    "module_name": "surface_dynamics.interval_exchanges.iet_family"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_21"
#define CYTHON_HEX_VERSION 0x001D15F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(U other) { return *ptr == other; }
    template<typename U> bool operator !=(U other) { return *ptr != other; }
  private:
    T *ptr;
};

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__surface_dynamics__interval_exchanges__iet_family
#define __PYX_HAVE_API__surface_dynamics__interval_exchanges__iet_family
/* Early includes */
#include "gmp.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "struct_signals.h"
#include "ios"
#include "new"
#include "stdexcept"
#include "typeinfo"
#include <vector>
#include "gmpxx.h"
#include "ppl.hh"
#include "ppl_shim.hh"
#include "ntlwrap.h"
#include "cython_metaclass.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "surface_dynamics/interval_exchanges/iet_family.pyx",
  "stringsource",
  "memory.pxd",
  "stdsage.pxd",
  "element.pxd",
  "type.pxd",
  "linear_algebra.pxd",
  "generator.pxd",
  "constraint.pxd",
  "polyhedron.pxd",
  "sage_object.pxd",
  "category_object.pxd",
  "coerce_dict.pxd",
  "parent.pxd",
  "inherit_comparison.pxd",
  "map.pxd",
  "morphism.pxd",
  "integer.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_3ppl_14linear_algebra_Variable;
struct __pyx_obj_3ppl_14linear_algebra_Variables_Set;
struct __pyx_obj_3ppl_14linear_algebra_Linear_Expression;
struct __pyx_obj_3ppl_9generator_Generator;
struct __pyx_obj_3ppl_9generator_Generator_System;
struct __pyx_obj_3ppl_9generator_Poly_Gen_Relation;
struct __pyx_obj_3ppl_10constraint_Constraint;
struct __pyx_obj_3ppl_10constraint_Constraint_System;
struct __pyx_obj_3ppl_10constraint_Poly_Con_Relation;
struct __pyx_obj_3ppl_10polyhedron_Polyhedron;
struct __pyx_obj_3ppl_10polyhedron_C_Polyhedron;
struct __pyx_obj_3ppl_10polyhedron_NNC_Polyhedron;
struct __pyx_obj_4sage_9structure_11sage_object_SageObject;
struct __pyx_obj_4sage_9structure_15category_object_CategoryObject;
struct __pyx_obj_4sage_9structure_11coerce_dict_MonoDict;
struct __pyx_obj_4sage_9structure_11coerce_dict_TripleDict;
struct __pyx_obj_4sage_9structure_6parent_Parent;
struct __pyx_obj_4sage_9structure_6parent_Set_generic;
struct __pyx_obj_4sage_4misc_18inherit_comparison_InheritComparisonMetaclass;
struct __pyx_obj_4sage_9structure_7element_Element;
struct __pyx_obj_4sage_9structure_7element_ElementWithCachedMethod;
struct __pyx_obj_4sage_9structure_7element_ModuleElement;
struct __pyx_obj_4sage_9structure_7element_MonoidElement;
struct __pyx_obj_4sage_9structure_7element_MultiplicativeGroupElement;
struct __pyx_obj_4sage_9structure_7element_AdditiveGroupElement;
struct __pyx_obj_4sage_9structure_7element_RingElement;
struct __pyx_obj_4sage_9structure_7element_CommutativeRingElement;
struct __pyx_obj_4sage_9structure_7element_IntegralDomainElement;
struct __pyx_obj_4sage_9structure_7element_DedekindDomainElement;
struct __pyx_obj_4sage_9structure_7element_PrincipalIdealDomainElement;
struct __pyx_obj_4sage_9structure_7element_EuclideanDomainElement;
struct __pyx_obj_4sage_9structure_7element_FieldElement;
struct __pyx_obj_4sage_9structure_7element_AlgebraElement;
struct __pyx_obj_4sage_9structure_7element_CommutativeAlgebraElement;
struct __pyx_obj_4sage_9structure_7element_InfinityElement;
struct __pyx_obj_4sage_9structure_7element_Vector;
struct __pyx_obj_4sage_9structure_7element_Matrix;
struct __pyx_obj_4sage_10categories_3map_Map;
struct __pyx_obj_4sage_10categories_3map_Section;
struct __pyx_obj_4sage_10categories_3map_FormalCompositeMap;
struct __pyx_obj_4sage_10categories_8morphism_Morphism;
struct __pyx_obj_4sage_10categories_8morphism_SetMorphism;
struct __pyx_obj_4sage_5rings_7integer_Integer;
struct __pyx_obj_4sage_5rings_7integer_int_to_Z;
struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily;
struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__;
struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr;
struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr;
struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr;
struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree;
struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element;
struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr;
struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics;
struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr;
struct __pyx_t_4sage_9structure_11coerce_dict_mono_cell;
struct __pyx_t_4sage_9structure_11coerce_dict_triple_cell;

/* "sage/structure/coerce_dict.pxd":5
 * 
 * 
 * cdef struct mono_cell:             # <<<<<<<<<<<<<<
 *     PyObject* key_id
 *     PyObject* key_weakref
 */
struct __pyx_t_4sage_9structure_11coerce_dict_mono_cell {
  PyObject *key_id;
  PyObject *key_weakref;
  PyObject *value;
};

/* "sage/structure/coerce_dict.pxd":27
 * 
 * 
 * cdef struct triple_cell:             # <<<<<<<<<<<<<<
 *     PyObject* key_id1
 *     PyObject* key_id2
 */
struct __pyx_t_4sage_9structure_11coerce_dict_triple_cell {
  PyObject *key_id1;
  PyObject *key_id2;
  PyObject *key_id3;
  PyObject *key_weakref1;
  PyObject *key_weakref2;
  PyObject *key_weakref3;
  PyObject *value;
};
struct __pyx_opt_args_4sage_9structure_6parent_6Parent_init_coerce;
struct __pyx_opt_args_4sage_9structure_6parent_6Parent_get_action;
struct __pyx_opt_args_4sage_9structure_6parent_6Parent__generic_convert_map;
struct __pyx_opt_args_4sage_9structure_6parent_6Parent_discover_action;

/* "parent.pxd":108
 * 
 * # Flags for Parent.flags
 * cdef enum:             # <<<<<<<<<<<<<<
 *     # If this flag is set, call __richcmp__ on elements without
 *     # coercion. This allows a completely custom comparison function.
 */
enum  {
  __pyx_e_4sage_9structure_6parent_Parent_richcmp_element_without_coercion = 1
};

/* "parent.pxd":34
 * 
 *     # Called from the __init__ method to set up coercion.
 *     cdef int init_coerce(self, bint warn=*) except -1             # <<<<<<<<<<<<<<
 * 
 *     # returns whether or not there is a Morphism from S to self
 */
struct __pyx_opt_args_4sage_9structure_6parent_6Parent_init_coerce {
  int __pyx_n;
  int warn;
};

/* "parent.pxd":52
 *     # returns the Action by/on self on/by S
 *     # corresponding to op and self_on_left
 *     cpdef get_action(self, S, op=*, bint self_on_left=*, self_el=*, S_el=*)             # <<<<<<<<<<<<<<
 *     cpdef _get_action_(self, S, op, bint self_on_left)
 * 
 */
struct __pyx_opt_args_4sage_9structure_6parent_6Parent_get_action {
  int __pyx_n;
  PyObject *op;
  int self_on_left;
  PyObject *self_el;
  PyObject *S_el;
};

/* "parent.pxd":62
 * 
 *     # For internal use
 *     cpdef _generic_convert_map(self, S, category=*)             # <<<<<<<<<<<<<<
 *     cpdef _generic_coerce_map(self, S)
 *     cdef discover_coerce_map_from(self, S)
 */
struct __pyx_opt_args_4sage_9structure_6parent_6Parent__generic_convert_map {
  int __pyx_n;
  PyObject *category;
};

/* "parent.pxd":66
 *     cdef discover_coerce_map_from(self, S)
 *     cdef discover_convert_map_from(self, S)
 *     cdef discover_action(self, S, op, bint self_on_left, self_el=*, S_el=*)             # <<<<<<<<<<<<<<
 * 
 *     # List consisting of Morphisms (from anything to self)
 */
struct __pyx_opt_args_4sage_9structure_6parent_6Parent_discover_action {
  int __pyx_n;
  PyObject *self_el;
  PyObject *S_el;
};
struct __pyx_opt_args_4sage_10categories_3map_3Map__call_with_args;

/* "map.pxd":15
 *     # these methods require x is an element of domain, and returns an element with parent codomain
 *     cpdef Element _call_(self, x)
 *     cpdef Element _call_with_args(self, x, args=*, kwds=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef public domain    # will be either a weakref or a constant map
 */
struct __pyx_opt_args_4sage_10categories_3map_3Map__call_with_args {
  int __pyx_n;
  PyObject *args;
  PyObject *kwds;
};

/* "ppl/linear_algebra.pxd":3
 * from ppl_decl cimport *
 * 
 * cdef class Variable(object):             # <<<<<<<<<<<<<<
 *     cdef PPL_Variable *thisptr
 * 
 */
struct __pyx_obj_3ppl_14linear_algebra_Variable {
  PyObject_HEAD
  Parma_Polyhedra_Library::Variable *thisptr;
};


/* "ppl/linear_algebra.pxd":6
 *     cdef PPL_Variable *thisptr
 * 
 * cdef class Variables_Set(object):             # <<<<<<<<<<<<<<
 *     cdef PPL_Variables_Set *thisptr
 * 
 */
struct __pyx_obj_3ppl_14linear_algebra_Variables_Set {
  PyObject_HEAD
  Parma_Polyhedra_Library::Variables_Set *thisptr;
};


/* "ppl/linear_algebra.pxd":9
 *     cdef PPL_Variables_Set *thisptr
 * 
 * cdef class Linear_Expression(object):             # <<<<<<<<<<<<<<
 *     cdef PPL_Linear_Expression *thisptr
 * 
 */
struct __pyx_obj_3ppl_14linear_algebra_Linear_Expression {
  PyObject_HEAD
  Parma_Polyhedra_Library::Linear_Expression *thisptr;
};


/* "ppl/generator.pxd":10
 * cdef PPL_GeneratorType_str(PPL_GeneratorType t)
 * 
 * cdef class Generator(object):             # <<<<<<<<<<<<<<
 *     cdef PPL_Generator *thisptr
 * 
 */
struct __pyx_obj_3ppl_9generator_Generator {
  PyObject_HEAD
  Parma_Polyhedra_Library::Generator *thisptr;
};


/* "ppl/generator.pxd":13
 *     cdef PPL_Generator *thisptr
 * 
 * cdef class Generator_System(object):             # <<<<<<<<<<<<<<
 *     cdef PPL_Generator_System *thisptr
 * 
 */
struct __pyx_obj_3ppl_9generator_Generator_System {
  PyObject_HEAD
  Parma_Polyhedra_Library::Generator_System *thisptr;
};


/* "ppl/generator.pxd":16
 *     cdef PPL_Generator_System *thisptr
 * 
 * cdef class Poly_Gen_Relation(object):             # <<<<<<<<<<<<<<
 *     cdef PPL_Poly_Gen_Relation *thisptr
 */
struct __pyx_obj_3ppl_9generator_Poly_Gen_Relation {
  PyObject_HEAD
  Parma_Polyhedra_Library::Poly_Gen_Relation *thisptr;
};


/* "constraint.pxd":9
 * cdef _make_Constraint_from_richcmp(lhs_, rhs_, op)
 * 
 * cdef class Constraint(object):             # <<<<<<<<<<<<<<
 *     cdef PPL_Constraint *thisptr
 * 
 */
struct __pyx_obj_3ppl_10constraint_Constraint {
  PyObject_HEAD
  Parma_Polyhedra_Library::Constraint *thisptr;
};


/* "constraint.pxd":12
 *     cdef PPL_Constraint *thisptr
 * 
 * cdef class Constraint_System(object):             # <<<<<<<<<<<<<<
 *     cdef PPL_Constraint_System *thisptr
 * 
 */
struct __pyx_obj_3ppl_10constraint_Constraint_System {
  PyObject_HEAD
  Parma_Polyhedra_Library::Constraint_System *thisptr;
};


/* "constraint.pxd":15
 *     cdef PPL_Constraint_System *thisptr
 * 
 * cdef class Poly_Con_Relation(object):             # <<<<<<<<<<<<<<
 *     cdef PPL_Poly_Con_Relation *thisptr
 */
struct __pyx_obj_3ppl_10constraint_Poly_Con_Relation {
  PyObject_HEAD
  Parma_Polyhedra_Library::Poly_Con_Relation *thisptr;
};


/* "ppl/polyhedron.pxd":5
 * from .constraint cimport Constraint
 * 
 * cdef class Polyhedron(object):             # <<<<<<<<<<<<<<
 *    cdef PPL_Polyhedron *thisptr
 *    cdef _relation_with_generator(Polyhedron self, Generator g)
 */
struct __pyx_obj_3ppl_10polyhedron_Polyhedron {
  PyObject_HEAD
  struct __pyx_vtabstruct_3ppl_10polyhedron_Polyhedron *__pyx_vtab;
  Parma_Polyhedra_Library::Polyhedron *thisptr;
};


/* "ppl/polyhedron.pxd":10
 *    cdef _relation_with_constraint(Polyhedron self, Constraint c)
 * 
 * cdef class C_Polyhedron(Polyhedron):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_3ppl_10polyhedron_C_Polyhedron {
  struct __pyx_obj_3ppl_10polyhedron_Polyhedron __pyx_base;
};


/* "ppl/polyhedron.pxd":13
 *     pass
 * 
 * cdef class NNC_Polyhedron(Polyhedron):             # <<<<<<<<<<<<<<
 *     pass
 */
struct __pyx_obj_3ppl_10polyhedron_NNC_Polyhedron {
  struct __pyx_obj_3ppl_10polyhedron_Polyhedron __pyx_base;
};


/* "sage_object.pxd":1
 * cdef class SageObject:             # <<<<<<<<<<<<<<
 *     pass
 */
struct __pyx_obj_4sage_9structure_11sage_object_SageObject {
  PyObject_HEAD
};


/* "sage/structure/category_object.pxd":15
 * cpdef check_default_category(default_category, category)
 * 
 * cdef class CategoryObject(SageObject):             # <<<<<<<<<<<<<<
 *     cdef public dict __cached_methods
 *     cdef _category
 */
struct __pyx_obj_4sage_9structure_15category_object_CategoryObject {
  struct __pyx_obj_4sage_9structure_11sage_object_SageObject __pyx_base;
  struct __pyx_vtabstruct_4sage_9structure_15category_object_CategoryObject *__pyx_vtab;
  PyObject *__cached_methods;
  PyObject *_category;
  PyObject *_base;
  PyObject *_names;
  PyObject *_factory_data;
  PyObject *__weakref__;
  long _hash_value;
};


/* "sage/structure/coerce_dict.pxd":12
 * 
 * @cython.final
 * cdef class MonoDict:             # <<<<<<<<<<<<<<
 *     cdef __weakref__
 *     cdef size_t mask  # size - 1 with 'size' the length of the table array
 */
struct __pyx_obj_4sage_9structure_11coerce_dict_MonoDict {
  PyObject_HEAD
  struct __pyx_vtabstruct_4sage_9structure_11coerce_dict_MonoDict *__pyx_vtab;
  PyObject *__weakref__;
  size_t mask;
  size_t used;
  size_t fill;
  struct __pyx_t_4sage_9structure_11coerce_dict_mono_cell *table;
  int weak_values;
  PyObject *eraser;
};


/* "sage/structure/coerce_dict.pxd":38
 * 
 * @cython.final
 * cdef class TripleDict:             # <<<<<<<<<<<<<<
 *     cdef __weakref__
 *     cdef size_t mask  # size - 1 with 'size' the length of the table array
 */
struct __pyx_obj_4sage_9structure_11coerce_dict_TripleDict {
  PyObject_HEAD
  struct __pyx_vtabstruct_4sage_9structure_11coerce_dict_TripleDict *__pyx_vtab;
  PyObject *__weakref__;
  size_t mask;
  size_t used;
  size_t fill;
  struct __pyx_t_4sage_9structure_11coerce_dict_triple_cell *table;
  int weak_values;
  PyObject *eraser;
};


/* "parent.pxd":12
 * from sage.structure.coerce_dict cimport MonoDict, TripleDict
 * 
 * cdef class Parent(sage.structure.category_object.CategoryObject):             # <<<<<<<<<<<<<<
 *     cdef _element_constructor
 *     cdef bint _element_init_pass_parent
 */
struct __pyx_obj_4sage_9structure_6parent_Parent {
  struct __pyx_obj_4sage_9structure_15category_object_CategoryObject __pyx_base;
  PyObject *_element_constructor;
  int _element_init_pass_parent;
  PyObject *_convert_method_name;
  PyObject *_initial_coerce_list;
  PyObject *_initial_action_list;
  PyObject *_initial_convert_list;
  int _coercions_used;
  int flags;
  PyObject *_cache_an_element;
  PyObject *_coerce_from_list;
  PyObject *_registered_domains;
  struct __pyx_obj_4sage_9structure_11coerce_dict_MonoDict *_coerce_from_hash;
  PyObject *_action_list;
  PyObject *_convert_from_list;
  struct __pyx_obj_4sage_9structure_11coerce_dict_MonoDict *_convert_from_hash;
  PyObject *_embedding;
  struct __pyx_obj_4sage_9structure_11coerce_dict_TripleDict *_action_hash;
};


/* "parent.pxd":103
 * 
 * 
 * cdef class Set_generic(Parent):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4sage_9structure_6parent_Set_generic {
  struct __pyx_obj_4sage_9structure_6parent_Parent __pyx_base;
};


/* "sage/misc/inherit_comparison.pxd":3
 * cimport sage.cpython.cython_metaclass
 * 
 * cdef class InheritComparisonMetaclass(type):             # <<<<<<<<<<<<<<
 *     pass
 */
struct __pyx_obj_4sage_4misc_18inherit_comparison_InheritComparisonMetaclass {
  PyHeapTypeObject __pyx_base;
};


/* "sage/structure/element.pxd":150
 * 
 * 
 * cdef class Element(SageObject):             # <<<<<<<<<<<<<<
 *     cdef Parent _parent
 *     cpdef _richcmp_(left, right, int op)
 */
struct __pyx_obj_4sage_9structure_7element_Element {
  struct __pyx_obj_4sage_9structure_11sage_object_SageObject __pyx_base;
  struct __pyx_vtabstruct_4sage_9structure_7element_Element *__pyx_vtab;
  struct __pyx_obj_4sage_9structure_6parent_Parent *_parent;
};


/* "sage/structure/element.pxd":177
 * 
 * 
 * cdef class ElementWithCachedMethod(Element):             # <<<<<<<<<<<<<<
 *     cdef public dict __cached_methods
 * 
 */
struct __pyx_obj_4sage_9structure_7element_ElementWithCachedMethod {
  struct __pyx_obj_4sage_9structure_7element_Element __pyx_base;
  PyObject *__cached_methods;
};


/* "sage/structure/element.pxd":180
 *     cdef public dict __cached_methods
 * 
 * cdef class ModuleElement(Element)       # forward declaration             # <<<<<<<<<<<<<<
 * 
 * cdef class RingElement(ModuleElement)   # forward declaration
 */
struct __pyx_obj_4sage_9structure_7element_ModuleElement {
  struct __pyx_obj_4sage_9structure_7element_Element __pyx_base;
};


/* "sage/structure/element.pxd":194
 *     cpdef _rmul_(self, Element left)
 * 
 * cdef class MonoidElement(Element):             # <<<<<<<<<<<<<<
 *     cpdef _pow_int(self, n)
 * 
 */
struct __pyx_obj_4sage_9structure_7element_MonoidElement {
  struct __pyx_obj_4sage_9structure_7element_Element __pyx_base;
};


/* "sage/structure/element.pxd":197
 *     cpdef _pow_int(self, n)
 * 
 * cdef class MultiplicativeGroupElement(MonoidElement):             # <<<<<<<<<<<<<<
 *     cpdef _div_(self, other)
 * 
 */
struct __pyx_obj_4sage_9structure_7element_MultiplicativeGroupElement {
  struct __pyx_obj_4sage_9structure_7element_MonoidElement __pyx_base;
};


/* "sage/structure/element.pxd":200
 *     cpdef _div_(self, other)
 * 
 * cdef class AdditiveGroupElement(ModuleElement):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4sage_9structure_7element_AdditiveGroupElement {
  struct __pyx_obj_4sage_9structure_7element_ModuleElement __pyx_base;
};


/* "sage/structure/element.pxd":182
 * cdef class ModuleElement(Element)       # forward declaration
 * 
 * cdef class RingElement(ModuleElement)   # forward declaration             # <<<<<<<<<<<<<<
 * 
 * cdef class ModuleElement(Element):
 */
struct __pyx_obj_4sage_9structure_7element_RingElement {
  struct __pyx_obj_4sage_9structure_7element_ModuleElement __pyx_base;
};


/* "sage/structure/element.pxd":208
 *     cpdef _pow_int(self, n)
 * 
 * cdef class CommutativeRingElement(RingElement):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4sage_9structure_7element_CommutativeRingElement {
  struct __pyx_obj_4sage_9structure_7element_RingElement __pyx_base;
};


/* "sage/structure/element.pxd":211
 *     pass
 * 
 * cdef class IntegralDomainElement(CommutativeRingElement):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4sage_9structure_7element_IntegralDomainElement {
  struct __pyx_obj_4sage_9structure_7element_CommutativeRingElement __pyx_base;
};


/* "sage/structure/element.pxd":214
 *     pass
 * 
 * cdef class DedekindDomainElement(IntegralDomainElement):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4sage_9structure_7element_DedekindDomainElement {
  struct __pyx_obj_4sage_9structure_7element_IntegralDomainElement __pyx_base;
};


/* "sage/structure/element.pxd":217
 *     pass
 * 
 * cdef class PrincipalIdealDomainElement(DedekindDomainElement):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4sage_9structure_7element_PrincipalIdealDomainElement {
  struct __pyx_obj_4sage_9structure_7element_DedekindDomainElement __pyx_base;
};


/* "sage/structure/element.pxd":220
 *     pass
 * 
 * cdef class EuclideanDomainElement(PrincipalIdealDomainElement):             # <<<<<<<<<<<<<<
 *     cpdef _floordiv_(self, other)
 *     cpdef _mod_(self, other)
 */
struct __pyx_obj_4sage_9structure_7element_EuclideanDomainElement {
  struct __pyx_obj_4sage_9structure_7element_PrincipalIdealDomainElement __pyx_base;
};


/* "sage/structure/element.pxd":224
 *     cpdef _mod_(self, other)
 * 
 * cdef class FieldElement(CommutativeRingElement):             # <<<<<<<<<<<<<<
 *     cpdef _floordiv_(self, other)
 * 
 */
struct __pyx_obj_4sage_9structure_7element_FieldElement {
  struct __pyx_obj_4sage_9structure_7element_CommutativeRingElement __pyx_base;
};


/* "sage/structure/element.pxd":227
 *     cpdef _floordiv_(self, other)
 * 
 * cdef class AlgebraElement(RingElement):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4sage_9structure_7element_AlgebraElement {
  struct __pyx_obj_4sage_9structure_7element_RingElement __pyx_base;
};


/* "sage/structure/element.pxd":230
 *     pass
 * 
 * cdef class CommutativeAlgebraElement(CommutativeRingElement):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4sage_9structure_7element_CommutativeAlgebraElement {
  struct __pyx_obj_4sage_9structure_7element_CommutativeRingElement __pyx_base;
};


/* "sage/structure/element.pxd":233
 *     pass
 * 
 * cdef class InfinityElement(RingElement):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4sage_9structure_7element_InfinityElement {
  struct __pyx_obj_4sage_9structure_7element_RingElement __pyx_base;
};


/* "sage/structure/element.pxd":237
 * 
 * 
 * cdef class Vector(ModuleElement):             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t _degree
 * 
 */
struct __pyx_obj_4sage_9structure_7element_Vector {
  struct __pyx_obj_4sage_9structure_7element_ModuleElement __pyx_base;
  Py_ssize_t _degree;
};


/* "sage/structure/element.pxd":252
 * 
 * 
 * cdef class Matrix(ModuleElement):             # <<<<<<<<<<<<<<
 *     # All matrix classes must be written in Cython
 *     cdef Py_ssize_t _nrows
 */
struct __pyx_obj_4sage_9structure_7element_Matrix {
  struct __pyx_obj_4sage_9structure_7element_ModuleElement __pyx_base;
  Py_ssize_t _nrows;
  Py_ssize_t _ncols;
};


/* "map.pxd":4
 * from sage.structure.element cimport Element
 * 
 * cdef class Map(Element):             # <<<<<<<<<<<<<<
 *     cdef object __weakref__
 * 
 */
struct __pyx_obj_4sage_10categories_3map_Map {
  struct __pyx_obj_4sage_9structure_7element_Element __pyx_base;
  PyObject *__weakref__;
  int _coerce_cost;
  PyObject *domain;
  PyObject *codomain;
  struct __pyx_obj_4sage_9structure_6parent_Parent *_codomain;
  PyObject *_category_for;
  PyObject *_repr_type_str;
  int _is_coercion;
};


/* "map.pxd":28
 * 
 * 
 * cdef class Section(Map):             # <<<<<<<<<<<<<<
 *     cdef Map _inverse
 * 
 */
struct __pyx_obj_4sage_10categories_3map_Section {
  struct __pyx_obj_4sage_10categories_3map_Map __pyx_base;
  struct __pyx_obj_4sage_10categories_3map_Map *_inverse;
};


/* "map.pxd":31
 *     cdef Map _inverse
 * 
 * cdef class FormalCompositeMap(Map):             # <<<<<<<<<<<<<<
 *     cdef __list
 */
struct __pyx_obj_4sage_10categories_3map_FormalCompositeMap {
  struct __pyx_obj_4sage_10categories_3map_Map __pyx_base;
  PyObject *__pyx___list;
};


/* "sage/categories/morphism.pxd":5
 * 
 * 
 * cdef class Morphism(Map):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_4sage_10categories_8morphism_Morphism {
  struct __pyx_obj_4sage_10categories_3map_Map __pyx_base;
};


/* "sage/categories/morphism.pxd":8
 *     pass
 * 
 * cdef class SetMorphism(Morphism):             # <<<<<<<<<<<<<<
 *     cdef object _function
 *     cpdef bint _eq_c_impl(left, Element right)
 */
struct __pyx_obj_4sage_10categories_8morphism_SetMorphism {
  struct __pyx_obj_4sage_10categories_8morphism_Morphism __pyx_base;
  PyObject *_function;
};


/* "sage/rings/integer.pxd":7
 * from sage.categories.morphism cimport Morphism
 * 
 * cdef class Integer(EuclideanDomainElement):             # <<<<<<<<<<<<<<
 *     # This is really of type mpz_t, but we don't use the mpz_t typedef
 *     # to work around Cython bug
 */
struct __pyx_obj_4sage_5rings_7integer_Integer {
  struct __pyx_obj_4sage_9structure_7element_EuclideanDomainElement __pyx_base;
  __mpz_struct value[1];
};


/* "sage/rings/integer.pxd":41
 * cdef Integer smallInteger(long value)
 * 
 * cdef class int_to_Z(Morphism):             # <<<<<<<<<<<<<<
 *     pass
 */
struct __pyx_obj_4sage_5rings_7integer_int_to_Z {
  struct __pyx_obj_4sage_10categories_8morphism_Morphism __pyx_base;
};


/* "surface_dynamics/interval_exchanges/iet_family.pyx":140
 *     perm[i + 1] = a
 * 
 * cdef class IETFamily(object):             # <<<<<<<<<<<<<<
 *     r"""
 *     A linear family of interval exchange transformations
 */
struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily {
  PyObject_HEAD
  struct __pyx_vtabstruct_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_vtab;
  int *perm;
  size_t dim;
  mpz_t *entries;
  mpz_t **rows;
  size_t alloc;
  size_t length;
  struct __pyx_obj_3ppl_10polyhedron_C_Polyhedron *poly;
};


/* "surface_dynamics/interval_exchanges/iet_family.pyx":274
 *         return [F([self.ray_coefficient(i, j) for j in range(self.dim)]) for i in range(self.length)]
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         r"""
 *         TESTS::
 */
struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__ {
  PyObject_HEAD
  size_t __pyx_v_i;
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self;
};


/* "surface_dynamics/interval_exchanges/iet_family.pyx":286
 *         cdef size_t i, j
 *         s = []
 *         s.append('top ' + ' '.join(str(self.perm[i]) for i in range(self.dim)))             # <<<<<<<<<<<<<<
 *         s.append('bot ' + ' '.join(str(self.perm[self.dim+i]) for i in range(self.dim)))
 *         for i in range(self.length):
 */
struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr {
  PyObject_HEAD
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__ *__pyx_outer_scope;
  size_t __pyx_v_i;
  size_t __pyx_t_0;
  size_t __pyx_t_1;
  size_t __pyx_t_2;
};


/* "surface_dynamics/interval_exchanges/iet_family.pyx":287
 *         s = []
 *         s.append('top ' + ' '.join(str(self.perm[i]) for i in range(self.dim)))
 *         s.append('bot ' + ' '.join(str(self.perm[self.dim+i]) for i in range(self.dim)))             # <<<<<<<<<<<<<<
 *         for i in range(self.length):
 *             s.append(' '.join(str(self.ray_coefficient(i, j)) for j in range(self.dim)))
 */
struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr {
  PyObject_HEAD
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__ *__pyx_outer_scope;
  PyObject *__pyx_v_i;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "surface_dynamics/interval_exchanges/iet_family.pyx":289
 *         s.append('bot ' + ' '.join(str(self.perm[self.dim+i]) for i in range(self.dim)))
 *         for i in range(self.length):
 *             s.append(' '.join(str(self.ray_coefficient(i, j)) for j in range(self.dim)))             # <<<<<<<<<<<<<<
 *         return '\n'.join(s)
 * 
 */
struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr {
  PyObject_HEAD
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__ *__pyx_outer_scope;
  size_t __pyx_v_j;
  size_t __pyx_t_0;
  size_t __pyx_t_1;
  size_t __pyx_t_2;
};


/* "surface_dynamics/interval_exchanges/iet_family.pyx":598
 *         return ans
 * 
 *     def tree(self, max_depth=5, verbose=False):             # <<<<<<<<<<<<<<
 *         r"""
 *         The tree should be smarter in several ways:
 */
struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree {
  PyObject_HEAD
  PyObject *__pyx_v_branch;
  PyObject *__pyx_v_f;
  PyObject *__pyx_v_ff;
  PyObject *__pyx_v_max_depth;
  PyObject *__pyx_v_s;
  PyObject *__pyx_v_seen;
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self;
  PyObject *__pyx_v_ss;
  PyObject *__pyx_v_verbose;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "surface_dynamics/interval_exchanges/iet_family.pyx":667
 *             seen.add(f)
 * 
 *     def random_element(self, ring, *args, **kwds):             # <<<<<<<<<<<<<<
 *         r"""
 *         Return a random point in the cone
 */
struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element {
  PyObject_HEAD
  PyObject *__pyx_v_coeffs;
  size_t __pyx_v_j;
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self;
};


/* "surface_dynamics/interval_exchanges/iet_family.pyx":703
 *                 c = ring.random_element(*args, **kwds)
 *             coeffs.append(c.abs())
 *         lengths = [sum(coeffs[i] * self.ray_coefficient(i, j) for i in range(self.length)) for j in range(self.dim)]             # <<<<<<<<<<<<<<
 * 
 *         return IET(p, lengths)
 */
struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr {
  PyObject_HEAD
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element *__pyx_outer_scope;
  PyObject *__pyx_v_i;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "surface_dynamics/interval_exchanges/iet_family.pyx":764
 *         print('%d saddles found on %d random points' % (num_saddles, num_exp))
 * 
 *     def random_integer_point_statistics(self, num_exp=100):             # <<<<<<<<<<<<<<
 *         r"""
 *         OUTPUT: a dictionary whose keys are integers ``i`` and values are the
 */
struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics {
  PyObject_HEAD
  PyObject *__pyx_v_ZZ;
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self;
};


/* "surface_dynamics/interval_exchanges/iet_family.pyx":777
 *         cdef dict cyls = {}
 *         for _ in range(num_exp):
 *             lengths = sum(ZZ.random_element(0,100000000) * r for r in self.rays())             # <<<<<<<<<<<<<<
 *             ncyls = iet_interface.number_of_cylinders(top, bot, lengths)
 *             if ncyls not in cyls:
 */
struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr {
  PyObject_HEAD
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics *__pyx_outer_scope;
  PyObject *__pyx_v_r;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};



/* "ppl/polyhedron.pxd":5
 * from .constraint cimport Constraint
 * 
 * cdef class Polyhedron(object):             # <<<<<<<<<<<<<<
 *    cdef PPL_Polyhedron *thisptr
 *    cdef _relation_with_generator(Polyhedron self, Generator g)
 */

struct __pyx_vtabstruct_3ppl_10polyhedron_Polyhedron {
  PyObject *(*_relation_with_generator)(struct __pyx_obj_3ppl_10polyhedron_Polyhedron *, struct __pyx_obj_3ppl_9generator_Generator *);
  PyObject *(*_relation_with_constraint)(struct __pyx_obj_3ppl_10polyhedron_Polyhedron *, struct __pyx_obj_3ppl_10constraint_Constraint *);
};
static struct __pyx_vtabstruct_3ppl_10polyhedron_Polyhedron *__pyx_vtabptr_3ppl_10polyhedron_Polyhedron;


/* "ppl/polyhedron.pxd":10
 *    cdef _relation_with_constraint(Polyhedron self, Constraint c)
 * 
 * cdef class C_Polyhedron(Polyhedron):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_3ppl_10polyhedron_C_Polyhedron {
  struct __pyx_vtabstruct_3ppl_10polyhedron_Polyhedron __pyx_base;
};
static struct __pyx_vtabstruct_3ppl_10polyhedron_C_Polyhedron *__pyx_vtabptr_3ppl_10polyhedron_C_Polyhedron;


/* "ppl/polyhedron.pxd":13
 *     pass
 * 
 * cdef class NNC_Polyhedron(Polyhedron):             # <<<<<<<<<<<<<<
 *     pass
 */

struct __pyx_vtabstruct_3ppl_10polyhedron_NNC_Polyhedron {
  struct __pyx_vtabstruct_3ppl_10polyhedron_Polyhedron __pyx_base;
};
static struct __pyx_vtabstruct_3ppl_10polyhedron_NNC_Polyhedron *__pyx_vtabptr_3ppl_10polyhedron_NNC_Polyhedron;


/* "sage/structure/category_object.pxd":15
 * cpdef check_default_category(default_category, category)
 * 
 * cdef class CategoryObject(SageObject):             # <<<<<<<<<<<<<<
 *     cdef public dict __cached_methods
 *     cdef _category
 */

struct __pyx_vtabstruct_4sage_9structure_15category_object_CategoryObject {
  PyObject *(*getattr_from_category)(struct __pyx_obj_4sage_9structure_15category_object_CategoryObject *, PyObject *);
};
static struct __pyx_vtabstruct_4sage_9structure_15category_object_CategoryObject *__pyx_vtabptr_4sage_9structure_15category_object_CategoryObject;


/* "sage/structure/coerce_dict.pxd":12
 * 
 * @cython.final
 * cdef class MonoDict:             # <<<<<<<<<<<<<<
 *     cdef __weakref__
 *     cdef size_t mask  # size - 1 with 'size' the length of the table array
 */

struct __pyx_vtabstruct_4sage_9structure_11coerce_dict_MonoDict {
  struct __pyx_t_4sage_9structure_11coerce_dict_mono_cell *(*lookup)(struct __pyx_obj_4sage_9structure_11coerce_dict_MonoDict *, PyObject *);
  PyObject *(*get)(struct __pyx_obj_4sage_9structure_11coerce_dict_MonoDict *, PyObject *);
  int (*set)(struct __pyx_obj_4sage_9structure_11coerce_dict_MonoDict *, PyObject *, PyObject *);
  int (*resize)(struct __pyx_obj_4sage_9structure_11coerce_dict_MonoDict *);
};
static struct __pyx_vtabstruct_4sage_9structure_11coerce_dict_MonoDict *__pyx_vtabptr_4sage_9structure_11coerce_dict_MonoDict;


/* "sage/structure/coerce_dict.pxd":38
 * 
 * @cython.final
 * cdef class TripleDict:             # <<<<<<<<<<<<<<
 *     cdef __weakref__
 *     cdef size_t mask  # size - 1 with 'size' the length of the table array
 */

struct __pyx_vtabstruct_4sage_9structure_11coerce_dict_TripleDict {
  struct __pyx_t_4sage_9structure_11coerce_dict_triple_cell *(*lookup)(struct __pyx_obj_4sage_9structure_11coerce_dict_TripleDict *, PyObject *, PyObject *, PyObject *);
  PyObject *(*get)(struct __pyx_obj_4sage_9structure_11coerce_dict_TripleDict *, PyObject *, PyObject *, PyObject *);
  int (*set)(struct __pyx_obj_4sage_9structure_11coerce_dict_TripleDict *, PyObject *, PyObject *, PyObject *, PyObject *);
  int (*resize)(struct __pyx_obj_4sage_9structure_11coerce_dict_TripleDict *);
};
static struct __pyx_vtabstruct_4sage_9structure_11coerce_dict_TripleDict *__pyx_vtabptr_4sage_9structure_11coerce_dict_TripleDict;


/* "parent.pxd":12
 * from sage.structure.coerce_dict cimport MonoDict, TripleDict
 * 
 * cdef class Parent(sage.structure.category_object.CategoryObject):             # <<<<<<<<<<<<<<
 *     cdef _element_constructor
 *     cdef bint _element_init_pass_parent
 */

struct __pyx_vtabstruct_4sage_9structure_6parent_Parent {
  struct __pyx_vtabstruct_4sage_9structure_15category_object_CategoryObject __pyx_base;
  int (*get_flag)(struct __pyx_obj_4sage_9structure_6parent_Parent *, int);
  PyObject *(*register_coercion)(struct __pyx_obj_4sage_9structure_6parent_Parent *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*register_action)(struct __pyx_obj_4sage_9structure_6parent_Parent *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*register_conversion)(struct __pyx_obj_4sage_9structure_6parent_Parent *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*register_embedding)(struct __pyx_obj_4sage_9structure_6parent_Parent *, PyObject *, int __pyx_skip_dispatch);
  int (*is_exact)(struct __pyx_obj_4sage_9structure_6parent_Parent *, int __pyx_skip_dispatch);
  int (*init_coerce)(struct __pyx_obj_4sage_9structure_6parent_Parent *, struct __pyx_opt_args_4sage_9structure_6parent_6Parent_init_coerce *__pyx_optional_args);
  int (*has_coerce_map_from)(struct __pyx_obj_4sage_9structure_6parent_Parent *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*coerce_map_from)(struct __pyx_obj_4sage_9structure_6parent_Parent *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_internal_coerce_map_from)(struct __pyx_obj_4sage_9structure_6parent_Parent *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_coerce_map_from_)(struct __pyx_obj_4sage_9structure_6parent_Parent *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*convert_map_from)(struct __pyx_obj_4sage_9structure_6parent_Parent *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_internal_convert_map_from)(struct __pyx_obj_4sage_9structure_6parent_Parent *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_convert_map_from_)(struct __pyx_obj_4sage_9structure_6parent_Parent *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*convert_method_map)(struct __pyx_obj_4sage_9structure_6parent_Parent *, PyObject *, PyObject *);
  PyObject *(*get_action)(struct __pyx_obj_4sage_9structure_6parent_Parent *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4sage_9structure_6parent_6Parent_get_action *__pyx_optional_args);
  PyObject *(*_get_action_)(struct __pyx_obj_4sage_9structure_6parent_Parent *, PyObject *, PyObject *, int, int __pyx_skip_dispatch);
  PyObject *(*coerce)(struct __pyx_obj_4sage_9structure_6parent_Parent *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*an_element)(struct __pyx_obj_4sage_9structure_6parent_Parent *, int __pyx_skip_dispatch);
  PyObject *(*_generic_convert_map)(struct __pyx_obj_4sage_9structure_6parent_Parent *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4sage_9structure_6parent_6Parent__generic_convert_map *__pyx_optional_args);
  PyObject *(*_generic_coerce_map)(struct __pyx_obj_4sage_9structure_6parent_Parent *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*discover_coerce_map_from)(struct __pyx_obj_4sage_9structure_6parent_Parent *, PyObject *);
  PyObject *(*discover_convert_map_from)(struct __pyx_obj_4sage_9structure_6parent_Parent *, PyObject *);
  PyObject *(*discover_action)(struct __pyx_obj_4sage_9structure_6parent_Parent *, PyObject *, PyObject *, int, struct __pyx_opt_args_4sage_9structure_6parent_6Parent_discover_action *__pyx_optional_args);
};
static struct __pyx_vtabstruct_4sage_9structure_6parent_Parent *__pyx_vtabptr_4sage_9structure_6parent_Parent;
static CYTHON_INLINE int __pyx_f_4sage_9structure_6parent_6Parent_get_flag(struct __pyx_obj_4sage_9structure_6parent_Parent *, int);


/* "parent.pxd":103
 * 
 * 
 * cdef class Set_generic(Parent):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4sage_9structure_6parent_Set_generic {
  struct __pyx_vtabstruct_4sage_9structure_6parent_Parent __pyx_base;
};
static struct __pyx_vtabstruct_4sage_9structure_6parent_Set_generic *__pyx_vtabptr_4sage_9structure_6parent_Set_generic;


/* "sage/structure/element.pxd":150
 * 
 * 
 * cdef class Element(SageObject):             # <<<<<<<<<<<<<<
 *     cdef Parent _parent
 *     cpdef _richcmp_(left, right, int op)
 */

struct __pyx_vtabstruct_4sage_9structure_7element_Element {
  PyObject *(*_richcmp_)(struct __pyx_obj_4sage_9structure_7element_Element *, PyObject *, int, int __pyx_skip_dispatch);
  int (*_cmp_)(struct __pyx_obj_4sage_9structure_7element_Element *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*base_extend)(struct __pyx_obj_4sage_9structure_7element_Element *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*getattr_from_category)(struct __pyx_obj_4sage_9structure_7element_Element *, PyObject *);
  PyObject *(*_act_on_)(struct __pyx_obj_4sage_9structure_7element_Element *, PyObject *, int, int __pyx_skip_dispatch);
  PyObject *(*_acted_upon_)(struct __pyx_obj_4sage_9structure_7element_Element *, PyObject *, int, int __pyx_skip_dispatch);
  PyObject *(*_add_)(struct __pyx_obj_4sage_9structure_7element_Element *, PyObject *);
  PyObject *(*_sub_)(struct __pyx_obj_4sage_9structure_7element_Element *, PyObject *);
  PyObject *(*_neg_)(struct __pyx_obj_4sage_9structure_7element_Element *);
  PyObject *(*_add_long)(struct __pyx_obj_4sage_9structure_7element_Element *, long);
  PyObject *(*_mul_)(struct __pyx_obj_4sage_9structure_7element_Element *, PyObject *);
  PyObject *(*_mul_long)(struct __pyx_obj_4sage_9structure_7element_Element *, long);
  PyObject *(*_div_)(struct __pyx_obj_4sage_9structure_7element_Element *, PyObject *);
  PyObject *(*_floordiv_)(struct __pyx_obj_4sage_9structure_7element_Element *, PyObject *);
  PyObject *(*_mod_)(struct __pyx_obj_4sage_9structure_7element_Element *, PyObject *);
  PyObject *(*_pow_)(struct __pyx_obj_4sage_9structure_7element_Element *, PyObject *);
  PyObject *(*_pow_int)(struct __pyx_obj_4sage_9structure_7element_Element *, PyObject *);
  PyObject *(*_pow_long)(struct __pyx_obj_4sage_9structure_7element_Element *, long);
};
static struct __pyx_vtabstruct_4sage_9structure_7element_Element *__pyx_vtabptr_4sage_9structure_7element_Element;


/* "sage/structure/element.pxd":177
 * 
 * 
 * cdef class ElementWithCachedMethod(Element):             # <<<<<<<<<<<<<<
 *     cdef public dict __cached_methods
 * 
 */

struct __pyx_vtabstruct_4sage_9structure_7element_ElementWithCachedMethod {
  struct __pyx_vtabstruct_4sage_9structure_7element_Element __pyx_base;
};
static struct __pyx_vtabstruct_4sage_9structure_7element_ElementWithCachedMethod *__pyx_vtabptr_4sage_9structure_7element_ElementWithCachedMethod;


/* "sage/structure/element.pxd":180
 *     cdef public dict __cached_methods
 * 
 * cdef class ModuleElement(Element)       # forward declaration             # <<<<<<<<<<<<<<
 * 
 * cdef class RingElement(ModuleElement)   # forward declaration
 */

struct __pyx_vtabstruct_4sage_9structure_7element_ModuleElement {
  struct __pyx_vtabstruct_4sage_9structure_7element_Element __pyx_base;
  PyObject *(*_add_)(struct __pyx_obj_4sage_9structure_7element_ModuleElement *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_sub_)(struct __pyx_obj_4sage_9structure_7element_ModuleElement *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_neg_)(struct __pyx_obj_4sage_9structure_7element_ModuleElement *, int __pyx_skip_dispatch);
  PyObject *(*_lmul_)(struct __pyx_obj_4sage_9structure_7element_ModuleElement *, struct __pyx_obj_4sage_9structure_7element_Element *, int __pyx_skip_dispatch);
  PyObject *(*_rmul_)(struct __pyx_obj_4sage_9structure_7element_ModuleElement *, struct __pyx_obj_4sage_9structure_7element_Element *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_4sage_9structure_7element_ModuleElement *__pyx_vtabptr_4sage_9structure_7element_ModuleElement;


/* "sage/structure/element.pxd":182
 * cdef class ModuleElement(Element)       # forward declaration
 * 
 * cdef class RingElement(ModuleElement)   # forward declaration             # <<<<<<<<<<<<<<
 * 
 * cdef class ModuleElement(Element):
 */

struct __pyx_vtabstruct_4sage_9structure_7element_RingElement {
  struct __pyx_vtabstruct_4sage_9structure_7element_ModuleElement __pyx_base;
  PyObject *(*_mul_)(struct __pyx_obj_4sage_9structure_7element_RingElement *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_div_)(struct __pyx_obj_4sage_9structure_7element_RingElement *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_pow_int)(struct __pyx_obj_4sage_9structure_7element_RingElement *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_4sage_9structure_7element_RingElement *__pyx_vtabptr_4sage_9structure_7element_RingElement;


/* "sage/structure/element.pxd":194
 *     cpdef _rmul_(self, Element left)
 * 
 * cdef class MonoidElement(Element):             # <<<<<<<<<<<<<<
 *     cpdef _pow_int(self, n)
 * 
 */

struct __pyx_vtabstruct_4sage_9structure_7element_MonoidElement {
  struct __pyx_vtabstruct_4sage_9structure_7element_Element __pyx_base;
  PyObject *(*_pow_int)(struct __pyx_obj_4sage_9structure_7element_MonoidElement *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_4sage_9structure_7element_MonoidElement *__pyx_vtabptr_4sage_9structure_7element_MonoidElement;


/* "sage/structure/element.pxd":197
 *     cpdef _pow_int(self, n)
 * 
 * cdef class MultiplicativeGroupElement(MonoidElement):             # <<<<<<<<<<<<<<
 *     cpdef _div_(self, other)
 * 
 */

struct __pyx_vtabstruct_4sage_9structure_7element_MultiplicativeGroupElement {
  struct __pyx_vtabstruct_4sage_9structure_7element_MonoidElement __pyx_base;
  PyObject *(*_div_)(struct __pyx_obj_4sage_9structure_7element_MultiplicativeGroupElement *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_4sage_9structure_7element_MultiplicativeGroupElement *__pyx_vtabptr_4sage_9structure_7element_MultiplicativeGroupElement;


/* "sage/structure/element.pxd":200
 *     cpdef _div_(self, other)
 * 
 * cdef class AdditiveGroupElement(ModuleElement):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4sage_9structure_7element_AdditiveGroupElement {
  struct __pyx_vtabstruct_4sage_9structure_7element_ModuleElement __pyx_base;
};
static struct __pyx_vtabstruct_4sage_9structure_7element_AdditiveGroupElement *__pyx_vtabptr_4sage_9structure_7element_AdditiveGroupElement;


/* "sage/structure/element.pxd":208
 *     cpdef _pow_int(self, n)
 * 
 * cdef class CommutativeRingElement(RingElement):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4sage_9structure_7element_CommutativeRingElement {
  struct __pyx_vtabstruct_4sage_9structure_7element_RingElement __pyx_base;
};
static struct __pyx_vtabstruct_4sage_9structure_7element_CommutativeRingElement *__pyx_vtabptr_4sage_9structure_7element_CommutativeRingElement;


/* "sage/structure/element.pxd":211
 *     pass
 * 
 * cdef class IntegralDomainElement(CommutativeRingElement):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4sage_9structure_7element_IntegralDomainElement {
  struct __pyx_vtabstruct_4sage_9structure_7element_CommutativeRingElement __pyx_base;
};
static struct __pyx_vtabstruct_4sage_9structure_7element_IntegralDomainElement *__pyx_vtabptr_4sage_9structure_7element_IntegralDomainElement;


/* "sage/structure/element.pxd":214
 *     pass
 * 
 * cdef class DedekindDomainElement(IntegralDomainElement):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4sage_9structure_7element_DedekindDomainElement {
  struct __pyx_vtabstruct_4sage_9structure_7element_IntegralDomainElement __pyx_base;
};
static struct __pyx_vtabstruct_4sage_9structure_7element_DedekindDomainElement *__pyx_vtabptr_4sage_9structure_7element_DedekindDomainElement;


/* "sage/structure/element.pxd":217
 *     pass
 * 
 * cdef class PrincipalIdealDomainElement(DedekindDomainElement):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4sage_9structure_7element_PrincipalIdealDomainElement {
  struct __pyx_vtabstruct_4sage_9structure_7element_DedekindDomainElement __pyx_base;
};
static struct __pyx_vtabstruct_4sage_9structure_7element_PrincipalIdealDomainElement *__pyx_vtabptr_4sage_9structure_7element_PrincipalIdealDomainElement;


/* "sage/structure/element.pxd":220
 *     pass
 * 
 * cdef class EuclideanDomainElement(PrincipalIdealDomainElement):             # <<<<<<<<<<<<<<
 *     cpdef _floordiv_(self, other)
 *     cpdef _mod_(self, other)
 */

struct __pyx_vtabstruct_4sage_9structure_7element_EuclideanDomainElement {
  struct __pyx_vtabstruct_4sage_9structure_7element_PrincipalIdealDomainElement __pyx_base;
  PyObject *(*_floordiv_)(struct __pyx_obj_4sage_9structure_7element_EuclideanDomainElement *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_mod_)(struct __pyx_obj_4sage_9structure_7element_EuclideanDomainElement *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_4sage_9structure_7element_EuclideanDomainElement *__pyx_vtabptr_4sage_9structure_7element_EuclideanDomainElement;


/* "sage/structure/element.pxd":224
 *     cpdef _mod_(self, other)
 * 
 * cdef class FieldElement(CommutativeRingElement):             # <<<<<<<<<<<<<<
 *     cpdef _floordiv_(self, other)
 * 
 */

struct __pyx_vtabstruct_4sage_9structure_7element_FieldElement {
  struct __pyx_vtabstruct_4sage_9structure_7element_CommutativeRingElement __pyx_base;
  PyObject *(*_floordiv_)(struct __pyx_obj_4sage_9structure_7element_FieldElement *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_4sage_9structure_7element_FieldElement *__pyx_vtabptr_4sage_9structure_7element_FieldElement;


/* "sage/structure/element.pxd":227
 *     cpdef _floordiv_(self, other)
 * 
 * cdef class AlgebraElement(RingElement):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4sage_9structure_7element_AlgebraElement {
  struct __pyx_vtabstruct_4sage_9structure_7element_RingElement __pyx_base;
};
static struct __pyx_vtabstruct_4sage_9structure_7element_AlgebraElement *__pyx_vtabptr_4sage_9structure_7element_AlgebraElement;


/* "sage/structure/element.pxd":230
 *     pass
 * 
 * cdef class CommutativeAlgebraElement(CommutativeRingElement):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4sage_9structure_7element_CommutativeAlgebraElement {
  struct __pyx_vtabstruct_4sage_9structure_7element_CommutativeRingElement __pyx_base;
};
static struct __pyx_vtabstruct_4sage_9structure_7element_CommutativeAlgebraElement *__pyx_vtabptr_4sage_9structure_7element_CommutativeAlgebraElement;


/* "sage/structure/element.pxd":233
 *     pass
 * 
 * cdef class InfinityElement(RingElement):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4sage_9structure_7element_InfinityElement {
  struct __pyx_vtabstruct_4sage_9structure_7element_RingElement __pyx_base;
};
static struct __pyx_vtabstruct_4sage_9structure_7element_InfinityElement *__pyx_vtabptr_4sage_9structure_7element_InfinityElement;


/* "sage/structure/element.pxd":237
 * 
 * 
 * cdef class Vector(ModuleElement):             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t _degree
 * 
 */

struct __pyx_vtabstruct_4sage_9structure_7element_Vector {
  struct __pyx_vtabstruct_4sage_9structure_7element_ModuleElement __pyx_base;
  PyObject *(*_dot_product_)(struct __pyx_obj_4sage_9structure_7element_Vector *, struct __pyx_obj_4sage_9structure_7element_Vector *, int __pyx_skip_dispatch);
  PyObject *(*_dot_product_coerce_)(struct __pyx_obj_4sage_9structure_7element_Vector *, struct __pyx_obj_4sage_9structure_7element_Vector *, int __pyx_skip_dispatch);
  PyObject *(*_pairwise_product_)(struct __pyx_obj_4sage_9structure_7element_Vector *, struct __pyx_obj_4sage_9structure_7element_Vector *, int __pyx_skip_dispatch);
  int (*is_sparse_c)(struct __pyx_obj_4sage_9structure_7element_Vector *);
  int (*is_dense_c)(struct __pyx_obj_4sage_9structure_7element_Vector *);
};
static struct __pyx_vtabstruct_4sage_9structure_7element_Vector *__pyx_vtabptr_4sage_9structure_7element_Vector;


/* "sage/structure/element.pxd":252
 * 
 * 
 * cdef class Matrix(ModuleElement):             # <<<<<<<<<<<<<<
 *     # All matrix classes must be written in Cython
 *     cdef Py_ssize_t _nrows
 */

struct __pyx_vtabstruct_4sage_9structure_7element_Matrix {
  struct __pyx_vtabstruct_4sage_9structure_7element_ModuleElement __pyx_base;
  PyObject *(*_vector_times_matrix_)(struct __pyx_obj_4sage_9structure_7element_Matrix *, struct __pyx_obj_4sage_9structure_7element_Vector *);
  PyObject *(*_matrix_times_vector_)(struct __pyx_obj_4sage_9structure_7element_Matrix *, struct __pyx_obj_4sage_9structure_7element_Vector *);
  PyObject *(*_matrix_times_matrix_)(struct __pyx_obj_4sage_9structure_7element_Matrix *, struct __pyx_obj_4sage_9structure_7element_Matrix *);
  int (*is_sparse_c)(struct __pyx_obj_4sage_9structure_7element_Matrix *);
  int (*is_dense_c)(struct __pyx_obj_4sage_9structure_7element_Matrix *);
};
static struct __pyx_vtabstruct_4sage_9structure_7element_Matrix *__pyx_vtabptr_4sage_9structure_7element_Matrix;


/* "map.pxd":4
 * from sage.structure.element cimport Element
 * 
 * cdef class Map(Element):             # <<<<<<<<<<<<<<
 *     cdef object __weakref__
 * 
 */

struct __pyx_vtabstruct_4sage_10categories_3map_Map {
  struct __pyx_vtabstruct_4sage_9structure_7element_Element __pyx_base;
  PyObject *(*_update_slots)(struct __pyx_obj_4sage_10categories_3map_Map *, PyObject *);
  PyObject *(*_extra_slots)(struct __pyx_obj_4sage_10categories_3map_Map *);
  struct __pyx_obj_4sage_9structure_7element_Element *(*_call_)(struct __pyx_obj_4sage_10categories_3map_Map *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_4sage_9structure_7element_Element *(*_call_with_args)(struct __pyx_obj_4sage_10categories_3map_Map *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_4sage_10categories_3map_3Map__call_with_args *__pyx_optional_args);
  PyObject *(*_pow_int)(struct __pyx_obj_4sage_10categories_3map_Map *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_4sage_10categories_3map_Map *__pyx_vtabptr_4sage_10categories_3map_Map;


/* "map.pxd":28
 * 
 * 
 * cdef class Section(Map):             # <<<<<<<<<<<<<<
 *     cdef Map _inverse
 * 
 */

struct __pyx_vtabstruct_4sage_10categories_3map_Section {
  struct __pyx_vtabstruct_4sage_10categories_3map_Map __pyx_base;
};
static struct __pyx_vtabstruct_4sage_10categories_3map_Section *__pyx_vtabptr_4sage_10categories_3map_Section;


/* "map.pxd":31
 *     cdef Map _inverse
 * 
 * cdef class FormalCompositeMap(Map):             # <<<<<<<<<<<<<<
 *     cdef __list
 */

struct __pyx_vtabstruct_4sage_10categories_3map_FormalCompositeMap {
  struct __pyx_vtabstruct_4sage_10categories_3map_Map __pyx_base;
};
static struct __pyx_vtabstruct_4sage_10categories_3map_FormalCompositeMap *__pyx_vtabptr_4sage_10categories_3map_FormalCompositeMap;


/* "sage/categories/morphism.pxd":5
 * 
 * 
 * cdef class Morphism(Map):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_4sage_10categories_8morphism_Morphism {
  struct __pyx_vtabstruct_4sage_10categories_3map_Map __pyx_base;
};
static struct __pyx_vtabstruct_4sage_10categories_8morphism_Morphism *__pyx_vtabptr_4sage_10categories_8morphism_Morphism;


/* "sage/categories/morphism.pxd":8
 *     pass
 * 
 * cdef class SetMorphism(Morphism):             # <<<<<<<<<<<<<<
 *     cdef object _function
 *     cpdef bint _eq_c_impl(left, Element right)
 */

struct __pyx_vtabstruct_4sage_10categories_8morphism_SetMorphism {
  struct __pyx_vtabstruct_4sage_10categories_8morphism_Morphism __pyx_base;
  int (*_eq_c_impl)(struct __pyx_obj_4sage_10categories_8morphism_SetMorphism *, struct __pyx_obj_4sage_9structure_7element_Element *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_4sage_10categories_8morphism_SetMorphism *__pyx_vtabptr_4sage_10categories_8morphism_SetMorphism;


/* "sage/rings/integer.pxd":7
 * from sage.categories.morphism cimport Morphism
 * 
 * cdef class Integer(EuclideanDomainElement):             # <<<<<<<<<<<<<<
 *     # This is really of type mpz_t, but we don't use the mpz_t typedef
 *     # to work around Cython bug
 */

struct __pyx_vtabstruct_4sage_5rings_7integer_Integer {
  struct __pyx_vtabstruct_4sage_9structure_7element_EuclideanDomainElement __pyx_base;
  int (*_to_ZZ)(struct __pyx_obj_4sage_5rings_7integer_Integer *, ZZ *);
  void (*set_from_mpz)(struct __pyx_obj_4sage_5rings_7integer_Integer *, __mpz_struct *);
  PyObject *(*hash_c)(struct __pyx_obj_4sage_5rings_7integer_Integer *);
  PyObject *(*__pyx___pari__)(struct __pyx_obj_4sage_5rings_7integer_Integer *, int __pyx_skip_dispatch);
  PyObject *(*_shift_helper)(struct __pyx_obj_4sage_5rings_7integer_Integer *, PyObject *, int, int __pyx_skip_dispatch);
  PyObject *(*_pow_)(struct __pyx_obj_4sage_5rings_7integer_Integer *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_and)(struct __pyx_obj_4sage_5rings_7integer_Integer *, struct __pyx_obj_4sage_5rings_7integer_Integer *);
  PyObject *(*_or)(struct __pyx_obj_4sage_5rings_7integer_Integer *, struct __pyx_obj_4sage_5rings_7integer_Integer *);
  PyObject *(*_xor)(struct __pyx_obj_4sage_5rings_7integer_Integer *, struct __pyx_obj_4sage_5rings_7integer_Integer *);
  size_t (*_exact_log_log2_iter)(struct __pyx_obj_4sage_5rings_7integer_Integer *, struct __pyx_obj_4sage_5rings_7integer_Integer *, int __pyx_skip_dispatch);
  size_t (*_exact_log_mpfi_log)(struct __pyx_obj_4sage_5rings_7integer_Integer *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_4sage_9structure_7element_RingElement *(*_valuation)(struct __pyx_obj_4sage_5rings_7integer_Integer *, struct __pyx_obj_4sage_5rings_7integer_Integer *, int __pyx_skip_dispatch);
  PyObject *(*_val_unit)(struct __pyx_obj_4sage_5rings_7integer_Integer *, struct __pyx_obj_4sage_5rings_7integer_Integer *);
  struct __pyx_obj_4sage_5rings_7integer_Integer *(*_divide_knowing_divisible_by)(struct __pyx_obj_4sage_5rings_7integer_Integer *, struct __pyx_obj_4sage_5rings_7integer_Integer *);
  int (*_is_power_of)(struct __pyx_obj_4sage_5rings_7integer_Integer *, struct __pyx_obj_4sage_5rings_7integer_Integer *);
  int (*_pseudoprime_is_prime)(struct __pyx_obj_4sage_5rings_7integer_Integer *, PyObject *);
  PyObject *(*_pari_divisors_small)(struct __pyx_obj_4sage_5rings_7integer_Integer *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_4sage_5rings_7integer_Integer *__pyx_vtabptr_4sage_5rings_7integer_Integer;


/* "sage/rings/integer.pxd":41
 * cdef Integer smallInteger(long value)
 * 
 * cdef class int_to_Z(Morphism):             # <<<<<<<<<<<<<<
 *     pass
 */

struct __pyx_vtabstruct_4sage_5rings_7integer_int_to_Z {
  struct __pyx_vtabstruct_4sage_10categories_8morphism_Morphism __pyx_base;
};
static struct __pyx_vtabstruct_4sage_5rings_7integer_int_to_Z *__pyx_vtabptr_4sage_5rings_7integer_int_to_Z;


/* "surface_dynamics/interval_exchanges/iet_family.pyx":140
 *     perm[i + 1] = a
 * 
 * cdef class IETFamily(object):             # <<<<<<<<<<<<<<
 *     r"""
 *     A linear family of interval exchange transformations
 */

struct __pyx_vtabstruct_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily {
  void (*fit_length)(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *, size_t);
  void (*fill_rays)(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *, Parma_Polyhedra_Library::Generator_System const &);
};
static struct __pyx_vtabstruct_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_vtabptr_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* StringJoin.proto */
#if PY_MAJOR_VERSION < 3
#define __Pyx_PyString_Join __Pyx_PyBytes_Join
#define __Pyx_PyBaseString_Join(s, v) (PyUnicode_CheckExact(s) ? PyUnicode_Join(s, v) : __Pyx_PyBytes_Join(s, v))
#else
#define __Pyx_PyString_Join PyUnicode_Join
#define __Pyx_PyBaseString_Join PyUnicode_Join
#endif
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION < 3
    #define __Pyx_PyBytes_Join _PyString_Join
    #else
    #define __Pyx_PyBytes_Join _PyBytes_Join
    #endif
#else
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values);
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* pyfrozenset_new.proto */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it);

/* PySetContains.proto */
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* pop_index.proto */
static PyObject* __Pyx__PyObject_PopNewIndex(PyObject* L, PyObject* py_ix);
static PyObject* __Pyx__PyObject_PopIndex(PyObject* L, PyObject* py_ix);
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static PyObject* __Pyx__PyList_PopIndex(PyObject* L, PyObject* py_ix, Py_ssize_t ix);
#define __Pyx_PyObject_PopIndex(L, py_ix, ix, is_signed, type, to_py_func) (\
    (likely(PyList_CheckExact(L) && __Pyx_fits_Py_ssize_t(ix, type, is_signed))) ?\
        __Pyx__PyList_PopIndex(L, py_ix, ix) : (\
        (unlikely((py_ix) == Py_None)) ? __Pyx__PyObject_PopNewIndex(L, to_py_func(ix)) :\
            __Pyx__PyObject_PopIndex(L, py_ix)))
#define __Pyx_PyList_PopIndex(L, py_ix, ix, is_signed, type, to_py_func) (\
    __Pyx_fits_Py_ssize_t(ix, type, is_signed) ?\
        __Pyx__PyList_PopIndex(L, py_ix, ix) : (\
        (unlikely((py_ix) == Py_None)) ? __Pyx__PyObject_PopNewIndex(L, to_py_func(ix)) :\
            __Pyx__PyObject_PopIndex(L, py_ix)))
#else
#define __Pyx_PyList_PopIndex(L, py_ix, ix, is_signed, type, to_py_func)\
    __Pyx_PyObject_PopIndex(L, py_ix, ix, is_signed, type, to_py_func)
#define __Pyx_PyObject_PopIndex(L, py_ix, ix, is_signed, type, to_py_func) (\
    (unlikely((py_ix) == Py_None)) ? __Pyx__PyObject_PopNewIndex(L, to_py_func(ix)) :\
        __Pyx__PyObject_PopIndex(L, py_ix))
#endif

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* pop.proto */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L);
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L);
#define __Pyx_PyObject_Pop(L) (likely(PyList_CheckExact(L)) ?\
    __Pyx_PyList_Pop(L) : __Pyx__PyObject_Pop(L))
#else
#define __Pyx_PyList_Pop(L)  __Pyx__PyObject_Pop(L)
#define __Pyx_PyObject_Pop(L)  __Pyx__PyObject_Pop(L)
#endif

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* py_set_remove.proto */
static CYTHON_INLINE int __Pyx_PySet_Remove(PyObject *set, PyObject *key);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* IncludeStringH.proto */
#include <string.h>

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* Print.proto */
static int __Pyx_Print(PyObject*, PyObject *, int);
#if CYTHON_COMPILING_IN_PYPY || PY_MAJOR_VERSION >= 3
static PyObject* __pyx_print = 0;
static PyObject* __pyx_print_kwargs = 0;
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* PrintOne.proto */
static int __Pyx_PrintOne(PyObject* stream, PyObject *o);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static void __pyx_f_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_fit_length(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self, size_t __pyx_v_length); /* proto*/
static void __pyx_f_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_fill_rays(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self, Parma_Polyhedra_Library::Generator_System const &__pyx_v_gs); /* proto*/
static CYTHON_INLINE int __pyx_f_4sage_9structure_6parent_6Parent_get_flag(struct __pyx_obj_4sage_9structure_6parent_Parent *__pyx_v_self, int __pyx_v_flag); /* proto*/

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cython' */

/* Module declarations from 'cysignals.signals' */
static cysigs_t *__pyx_vp_9cysignals_7signals_cysigs = 0;
#define cysigs (*__pyx_vp_9cysignals_7signals_cysigs)
static void (*__pyx_f_9cysignals_7signals_verify_exc_value)(void); /*proto*/
static void (*_sig_on_interrupt_received)(void); /*proto*/
static void (*_sig_on_recover)(void); /*proto*/
static void (*_sig_off_warning)(char const *, int); /*proto*/
static void (*print_backtrace)(void); /*proto*/

/* Module declarations from 'cysignals.memory' */
static CYTHON_INLINE void *sig_malloc(size_t); /*proto*/
static CYTHON_INLINE void *sig_realloc(void *, size_t); /*proto*/
static CYTHON_INLINE void *sig_calloc(size_t, size_t); /*proto*/
static CYTHON_INLINE void sig_free(void *); /*proto*/
static CYTHON_INLINE size_t __pyx_f_9cysignals_6memory_mul_overflowcheck(size_t, size_t); /*proto*/
static CYTHON_INLINE void *__pyx_f_9cysignals_6memory_check_malloc(size_t); /*proto*/
static CYTHON_INLINE void *__pyx_f_9cysignals_6memory_check_realloc(void *, size_t); /*proto*/
static CYTHON_INLINE void *__pyx_f_9cysignals_6memory_check_calloc(size_t, size_t); /*proto*/

/* Module declarations from 'libcpp' */

/* Module declarations from 'libcpp.vector' */

/* Module declarations from 'ppl.ppl_decl' */

/* Module declarations from 'ppl.linear_algebra' */
static PyTypeObject *__pyx_ptype_3ppl_14linear_algebra_Variable = 0;
static PyTypeObject *__pyx_ptype_3ppl_14linear_algebra_Variables_Set = 0;
static PyTypeObject *__pyx_ptype_3ppl_14linear_algebra_Linear_Expression = 0;

/* Module declarations from 'ppl.generator' */
static PyTypeObject *__pyx_ptype_3ppl_9generator_Generator = 0;
static PyTypeObject *__pyx_ptype_3ppl_9generator_Generator_System = 0;
static PyTypeObject *__pyx_ptype_3ppl_9generator_Poly_Gen_Relation = 0;

/* Module declarations from 'ppl.constraint' */
static PyTypeObject *__pyx_ptype_3ppl_10constraint_Constraint = 0;
static PyTypeObject *__pyx_ptype_3ppl_10constraint_Constraint_System = 0;
static PyTypeObject *__pyx_ptype_3ppl_10constraint_Poly_Con_Relation = 0;

/* Module declarations from 'ppl.polyhedron' */
static PyTypeObject *__pyx_ptype_3ppl_10polyhedron_Polyhedron = 0;
static PyTypeObject *__pyx_ptype_3ppl_10polyhedron_C_Polyhedron = 0;
static PyTypeObject *__pyx_ptype_3ppl_10polyhedron_NNC_Polyhedron = 0;

/* Module declarations from 'sage.ext.stdsage' */
static CYTHON_INLINE PyObject *__pyx_f_4sage_3ext_7stdsage_PY_NEW(PyTypeObject *); /*proto*/

/* Module declarations from 'sage.libs.gmp.types' */

/* Module declarations from 'sage.libs.ntl.types' */

/* Module declarations from 'sage.structure.sage_object' */
static PyTypeObject *__pyx_ptype_4sage_9structure_11sage_object_SageObject = 0;

/* Module declarations from 'sage.structure.category_object' */
static PyTypeObject *__pyx_ptype_4sage_9structure_15category_object_CategoryObject = 0;

/* Module declarations from 'sage.structure' */

/* Module declarations from 'sage.cpython.cython_metaclass' */

/* Module declarations from 'sage.cpython' */

/* Module declarations from 'sage' */

/* Module declarations from 'sage.structure.coerce_dict' */
static PyTypeObject *__pyx_ptype_4sage_9structure_11coerce_dict_MonoDict = 0;
static PyTypeObject *__pyx_ptype_4sage_9structure_11coerce_dict_TripleDict = 0;

/* Module declarations from 'sage.structure.parent' */
static PyTypeObject *__pyx_ptype_4sage_9structure_6parent_Parent = 0;
static PyTypeObject *__pyx_ptype_4sage_9structure_6parent_Set_generic = 0;

/* Module declarations from 'sage.misc.inherit_comparison' */
static PyTypeObject *__pyx_ptype_4sage_4misc_18inherit_comparison_InheritComparisonMetaclass = 0;

/* Module declarations from 'sage.structure.element' */
static PyTypeObject *__pyx_ptype_4sage_9structure_7element_Element = 0;
static PyTypeObject *__pyx_ptype_4sage_9structure_7element_ElementWithCachedMethod = 0;
static PyTypeObject *__pyx_ptype_4sage_9structure_7element_ModuleElement = 0;
static PyTypeObject *__pyx_ptype_4sage_9structure_7element_RingElement = 0;
static PyTypeObject *__pyx_ptype_4sage_9structure_7element_MonoidElement = 0;
static PyTypeObject *__pyx_ptype_4sage_9structure_7element_MultiplicativeGroupElement = 0;
static PyTypeObject *__pyx_ptype_4sage_9structure_7element_AdditiveGroupElement = 0;
static PyTypeObject *__pyx_ptype_4sage_9structure_7element_CommutativeRingElement = 0;
static PyTypeObject *__pyx_ptype_4sage_9structure_7element_IntegralDomainElement = 0;
static PyTypeObject *__pyx_ptype_4sage_9structure_7element_DedekindDomainElement = 0;
static PyTypeObject *__pyx_ptype_4sage_9structure_7element_PrincipalIdealDomainElement = 0;
static PyTypeObject *__pyx_ptype_4sage_9structure_7element_EuclideanDomainElement = 0;
static PyTypeObject *__pyx_ptype_4sage_9structure_7element_FieldElement = 0;
static PyTypeObject *__pyx_ptype_4sage_9structure_7element_AlgebraElement = 0;
static PyTypeObject *__pyx_ptype_4sage_9structure_7element_CommutativeAlgebraElement = 0;
static PyTypeObject *__pyx_ptype_4sage_9structure_7element_InfinityElement = 0;
static PyTypeObject *__pyx_ptype_4sage_9structure_7element_Vector = 0;
static PyTypeObject *__pyx_ptype_4sage_9structure_7element_Matrix = 0;
static CYTHON_INLINE PyObject *__pyx_f_4sage_9structure_7element_parent(PyObject *, int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE int __pyx_f_4sage_9structure_7element_classify_elements(PyObject *, PyObject *); /*proto*/
static CYTHON_INLINE int __pyx_f_4sage_9structure_7element_HAVE_SAME_PARENT(int); /*proto*/
static CYTHON_INLINE int __pyx_f_4sage_9structure_7element_have_same_parent(PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/

/* Module declarations from 'sage.categories.map' */
static PyTypeObject *__pyx_ptype_4sage_10categories_3map_Map = 0;
static PyTypeObject *__pyx_ptype_4sage_10categories_3map_Section = 0;
static PyTypeObject *__pyx_ptype_4sage_10categories_3map_FormalCompositeMap = 0;

/* Module declarations from 'sage.categories.morphism' */
static PyTypeObject *__pyx_ptype_4sage_10categories_8morphism_Morphism = 0;
static PyTypeObject *__pyx_ptype_4sage_10categories_8morphism_SetMorphism = 0;

/* Module declarations from 'sage.rings.integer' */
static PyTypeObject *__pyx_ptype_4sage_5rings_7integer_Integer = 0;
static PyTypeObject *__pyx_ptype_4sage_5rings_7integer_int_to_Z = 0;

/* Module declarations from 'surface_dynamics.interval_exchanges.iet_family' */
static PyTypeObject *__pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily = 0;
static PyTypeObject *__pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__ = 0;
static PyTypeObject *__pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr = 0;
static PyTypeObject *__pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr = 0;
static PyTypeObject *__pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr = 0;
static PyTypeObject *__pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree = 0;
static PyTypeObject *__pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element = 0;
static PyTypeObject *__pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr = 0;
static PyTypeObject *__pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics = 0;
static PyTypeObject *__pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr = 0;
static int __pyx_f_16surface_dynamics_18interval_exchanges_10iet_family_compare(void const *, void const *); /*proto*/
static CYTHON_INLINE int __pyx_f_16surface_dynamics_18interval_exchanges_10iet_family_mpz_vec_any_smaller(mpz_t *, mpz_t *, size_t); /*proto*/
static CYTHON_INLINE int __pyx_f_16surface_dynamics_18interval_exchanges_10iet_family_mpz_vec_equal(mpz_t *, mpz_t *, size_t); /*proto*/
static Py_hash_t __pyx_f_16surface_dynamics_18interval_exchanges_10iet_family_mpz_pythonhash(mpz_srcptr); /*proto*/
static CYTHON_INLINE void __pyx_f_16surface_dynamics_18interval_exchanges_10iet_family_rauzy_top(int *, size_t); /*proto*/
static CYTHON_INLINE void __pyx_f_16surface_dynamics_18interval_exchanges_10iet_family_rauzy_bot(int *, size_t); /*proto*/
#define __Pyx_MODULE_NAME "surface_dynamics.interval_exchanges.iet_family"
extern int __pyx_module_is_main_surface_dynamics__interval_exchanges__iet_family;
int __pyx_module_is_main_surface_dynamics__interval_exchanges__iet_family = 0;

/* Implementation of 'surface_dynamics.interval_exchanges.iet_family' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_sum;
static PyObject *__pyx_builtin_MemoryError;
static const char __pyx_k_C[] = "C";
static const char __pyx_k_b[] = "b";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_p[] = "p";
static const char __pyx_k_t[] = "t";
static const char __pyx_k_QQ[] = "QQ";
static const char __pyx_k_ZZ[] = "ZZ";
static const char __pyx_k__5[] = " ";
static const char __pyx_k__6[] = "\n";
static const char __pyx_k__8[] = "";
static const char __pyx_k_IET[] = "IET";
static const char __pyx_k_abs[] = "abs";
static const char __pyx_k_bot[] = "bot ";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_s_f[] = "s = {}\nf =\n{}";
static const char __pyx_k_sum[] = "sum";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_top[] = "top ";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_cone[] = "cone";
static const char __pyx_k_file[] = "file";
static const char __pyx_k_join[] = "join";
static const char __pyx_k_left[] = "left";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_nray[] = "nray";
static const char __pyx_k_rays[] = "rays";
static const char __pyx_k_ring[] = "ring";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_tree[] = "tree";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_flush[] = "flush";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_right[] = "right";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_write[] = "write";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_branch[] = "branch:";
static const char __pyx_k_dim_lu[] = "dim      %lu\n";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_labels[] = "_labels";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_saddle[] = "saddle";
static const char __pyx_k_stdout[] = "stdout";
static const char __pyx_k_autosim[] = "autosim";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_is_zero[] = "is_zero";
static const char __pyx_k_num_exp[] = "num_exp";
static const char __pyx_k_rows_lu[] = "rows     %lu\n";
static const char __pyx_k_verbose[] = "verbose";
static const char __pyx_k_alloc_lu[] = "alloc    %lu\n";
static const char __pyx_k_alphabet[] = "alphabet";
static const char __pyx_k_children[] = "children";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_IETFamily[] = "IETFamily";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_length_lu[] = "length   %lu\n";
static const char __pyx_k_max_depth[] = "max_depth";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_rows_d_lu[] = "rows[%d]  %lu\n";
static const char __pyx_k_FreeModule[] = "FreeModule";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_entries_lu[] = "entries  %lu\n";
static const char __pyx_k_iterations[] = "iterations";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_Permutation[] = "Permutation";
static const char __pyx_k_permutation[] = "permutation";
static const char __pyx_k_zorich_move[] = "zorich_move";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_affine_image[] = "affine_image";
static const char __pyx_k_constructors[] = "constructors";
static const char __pyx_k_iet_interface[] = "iet_interface";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_IETFamily_tree[] = "IETFamily.tree";
static const char __pyx_k_has_connection[] = "has_connection";
static const char __pyx_k_num_iterations[] = "num_iterations";
static const char __pyx_k_random_element[] = "random_element";
static const char __pyx_k_ray_coefficient[] = "ray_coefficient";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_space_dimension[] = "space_dimension";
static const char __pyx_k_affine_dimension[] = "affine_dimension";
static const char __pyx_k_looking_children[] = "looking children";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_dimension_mismatch[] = "dimension mismatch";
static const char __pyx_k_intersection_assign[] = "intersection_assign";
static const char __pyx_k_number_of_cylinders[] = "number_of_cylinders";
static const char __pyx_k_repr___locals_genexpr[] = "__repr__.<locals>.genexpr";
static const char __pyx_k_LabelledPermutationIET[] = "LabelledPermutationIET";
static const char __pyx_k_row_index_out_of_range[] = "row index out of range";
static const char __pyx_k_sage_rings_integer_ring[] = "sage.rings.integer_ring";
static const char __pyx_k_sage_modules_free_module[] = "sage.modules.free_module";
static const char __pyx_k_column_index_out_of_range[] = "column index out of range";
static const char __pyx_k_sage_rings_rational_field[] = "sage.rings.rational_field";
static const char __pyx_k_failed_to_allocate_s_bytes[] = "failed to allocate %s bytes";
static const char __pyx_k_ppl_check_non_negative_cone[] = "ppl_check_non_negative_cone";
static const char __pyx_k_failed_to_allocate_s_s_bytes[] = "failed to allocate %s * %s bytes";
static const char __pyx_k_random_element_locals_genexpr[] = "random_element.<locals>.genexpr";
static const char __pyx_k_add_space_dimensions_and_embed[] = "add_space_dimensions_and_embed";
static const char __pyx_k_Linear_families_of_interval_exc[] = "\nLinear families of interval exchange transformations\n";
static const char __pyx_k_d_saddles_found_on_d_random_poi[] = "%d saddles found on %d random points";
static const char __pyx_k_random_integer_point_statistics[] = "random_integer_point_statistics.<locals>.genexpr";
static const char __pyx_k_surface_dynamics_misc_ppl_utils[] = "surface_dynamics.misc.ppl_utils";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_only_equality_and_difference_tes[] = "only equality and difference tests available";
static const char __pyx_k_p_must_be_a_labelled_permutation[] = "p must be a labelled permutation";
static const char __pyx_k_surface_dynamics_interval_exchan[] = "surface_dynamics.interval_exchanges.labelled";
static const char __pyx_k_surface_dynamics_interval_exchan_2[] = "surface_dynamics.interval_exchanges.iet_family";
static PyObject *__pyx_n_s_C;
static PyObject *__pyx_n_s_FreeModule;
static PyObject *__pyx_n_s_IET;
static PyObject *__pyx_n_s_IETFamily;
static PyObject *__pyx_n_s_IETFamily_tree;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_n_s_LabelledPermutationIET;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_n_s_Permutation;
static PyObject *__pyx_n_s_QQ;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_ZZ;
static PyObject *__pyx_kp_s__5;
static PyObject *__pyx_kp_s__6;
static PyObject *__pyx_kp_s__8;
static PyObject *__pyx_n_s_abs;
static PyObject *__pyx_n_s_add_space_dimensions_and_embed;
static PyObject *__pyx_n_s_affine_dimension;
static PyObject *__pyx_n_s_affine_image;
static PyObject *__pyx_kp_s_alloc_lu;
static PyObject *__pyx_n_s_alphabet;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_autosim;
static PyObject *__pyx_n_s_b;
static PyObject *__pyx_kp_s_bot;
static PyObject *__pyx_kp_s_branch;
static PyObject *__pyx_n_s_children;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_kp_s_column_index_out_of_range;
static PyObject *__pyx_n_s_cone;
static PyObject *__pyx_n_s_constructors;
static PyObject *__pyx_kp_s_d_saddles_found_on_d_random_poi;
static PyObject *__pyx_kp_s_dim_lu;
static PyObject *__pyx_kp_s_dimension_mismatch;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_kp_s_entries_lu;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_kp_s_failed_to_allocate_s_bytes;
static PyObject *__pyx_kp_s_failed_to_allocate_s_s_bytes;
static PyObject *__pyx_n_s_file;
static PyObject *__pyx_n_s_flush;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_genexpr;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_has_connection;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_iet_interface;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_intersection_assign;
static PyObject *__pyx_n_s_is_zero;
static PyObject *__pyx_n_s_iterations;
static PyObject *__pyx_n_s_join;
static PyObject *__pyx_n_s_labels;
static PyObject *__pyx_n_s_left;
static PyObject *__pyx_kp_s_length_lu;
static PyObject *__pyx_kp_s_looking_children;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_max_depth;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_nray;
static PyObject *__pyx_n_s_num_exp;
static PyObject *__pyx_n_s_num_iterations;
static PyObject *__pyx_n_s_number_of_cylinders;
static PyObject *__pyx_kp_s_only_equality_and_difference_tes;
static PyObject *__pyx_n_s_p;
static PyObject *__pyx_kp_s_p_must_be_a_labelled_permutation;
static PyObject *__pyx_n_s_permutation;
static PyObject *__pyx_n_s_pop;
static PyObject *__pyx_n_s_ppl_check_non_negative_cone;
static PyObject *__pyx_n_s_print;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_random_element;
static PyObject *__pyx_n_s_random_element_locals_genexpr;
static PyObject *__pyx_n_s_random_integer_point_statistics;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_ray_coefficient;
static PyObject *__pyx_n_s_rays;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_repr___locals_genexpr;
static PyObject *__pyx_n_s_right;
static PyObject *__pyx_n_s_ring;
static PyObject *__pyx_kp_s_row_index_out_of_range;
static PyObject *__pyx_kp_s_rows_d_lu;
static PyObject *__pyx_kp_s_rows_lu;
static PyObject *__pyx_kp_s_s_f;
static PyObject *__pyx_n_s_saddle;
static PyObject *__pyx_n_s_sage_modules_free_module;
static PyObject *__pyx_n_s_sage_rings_integer_ring;
static PyObject *__pyx_n_s_sage_rings_rational_field;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_space_dimension;
static PyObject *__pyx_n_s_stdout;
static PyObject *__pyx_n_s_sum;
static PyObject *__pyx_n_s_surface_dynamics_interval_exchan;
static PyObject *__pyx_n_s_surface_dynamics_interval_exchan_2;
static PyObject *__pyx_n_s_surface_dynamics_misc_ppl_utils;
static PyObject *__pyx_n_s_sys;
static PyObject *__pyx_n_s_t;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_kp_s_top;
static PyObject *__pyx_n_s_tree;
static PyObject *__pyx_n_s_verbose;
static PyObject *__pyx_n_s_write;
static PyObject *__pyx_n_s_zorich_move;
static int __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily___cinit__(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self); /* proto */
static void __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_2__dealloc__(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self); /* proto */
static int __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_4__init__(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self, PyObject *__pyx_v_p, struct __pyx_obj_3ppl_10polyhedron_C_Polyhedron *__pyx_v_C); /* proto */
static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_6ray_coefficient(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self, size_t __pyx_v_nray, size_t __pyx_v_i); /* proto */
static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_8rays(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_8__repr___genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_8__repr___3genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_8__repr___6genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_10__repr__(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_12_debug_info(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_14permutation(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self); /* proto */
static Py_hash_t __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_16__hash__(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_18__richcmp__(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self, struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_other, int __pyx_v_op); /* proto */
static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_20has_connection(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_22children(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_24tree(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self, PyObject *__pyx_v_max_depth, PyObject *__pyx_v_verbose); /* proto */
static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_14random_element_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_27random_element(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self, PyObject *__pyx_v_ring, PyObject *__pyx_v_args, PyObject *__pyx_v_kwds); /* proto */
static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_29random_element_statistics(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self, PyObject *__pyx_v_ring, PyObject *__pyx_v_num_exp, PyObject *__pyx_v_num_iterations, PyObject *__pyx_v_args, PyObject *__pyx_v_kwds); /* proto */
static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_31random_integer_point_statistics_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_31random_integer_point_statistics(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self, PyObject *__pyx_v_num_exp); /* proto */
static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_33__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_35__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_4sage_9structure_7element_parent(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_4sage_9structure_7element_2have_same_parent(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_left, PyObject *__pyx_v_right); /* proto */
static PyObject *__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyList_Type_pop = {0, &__pyx_n_s_pop, 0, 0, 0};
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_5;
static PyObject *__pyx_int_100;
static PyObject *__pyx_int_200;
static PyObject *__pyx_int_100000000;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
/* Late includes */
#include "macros.h"

/* "surface_dynamics/interval_exchanges/iet_family.pyx":70
 * ########################
 * 
 * cdef int compare(const void * _left, const void * _right) nogil:             # <<<<<<<<<<<<<<
 *     cdef mpz_t * left = (<mpz_t **> _left)[0]
 *     cdef mpz_t * right = (<mpz_t **> _right)[0]
 */

static int __pyx_f_16surface_dynamics_18interval_exchanges_10iet_family_compare(void const *__pyx_v__left, void const *__pyx_v__right) {
  mpz_t *__pyx_v_left;
  mpz_t *__pyx_v_right;
  int __pyx_v_c;
  int __pyx_r;
  int __pyx_t_1;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":71
 * 
 * cdef int compare(const void * _left, const void * _right) nogil:
 *     cdef mpz_t * left = (<mpz_t **> _left)[0]             # <<<<<<<<<<<<<<
 *     cdef mpz_t * right = (<mpz_t **> _right)[0]
 *     cdef int c = mpz_cmp(left[0], right[0])
 */
  __pyx_v_left = (((mpz_t **)__pyx_v__left)[0]);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":72
 * cdef int compare(const void * _left, const void * _right) nogil:
 *     cdef mpz_t * left = (<mpz_t **> _left)[0]
 *     cdef mpz_t * right = (<mpz_t **> _right)[0]             # <<<<<<<<<<<<<<
 *     cdef int c = mpz_cmp(left[0], right[0])
 *     while c == 0:
 */
  __pyx_v_right = (((mpz_t **)__pyx_v__right)[0]);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":73
 *     cdef mpz_t * left = (<mpz_t **> _left)[0]
 *     cdef mpz_t * right = (<mpz_t **> _right)[0]
 *     cdef int c = mpz_cmp(left[0], right[0])             # <<<<<<<<<<<<<<
 *     while c == 0:
 *         left += 1
 */
  __pyx_v_c = mpz_cmp((__pyx_v_left[0]), (__pyx_v_right[0]));

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":74
 *     cdef mpz_t * right = (<mpz_t **> _right)[0]
 *     cdef int c = mpz_cmp(left[0], right[0])
 *     while c == 0:             # <<<<<<<<<<<<<<
 *         left += 1
 *         right += 1
 */
  while (1) {
    __pyx_t_1 = ((__pyx_v_c == 0) != 0);
    if (!__pyx_t_1) break;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":75
 *     cdef int c = mpz_cmp(left[0], right[0])
 *     while c == 0:
 *         left += 1             # <<<<<<<<<<<<<<
 *         right += 1
 *         c = mpz_cmp(left[0], right[0])
 */
    __pyx_v_left = (__pyx_v_left + 1);

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":76
 *     while c == 0:
 *         left += 1
 *         right += 1             # <<<<<<<<<<<<<<
 *         c = mpz_cmp(left[0], right[0])
 *     return c
 */
    __pyx_v_right = (__pyx_v_right + 1);

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":77
 *         left += 1
 *         right += 1
 *         c = mpz_cmp(left[0], right[0])             # <<<<<<<<<<<<<<
 *     return c
 * 
 */
    __pyx_v_c = mpz_cmp((__pyx_v_left[0]), (__pyx_v_right[0]));
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":78
 *         right += 1
 *         c = mpz_cmp(left[0], right[0])
 *     return c             # <<<<<<<<<<<<<<
 * 
 * cdef inline int mpz_vec_any_smaller(mpz_t * x, mpz_t * y, size_t length):
 */
  __pyx_r = __pyx_v_c;
  goto __pyx_L0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":70
 * ########################
 * 
 * cdef int compare(const void * _left, const void * _right) nogil:             # <<<<<<<<<<<<<<
 *     cdef mpz_t * left = (<mpz_t **> _left)[0]
 *     cdef mpz_t * right = (<mpz_t **> _right)[0]
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "surface_dynamics/interval_exchanges/iet_family.pyx":80
 *     return c
 * 
 * cdef inline int mpz_vec_any_smaller(mpz_t * x, mpz_t * y, size_t length):             # <<<<<<<<<<<<<<
 *     r"""
 *     Check whether any x[i] < y[i]
 */

static CYTHON_INLINE int __pyx_f_16surface_dynamics_18interval_exchanges_10iet_family_mpz_vec_any_smaller(mpz_t *__pyx_v_x, mpz_t *__pyx_v_y, size_t __pyx_v_length) {
  size_t __pyx_v_k;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("mpz_vec_any_smaller", 0);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":85
 *     """
 *     cdef size_t k
 *     for k in range(length):             # <<<<<<<<<<<<<<
 *         if mpz_cmp(x[k], y[k]) < 0:
 *             return 1
 */
  __pyx_t_1 = __pyx_v_length;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_k = __pyx_t_3;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":86
 *     cdef size_t k
 *     for k in range(length):
 *         if mpz_cmp(x[k], y[k]) < 0:             # <<<<<<<<<<<<<<
 *             return 1
 *     return 0
 */
    __pyx_t_4 = ((mpz_cmp((__pyx_v_x[__pyx_v_k]), (__pyx_v_y[__pyx_v_k])) < 0) != 0);
    if (__pyx_t_4) {

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":87
 *     for k in range(length):
 *         if mpz_cmp(x[k], y[k]) < 0:
 *             return 1             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
      __pyx_r = 1;
      goto __pyx_L0;

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":86
 *     cdef size_t k
 *     for k in range(length):
 *         if mpz_cmp(x[k], y[k]) < 0:             # <<<<<<<<<<<<<<
 *             return 1
 *     return 0
 */
    }
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":88
 *         if mpz_cmp(x[k], y[k]) < 0:
 *             return 1
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * cdef inline int mpz_vec_equal(mpz_t * x, mpz_t * y, size_t length):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":80
 *     return c
 * 
 * cdef inline int mpz_vec_any_smaller(mpz_t * x, mpz_t * y, size_t length):             # <<<<<<<<<<<<<<
 *     r"""
 *     Check whether any x[i] < y[i]
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "surface_dynamics/interval_exchanges/iet_family.pyx":90
 *     return 0
 * 
 * cdef inline int mpz_vec_equal(mpz_t * x, mpz_t * y, size_t length):             # <<<<<<<<<<<<<<
 *     cdef size_t k
 *     for k in range(length):
 */

static CYTHON_INLINE int __pyx_f_16surface_dynamics_18interval_exchanges_10iet_family_mpz_vec_equal(mpz_t *__pyx_v_x, mpz_t *__pyx_v_y, size_t __pyx_v_length) {
  size_t __pyx_v_k;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("mpz_vec_equal", 0);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":92
 * cdef inline int mpz_vec_equal(mpz_t * x, mpz_t * y, size_t length):
 *     cdef size_t k
 *     for k in range(length):             # <<<<<<<<<<<<<<
 *         if mpz_cmp(x[k], y[k]): return 0
 *     return 1
 */
  __pyx_t_1 = __pyx_v_length;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_k = __pyx_t_3;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":93
 *     cdef size_t k
 *     for k in range(length):
 *         if mpz_cmp(x[k], y[k]): return 0             # <<<<<<<<<<<<<<
 *     return 1
 * 
 */
    __pyx_t_4 = (mpz_cmp((__pyx_v_x[__pyx_v_k]), (__pyx_v_y[__pyx_v_k])) != 0);
    if (__pyx_t_4) {
      __pyx_r = 0;
      goto __pyx_L0;
    }
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":94
 *     for k in range(length):
 *         if mpz_cmp(x[k], y[k]): return 0
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * cdef Py_hash_t mpz_pythonhash(mpz_srcptr z):
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":90
 *     return 0
 * 
 * cdef inline int mpz_vec_equal(mpz_t * x, mpz_t * y, size_t length):             # <<<<<<<<<<<<<<
 *     cdef size_t k
 *     for k in range(length):
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "surface_dynamics/interval_exchanges/iet_family.pyx":96
 *     return 1
 * 
 * cdef Py_hash_t mpz_pythonhash(mpz_srcptr z):             # <<<<<<<<<<<<<<
 *     """
 *     Hash an ``mpz``, where the hash value is the same as the hash value
 */

static Py_hash_t __pyx_f_16surface_dynamics_18interval_exchanges_10iet_family_mpz_pythonhash(mpz_srcptr __pyx_v_z) {
  mp_limb_t __pyx_v_h1;
  mp_limb_t __pyx_v_h0;
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  Py_hash_t __pyx_v_h;
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("mpz_pythonhash", 0);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":103
 *     """
 *     # Add all limbs, adding 1 for every carry
 *     cdef mp_limb_t h1 = 0             # <<<<<<<<<<<<<<
 *     cdef mp_limb_t h0
 *     cdef size_t i, n
 */
  __pyx_v_h1 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":106
 *     cdef mp_limb_t h0
 *     cdef size_t i, n
 *     n = mpz_size(z)             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         h0 = h1
 */
  __pyx_v_n = mpz_size(__pyx_v_z);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":107
 *     cdef size_t i, n
 *     n = mpz_size(z)
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         h0 = h1
 *         h1 += mpz_getlimbn(z, i)
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":108
 *     n = mpz_size(z)
 *     for i in range(n):
 *         h0 = h1             # <<<<<<<<<<<<<<
 *         h1 += mpz_getlimbn(z, i)
 *         # Add 1 on overflow
 */
    __pyx_v_h0 = __pyx_v_h1;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":109
 *     for i in range(n):
 *         h0 = h1
 *         h1 += mpz_getlimbn(z, i)             # <<<<<<<<<<<<<<
 *         # Add 1 on overflow
 *         if h1 < h0: h1 += 1
 */
    __pyx_v_h1 = (__pyx_v_h1 + mpz_getlimbn(__pyx_v_z, __pyx_v_i));

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":111
 *         h1 += mpz_getlimbn(z, i)
 *         # Add 1 on overflow
 *         if h1 < h0: h1 += 1             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_hash_t h = h1
 */
    __pyx_t_4 = ((__pyx_v_h1 < __pyx_v_h0) != 0);
    if (__pyx_t_4) {
      __pyx_v_h1 = (__pyx_v_h1 + 1);
    }
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":113
 *         if h1 < h0: h1 += 1
 * 
 *     cdef Py_hash_t h = h1             # <<<<<<<<<<<<<<
 *     if mpz_sgn(z) < 0:
 *         return -h
 */
  __pyx_v_h = __pyx_v_h1;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":114
 * 
 *     cdef Py_hash_t h = h1
 *     if mpz_sgn(z) < 0:             # <<<<<<<<<<<<<<
 *         return -h
 *     return h
 */
  __pyx_t_4 = ((mpz_sgn(__pyx_v_z) < 0) != 0);
  if (__pyx_t_4) {

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":115
 *     cdef Py_hash_t h = h1
 *     if mpz_sgn(z) < 0:
 *         return -h             # <<<<<<<<<<<<<<
 *     return h
 * 
 */
    __pyx_r = (-__pyx_v_h);
    goto __pyx_L0;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":114
 * 
 *     cdef Py_hash_t h = h1
 *     if mpz_sgn(z) < 0:             # <<<<<<<<<<<<<<
 *         return -h
 *     return h
 */
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":116
 *     if mpz_sgn(z) < 0:
 *         return -h
 *     return h             # <<<<<<<<<<<<<<
 * 
 * cdef inline void rauzy_top(int * perm, size_t dim):
 */
  __pyx_r = __pyx_v_h;
  goto __pyx_L0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":96
 *     return 1
 * 
 * cdef Py_hash_t mpz_pythonhash(mpz_srcptr z):             # <<<<<<<<<<<<<<
 *     """
 *     Hash an ``mpz``, where the hash value is the same as the hash value
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "surface_dynamics/interval_exchanges/iet_family.pyx":118
 *     return h
 * 
 * cdef inline void rauzy_top(int * perm, size_t dim):             # <<<<<<<<<<<<<<
 *     cdef int a = perm[dim - 1]
 *     cdef size_t i, j
 */

static CYTHON_INLINE void __pyx_f_16surface_dynamics_18interval_exchanges_10iet_family_rauzy_top(int *__pyx_v_perm, size_t __pyx_v_dim) {
  int __pyx_v_a;
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("rauzy_top", 0);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":119
 * 
 * cdef inline void rauzy_top(int * perm, size_t dim):
 *     cdef int a = perm[dim - 1]             # <<<<<<<<<<<<<<
 *     cdef size_t i, j
 *     for i in range(dim, 2*dim):
 */
  __pyx_v_a = (__pyx_v_perm[(__pyx_v_dim - 1)]);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":121
 *     cdef int a = perm[dim - 1]
 *     cdef size_t i, j
 *     for i in range(dim, 2*dim):             # <<<<<<<<<<<<<<
 *         if perm[i] == a:
 *             break
 */
  __pyx_t_1 = (2 * __pyx_v_dim);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = __pyx_v_dim; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":122
 *     cdef size_t i, j
 *     for i in range(dim, 2*dim):
 *         if perm[i] == a:             # <<<<<<<<<<<<<<
 *             break
 *     a = perm[2 * dim - 1]
 */
    __pyx_t_4 = (((__pyx_v_perm[__pyx_v_i]) == __pyx_v_a) != 0);
    if (__pyx_t_4) {

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":123
 *     for i in range(dim, 2*dim):
 *         if perm[i] == a:
 *             break             # <<<<<<<<<<<<<<
 *     a = perm[2 * dim - 1]
 *     for j in range(2 * dim - 1, i + 1, -1):
 */
      goto __pyx_L4_break;

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":122
 *     cdef size_t i, j
 *     for i in range(dim, 2*dim):
 *         if perm[i] == a:             # <<<<<<<<<<<<<<
 *             break
 *     a = perm[2 * dim - 1]
 */
    }
  }
  __pyx_L4_break:;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":124
 *         if perm[i] == a:
 *             break
 *     a = perm[2 * dim - 1]             # <<<<<<<<<<<<<<
 *     for j in range(2 * dim - 1, i + 1, -1):
 *         perm[j] = perm[j - 1]
 */
  __pyx_v_a = (__pyx_v_perm[((2 * __pyx_v_dim) - 1)]);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":125
 *             break
 *     a = perm[2 * dim - 1]
 *     for j in range(2 * dim - 1, i + 1, -1):             # <<<<<<<<<<<<<<
 *         perm[j] = perm[j - 1]
 *     perm[i + 1] = a
 */
  __pyx_t_1 = (__pyx_v_i + 1);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = ((2 * __pyx_v_dim) - 1) + 1; __pyx_t_3 > __pyx_t_2 + 1; ) { __pyx_t_3-=1;
    __pyx_v_j = __pyx_t_3;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":126
 *     a = perm[2 * dim - 1]
 *     for j in range(2 * dim - 1, i + 1, -1):
 *         perm[j] = perm[j - 1]             # <<<<<<<<<<<<<<
 *     perm[i + 1] = a
 * 
 */
    (__pyx_v_perm[__pyx_v_j]) = (__pyx_v_perm[(__pyx_v_j - 1)]);
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":127
 *     for j in range(2 * dim - 1, i + 1, -1):
 *         perm[j] = perm[j - 1]
 *     perm[i + 1] = a             # <<<<<<<<<<<<<<
 * 
 * cdef inline void rauzy_bot(int * perm, size_t dim):
 */
  (__pyx_v_perm[(__pyx_v_i + 1)]) = __pyx_v_a;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":118
 *     return h
 * 
 * cdef inline void rauzy_top(int * perm, size_t dim):             # <<<<<<<<<<<<<<
 *     cdef int a = perm[dim - 1]
 *     cdef size_t i, j
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "surface_dynamics/interval_exchanges/iet_family.pyx":129
 *     perm[i + 1] = a
 * 
 * cdef inline void rauzy_bot(int * perm, size_t dim):             # <<<<<<<<<<<<<<
 *     cdef int a = perm[2*dim - 1]
 *     cdef size_t i, j
 */

static CYTHON_INLINE void __pyx_f_16surface_dynamics_18interval_exchanges_10iet_family_rauzy_bot(int *__pyx_v_perm, size_t __pyx_v_dim) {
  int __pyx_v_a;
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("rauzy_bot", 0);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":130
 * 
 * cdef inline void rauzy_bot(int * perm, size_t dim):
 *     cdef int a = perm[2*dim - 1]             # <<<<<<<<<<<<<<
 *     cdef size_t i, j
 *     for i in range(dim):
 */
  __pyx_v_a = (__pyx_v_perm[((2 * __pyx_v_dim) - 1)]);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":132
 *     cdef int a = perm[2*dim - 1]
 *     cdef size_t i, j
 *     for i in range(dim):             # <<<<<<<<<<<<<<
 *         if perm[i] == a:
 *             break
 */
  __pyx_t_1 = __pyx_v_dim;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":133
 *     cdef size_t i, j
 *     for i in range(dim):
 *         if perm[i] == a:             # <<<<<<<<<<<<<<
 *             break
 *     a = perm[dim - 1]
 */
    __pyx_t_4 = (((__pyx_v_perm[__pyx_v_i]) == __pyx_v_a) != 0);
    if (__pyx_t_4) {

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":134
 *     for i in range(dim):
 *         if perm[i] == a:
 *             break             # <<<<<<<<<<<<<<
 *     a = perm[dim - 1]
 *     for j in range(dim - 1, i + 1, -1):
 */
      goto __pyx_L4_break;

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":133
 *     cdef size_t i, j
 *     for i in range(dim):
 *         if perm[i] == a:             # <<<<<<<<<<<<<<
 *             break
 *     a = perm[dim - 1]
 */
    }
  }
  __pyx_L4_break:;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":135
 *         if perm[i] == a:
 *             break
 *     a = perm[dim - 1]             # <<<<<<<<<<<<<<
 *     for j in range(dim - 1, i + 1, -1):
 *         perm[j] = perm[j - 1]
 */
  __pyx_v_a = (__pyx_v_perm[(__pyx_v_dim - 1)]);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":136
 *             break
 *     a = perm[dim - 1]
 *     for j in range(dim - 1, i + 1, -1):             # <<<<<<<<<<<<<<
 *         perm[j] = perm[j - 1]
 *     perm[i + 1] = a
 */
  __pyx_t_1 = (__pyx_v_i + 1);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = (__pyx_v_dim - 1) + 1; __pyx_t_3 > __pyx_t_2 + 1; ) { __pyx_t_3-=1;
    __pyx_v_j = __pyx_t_3;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":137
 *     a = perm[dim - 1]
 *     for j in range(dim - 1, i + 1, -1):
 *         perm[j] = perm[j - 1]             # <<<<<<<<<<<<<<
 *     perm[i + 1] = a
 * 
 */
    (__pyx_v_perm[__pyx_v_j]) = (__pyx_v_perm[(__pyx_v_j - 1)]);
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":138
 *     for j in range(dim - 1, i + 1, -1):
 *         perm[j] = perm[j - 1]
 *     perm[i + 1] = a             # <<<<<<<<<<<<<<
 * 
 * cdef class IETFamily(object):
 */
  (__pyx_v_perm[(__pyx_v_i + 1)]) = __pyx_v_a;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":129
 *     perm[i + 1] = a
 * 
 * cdef inline void rauzy_bot(int * perm, size_t dim):             # <<<<<<<<<<<<<<
 *     cdef int a = perm[2*dim - 1]
 *     cdef size_t i, j
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "surface_dynamics/interval_exchanges/iet_family.pyx":187
 *     cdef C_Polyhedron poly  # the polyhedron (that we should get rid of)
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.perm = NULL
 *         self.entries = NULL
 */

/* Python wrapper */
static int __pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily___cinit__(((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily___cinit__(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":188
 * 
 *     def __cinit__(self):
 *         self.perm = NULL             # <<<<<<<<<<<<<<
 *         self.entries = NULL
 *         self.rows = NULL
 */
  __pyx_v_self->perm = NULL;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":189
 *     def __cinit__(self):
 *         self.perm = NULL
 *         self.entries = NULL             # <<<<<<<<<<<<<<
 *         self.rows = NULL
 *         self.alloc = 0
 */
  __pyx_v_self->entries = NULL;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":190
 *         self.perm = NULL
 *         self.entries = NULL
 *         self.rows = NULL             # <<<<<<<<<<<<<<
 *         self.alloc = 0
 *         self.dim = 0
 */
  __pyx_v_self->rows = NULL;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":191
 *         self.entries = NULL
 *         self.rows = NULL
 *         self.alloc = 0             # <<<<<<<<<<<<<<
 *         self.dim = 0
 *         self.length = 0
 */
  __pyx_v_self->alloc = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":192
 *         self.rows = NULL
 *         self.alloc = 0
 *         self.dim = 0             # <<<<<<<<<<<<<<
 *         self.length = 0
 * 
 */
  __pyx_v_self->dim = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":193
 *         self.alloc = 0
 *         self.dim = 0
 *         self.length = 0             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_v_self->length = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":187
 *     cdef C_Polyhedron poly  # the polyhedron (that we should get rid of)
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.perm = NULL
 *         self.entries = NULL
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "surface_dynamics/interval_exchanges/iet_family.pyx":195
 *         self.length = 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         for i in range(self.dim * self.alloc):
 */

/* Python wrapper */
static void __pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_2__dealloc__(((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_2__dealloc__(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self) {
  size_t __pyx_v_i;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":197
 *     def __dealloc__(self):
 *         cdef size_t i
 *         for i in range(self.dim * self.alloc):             # <<<<<<<<<<<<<<
 *             mpz_clear(self.entries[i])
 *         free(self.perm)
 */
  __pyx_t_1 = (__pyx_v_self->dim * __pyx_v_self->alloc);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":198
 *         cdef size_t i
 *         for i in range(self.dim * self.alloc):
 *             mpz_clear(self.entries[i])             # <<<<<<<<<<<<<<
 *         free(self.perm)
 *         free(self.entries)
 */
    mpz_clear((__pyx_v_self->entries[__pyx_v_i]));
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":199
 *         for i in range(self.dim * self.alloc):
 *             mpz_clear(self.entries[i])
 *         free(self.perm)             # <<<<<<<<<<<<<<
 *         free(self.entries)
 *         free(self.rows)
 */
  free(__pyx_v_self->perm);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":200
 *             mpz_clear(self.entries[i])
 *         free(self.perm)
 *         free(self.entries)             # <<<<<<<<<<<<<<
 *         free(self.rows)
 * 
 */
  free(__pyx_v_self->entries);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":201
 *         free(self.perm)
 *         free(self.entries)
 *         free(self.rows)             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, p, C_Polyhedron C):
 */
  free(__pyx_v_self->rows);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":195
 *         self.length = 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         for i in range(self.dim * self.alloc):
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "surface_dynamics/interval_exchanges/iet_family.pyx":203
 *         free(self.rows)
 * 
 *     def __init__(self, p, C_Polyhedron C):             # <<<<<<<<<<<<<<
 *         # convert and check input
 *         from surface_dynamics.interval_exchanges.labelled import LabelledPermutationIET
 */

/* Python wrapper */
static int __pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_5__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_5__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_p = 0;
  struct __pyx_obj_3ppl_10polyhedron_C_Polyhedron *__pyx_v_C = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_p,&__pyx_n_s_C,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_p)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 203, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 203, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_p = values[0];
    __pyx_v_C = ((struct __pyx_obj_3ppl_10polyhedron_C_Polyhedron *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 203, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_C), __pyx_ptype_3ppl_10polyhedron_C_Polyhedron, 1, "C", 0))) __PYX_ERR(0, 203, __pyx_L1_error)
  __pyx_r = __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_4__init__(((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)__pyx_v_self), __pyx_v_p, __pyx_v_C);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_4__init__(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self, PyObject *__pyx_v_p, struct __pyx_obj_3ppl_10polyhedron_C_Polyhedron *__pyx_v_C) {
  PyObject *__pyx_v_LabelledPermutationIET = NULL;
  PyObject *__pyx_v_ppl_check_non_negative_cone = NULL;
  size_t __pyx_v_i;
  int __pyx_v_j;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  void *__pyx_t_7;
  size_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":205
 *     def __init__(self, p, C_Polyhedron C):
 *         # convert and check input
 *         from surface_dynamics.interval_exchanges.labelled import LabelledPermutationIET             # <<<<<<<<<<<<<<
 *         if not isinstance(p, LabelledPermutationIET):
 *             raise ValueError('p must be a labelled permutation')
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_LabelledPermutationIET);
  __Pyx_GIVEREF(__pyx_n_s_LabelledPermutationIET);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_LabelledPermutationIET);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_surface_dynamics_interval_exchan, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_LabelledPermutationIET); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_LabelledPermutationIET = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":206
 *         # convert and check input
 *         from surface_dynamics.interval_exchanges.labelled import LabelledPermutationIET
 *         if not isinstance(p, LabelledPermutationIET):             # <<<<<<<<<<<<<<
 *             raise ValueError('p must be a labelled permutation')
 * 
 */
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_p, __pyx_v_LabelledPermutationIET); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 206, __pyx_L1_error)
  __pyx_t_4 = ((!(__pyx_t_3 != 0)) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":207
 *         from surface_dynamics.interval_exchanges.labelled import LabelledPermutationIET
 *         if not isinstance(p, LabelledPermutationIET):
 *             raise ValueError('p must be a labelled permutation')             # <<<<<<<<<<<<<<
 * 
 *         from surface_dynamics.misc.ppl_utils import ppl_check_non_negative_cone
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 207, __pyx_L1_error)

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":206
 *         # convert and check input
 *         from surface_dynamics.interval_exchanges.labelled import LabelledPermutationIET
 *         if not isinstance(p, LabelledPermutationIET):             # <<<<<<<<<<<<<<
 *             raise ValueError('p must be a labelled permutation')
 * 
 */
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":209
 *             raise ValueError('p must be a labelled permutation')
 * 
 *         from surface_dynamics.misc.ppl_utils import ppl_check_non_negative_cone             # <<<<<<<<<<<<<<
 *         ppl_check_non_negative_cone(C)
 *         if C.space_dimension() != len(p):
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_ppl_check_non_negative_cone);
  __Pyx_GIVEREF(__pyx_n_s_ppl_check_non_negative_cone);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_ppl_check_non_negative_cone);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_surface_dynamics_misc_ppl_utils, __pyx_t_2, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_ppl_check_non_negative_cone); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_ppl_check_non_negative_cone = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":210
 * 
 *         from surface_dynamics.misc.ppl_utils import ppl_check_non_negative_cone
 *         ppl_check_non_negative_cone(C)             # <<<<<<<<<<<<<<
 *         if C.space_dimension() != len(p):
 *             raise ValueError('dimension mismatch')
 */
  __Pyx_INCREF(__pyx_v_ppl_check_non_negative_cone);
  __pyx_t_2 = __pyx_v_ppl_check_non_negative_cone; __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, ((PyObject *)__pyx_v_C)) : __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)__pyx_v_C));
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":211
 *         from surface_dynamics.misc.ppl_utils import ppl_check_non_negative_cone
 *         ppl_check_non_negative_cone(C)
 *         if C.space_dimension() != len(p):             # <<<<<<<<<<<<<<
 *             raise ValueError('dimension mismatch')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_C), __pyx_n_s_space_dimension); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = PyObject_Length(__pyx_v_p); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 211, __pyx_L1_error)
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(__pyx_t_4)) {

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":212
 *         ppl_check_non_negative_cone(C)
 *         if C.space_dimension() != len(p):
 *             raise ValueError('dimension mismatch')             # <<<<<<<<<<<<<<
 * 
 *         # fill C data from input
 */
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 212, __pyx_L1_error)

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":211
 *         from surface_dynamics.misc.ppl_utils import ppl_check_non_negative_cone
 *         ppl_check_non_negative_cone(C)
 *         if C.space_dimension() != len(p):             # <<<<<<<<<<<<<<
 *             raise ValueError('dimension mismatch')
 * 
 */
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":215
 * 
 *         # fill C data from input
 *         self.dim = len(p)             # <<<<<<<<<<<<<<
 *         self.perm = <int *> check_malloc(2 * self.dim * sizeof(int))
 *         cdef size_t i
 */
  __pyx_t_6 = PyObject_Length(__pyx_v_p); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 215, __pyx_L1_error)
  __pyx_v_self->dim = __pyx_t_6;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":216
 *         # fill C data from input
 *         self.dim = len(p)
 *         self.perm = <int *> check_malloc(2 * self.dim * sizeof(int))             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         cdef int j
 */
  __pyx_t_7 = __pyx_f_9cysignals_6memory_check_malloc(((2 * __pyx_v_self->dim) * (sizeof(int)))); if (unlikely(__pyx_t_7 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 216, __pyx_L1_error)
  __pyx_v_self->perm = ((int *)__pyx_t_7);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":219
 *         cdef size_t i
 *         cdef int j
 *         for i,j in enumerate(p._labels[0]):             # <<<<<<<<<<<<<<
 *             self.perm[i] = j
 *         for i,j in enumerate(p._labels[1]):
 */
  __pyx_t_8 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_p, __pyx_n_s_labels); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_5, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_5 = __pyx_t_2; __Pyx_INCREF(__pyx_t_5); __pyx_t_6 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 219, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 219, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 219, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 219, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 219, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 219, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_9(__pyx_t_5);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 219, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 219, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_j = __pyx_t_10;
    __pyx_v_i = __pyx_t_8;
    __pyx_t_8 = (__pyx_t_8 + 1);

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":220
 *         cdef int j
 *         for i,j in enumerate(p._labels[0]):
 *             self.perm[i] = j             # <<<<<<<<<<<<<<
 *         for i,j in enumerate(p._labels[1]):
 *             self.perm[self.dim + i] = j
 */
    (__pyx_v_self->perm[__pyx_v_i]) = __pyx_v_j;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":219
 *         cdef size_t i
 *         cdef int j
 *         for i,j in enumerate(p._labels[0]):             # <<<<<<<<<<<<<<
 *             self.perm[i] = j
 *         for i,j in enumerate(p._labels[1]):
 */
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":221
 *         for i,j in enumerate(p._labels[0]):
 *             self.perm[i] = j
 *         for i,j in enumerate(p._labels[1]):             # <<<<<<<<<<<<<<
 *             self.perm[self.dim + i] = j
 * 
 */
  __pyx_t_8 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_p, __pyx_n_s_labels); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_5, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_5 = __pyx_t_2; __Pyx_INCREF(__pyx_t_5); __pyx_t_6 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 221, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 221, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 221, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 221, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 221, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 221, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_9(__pyx_t_5);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 221, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 221, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_j = __pyx_t_10;
    __pyx_v_i = __pyx_t_8;
    __pyx_t_8 = (__pyx_t_8 + 1);

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":222
 *             self.perm[i] = j
 *         for i,j in enumerate(p._labels[1]):
 *             self.perm[self.dim + i] = j             # <<<<<<<<<<<<<<
 * 
 *         self.poly = <C_Polyhedron> C
 */
    (__pyx_v_self->perm[(__pyx_v_self->dim + __pyx_v_i)]) = __pyx_v_j;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":221
 *         for i,j in enumerate(p._labels[0]):
 *             self.perm[i] = j
 *         for i,j in enumerate(p._labels[1]):             # <<<<<<<<<<<<<<
 *             self.perm[self.dim + i] = j
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":224
 *             self.perm[self.dim + i] = j
 * 
 *         self.poly = <C_Polyhedron> C             # <<<<<<<<<<<<<<
 *         self.fill_rays(self.poly.thisptr.minimized_generators())
 * 
 */
  __pyx_t_5 = ((PyObject *)__pyx_v_C);
  __Pyx_INCREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF(__pyx_v_self->poly);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->poly));
  __pyx_v_self->poly = ((struct __pyx_obj_3ppl_10polyhedron_C_Polyhedron *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":225
 * 
 *         self.poly = <C_Polyhedron> C
 *         self.fill_rays(self.poly.thisptr.minimized_generators())             # <<<<<<<<<<<<<<
 * 
 *     def ray_coefficient(self, size_t nray, size_t i):
 */
  ((struct __pyx_vtabstruct_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)__pyx_v_self->__pyx_vtab)->fill_rays(__pyx_v_self, __pyx_v_self->poly->__pyx_base.thisptr->minimized_generators());

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":203
 *         free(self.rows)
 * 
 *     def __init__(self, p, C_Polyhedron C):             # <<<<<<<<<<<<<<
 *         # convert and check input
 *         from surface_dynamics.interval_exchanges.labelled import LabelledPermutationIET
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_LabelledPermutationIET);
  __Pyx_XDECREF(__pyx_v_ppl_check_non_negative_cone);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "surface_dynamics/interval_exchanges/iet_family.pyx":227
 *         self.fill_rays(self.poly.thisptr.minimized_generators())
 * 
 *     def ray_coefficient(self, size_t nray, size_t i):             # <<<<<<<<<<<<<<
 *         r"""
 *         Return the ``i``-th coefficient of the ``nray``-th ray
 */

/* Python wrapper */
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_7ray_coefficient(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_6ray_coefficient[] = "\n        Return the ``i``-th coefficient of the ``nray``-th ray\n\n        EXAMPLES::\n\n            sage: from surface_dynamics import *\n            sage: p = iet.Permutation('a b c d', 'd c b a')\n            sage: F = iet.IETFamily(p, [(2,3,0,0), (0,1,1,1)]) # optional - pplpy\n            sage: F.ray_coefficient(0, 2) # optional - pplpy\n            1\n            sage: F.ray_coefficient(1, 0) # optional - pplpy\n            2\n            sage: F.ray_coefficient(3, 0) # optional - pplpy\n            Traceback (most recent call last):\n            ...\n            IndexError: row index out of range\n            sage: F.ray_coefficient(1, 5) # optional - pplpy\n            Traceback (most recent call last):\n            ...\n            IndexError: column index out of range\n        ";
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_7ray_coefficient(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_nray;
  size_t __pyx_v_i;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ray_coefficient (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_nray,&__pyx_n_s_i,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nray)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_i)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ray_coefficient", 1, 2, 2, 1); __PYX_ERR(0, 227, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ray_coefficient") < 0)) __PYX_ERR(0, 227, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_nray = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_nray == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 227, __pyx_L3_error)
    __pyx_v_i = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_i == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 227, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ray_coefficient", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 227, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.ray_coefficient", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_6ray_coefficient(((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)__pyx_v_self), __pyx_v_nray, __pyx_v_i);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_6ray_coefficient(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self, size_t __pyx_v_nray, size_t __pyx_v_i) {
  struct __pyx_obj_4sage_5rings_7integer_Integer *__pyx_v_z = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ray_coefficient", 0);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":249
 *             IndexError: column index out of range
 *         """
 *         if nray >= self.length:             # <<<<<<<<<<<<<<
 *             raise IndexError('row index out of range')
 *         if i >= self.dim:
 */
  __pyx_t_1 = ((__pyx_v_nray >= __pyx_v_self->length) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":250
 *         """
 *         if nray >= self.length:
 *             raise IndexError('row index out of range')             # <<<<<<<<<<<<<<
 *         if i >= self.dim:
 *             raise IndexError('column index out of range')
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 250, __pyx_L1_error)

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":249
 *             IndexError: column index out of range
 *         """
 *         if nray >= self.length:             # <<<<<<<<<<<<<<
 *             raise IndexError('row index out of range')
 *         if i >= self.dim:
 */
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":251
 *         if nray >= self.length:
 *             raise IndexError('row index out of range')
 *         if i >= self.dim:             # <<<<<<<<<<<<<<
 *             raise IndexError('column index out of range')
 *         cdef Integer z= PY_NEW(Integer)
 */
  __pyx_t_1 = ((__pyx_v_i >= __pyx_v_self->dim) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":252
 *             raise IndexError('row index out of range')
 *         if i >= self.dim:
 *             raise IndexError('column index out of range')             # <<<<<<<<<<<<<<
 *         cdef Integer z= PY_NEW(Integer)
 *         mpz_set(z.value, self.rows[nray][i])
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 252, __pyx_L1_error)

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":251
 *         if nray >= self.length:
 *             raise IndexError('row index out of range')
 *         if i >= self.dim:             # <<<<<<<<<<<<<<
 *             raise IndexError('column index out of range')
 *         cdef Integer z= PY_NEW(Integer)
 */
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":253
 *         if i >= self.dim:
 *             raise IndexError('column index out of range')
 *         cdef Integer z= PY_NEW(Integer)             # <<<<<<<<<<<<<<
 *         mpz_set(z.value, self.rows[nray][i])
 *         return z
 */
  __pyx_t_2 = __pyx_f_4sage_3ext_7stdsage_PY_NEW(__pyx_ptype_4sage_5rings_7integer_Integer); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_4sage_5rings_7integer_Integer))))) __PYX_ERR(0, 253, __pyx_L1_error)
  __pyx_v_z = ((struct __pyx_obj_4sage_5rings_7integer_Integer *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":254
 *             raise IndexError('column index out of range')
 *         cdef Integer z= PY_NEW(Integer)
 *         mpz_set(z.value, self.rows[nray][i])             # <<<<<<<<<<<<<<
 *         return z
 * 
 */
  mpz_set(__pyx_v_z->value, ((__pyx_v_self->rows[__pyx_v_nray])[__pyx_v_i]));

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":255
 *         cdef Integer z= PY_NEW(Integer)
 *         mpz_set(z.value, self.rows[nray][i])
 *         return z             # <<<<<<<<<<<<<<
 * 
 *     def rays(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_z));
  __pyx_r = ((PyObject *)__pyx_v_z);
  goto __pyx_L0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":227
 *         self.fill_rays(self.poly.thisptr.minimized_generators())
 * 
 *     def ray_coefficient(self, size_t nray, size_t i):             # <<<<<<<<<<<<<<
 *         r"""
 *         Return the ``i``-th coefficient of the ``nray``-th ray
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.ray_coefficient", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_z);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "surface_dynamics/interval_exchanges/iet_family.pyx":257
 *         return z
 * 
 *     def rays(self):             # <<<<<<<<<<<<<<
 *         r"""
 *         Return the rays as vectors
 */

/* Python wrapper */
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_9rays(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_8rays[] = "\n        Return the rays as vectors\n\n        EXAMPLES::\n\n            sage: from surface_dynamics import *\n            sage: p = iet.Permutation('a b c d', 'd c b a')\n            sage: F = iet.IETFamily(p, [(2,3,0,0), (0,1,1,1)]) # optional - pplpy\n            sage: F.rays() # optional - pplpy\n            [(0, 1, 1, 1), (2, 3, 0, 0)]\n        ";
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_9rays(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("rays (wrapper)", 0);
  __pyx_r = __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_8rays(((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_8rays(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self) {
  PyObject *__pyx_v_FreeModule = NULL;
  PyObject *__pyx_v_QQ = NULL;
  PyObject *__pyx_v_F = NULL;
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  size_t __pyx_t_10;
  size_t __pyx_t_11;
  size_t __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("rays", 0);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":269
 *             [(0, 1, 1, 1), (2, 3, 0, 0)]
 *         """
 *         from sage.modules.free_module import FreeModule             # <<<<<<<<<<<<<<
 *         from sage.rings.rational_field import QQ
 *         F = FreeModule(QQ, self.dim)
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_FreeModule);
  __Pyx_GIVEREF(__pyx_n_s_FreeModule);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_FreeModule);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_sage_modules_free_module, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_FreeModule); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_FreeModule = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":270
 *         """
 *         from sage.modules.free_module import FreeModule
 *         from sage.rings.rational_field import QQ             # <<<<<<<<<<<<<<
 *         F = FreeModule(QQ, self.dim)
 *         return [F([self.ray_coefficient(i, j) for j in range(self.dim)]) for i in range(self.length)]
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_QQ);
  __Pyx_GIVEREF(__pyx_n_s_QQ);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_QQ);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_sage_rings_rational_field, __pyx_t_2, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_QQ); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_QQ = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":271
 *         from sage.modules.free_module import FreeModule
 *         from sage.rings.rational_field import QQ
 *         F = FreeModule(QQ, self.dim)             # <<<<<<<<<<<<<<
 *         return [F([self.ray_coefficient(i, j) for j in range(self.dim)]) for i in range(self.length)]
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->dim); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_FreeModule);
  __pyx_t_3 = __pyx_v_FreeModule; __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_QQ, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 271, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_QQ, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 271, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_QQ);
    __Pyx_GIVEREF(__pyx_v_QQ);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_QQ);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_F = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":272
 *         from sage.rings.rational_field import QQ
 *         F = FreeModule(QQ, self.dim)
 *         return [F([self.ray_coefficient(i, j) for j in range(self.dim)]) for i in range(self.length)]             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __pyx_v_self->length;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;
    __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_10 = __pyx_v_self->dim;
    __pyx_t_11 = __pyx_t_10;
    for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
      __pyx_v_j = __pyx_t_12;
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_ray_coefficient); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 272, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_13 = __Pyx_PyInt_FromSize_t(__pyx_v_i); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 272, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_14 = __Pyx_PyInt_FromSize_t(__pyx_v_j); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 272, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_15 = NULL;
      __pyx_t_5 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_15)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_15);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_5 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[3] = {__pyx_t_15, __pyx_t_13, __pyx_t_14};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 272, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[3] = {__pyx_t_15, __pyx_t_13, __pyx_t_14};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 272, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      } else
      #endif
      {
        __pyx_t_16 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 272, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        if (__pyx_t_15) {
          __Pyx_GIVEREF(__pyx_t_15); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_15); __pyx_t_15 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_13);
        PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_5, __pyx_t_13);
        __Pyx_GIVEREF(__pyx_t_14);
        PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_5, __pyx_t_14);
        __pyx_t_13 = 0;
        __pyx_t_14 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_16, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 272, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_6, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 272, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_INCREF(__pyx_v_F);
    __pyx_t_2 = __pyx_v_F; __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 272, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":257
 *         return z
 * 
 *     def rays(self):             # <<<<<<<<<<<<<<
 *         r"""
 *         Return the rays as vectors
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.rays", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_FreeModule);
  __Pyx_XDECREF(__pyx_v_QQ);
  __Pyx_XDECREF(__pyx_v_F);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "surface_dynamics/interval_exchanges/iet_family.pyx":274
 *         return [F([self.ray_coefficient(i, j) for j in range(self.dim)]) for i in range(self.length)]
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         r"""
 *         TESTS::
 */

/* Python wrapper */
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_11__repr__(PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_10__repr__[] = "\n        TESTS::\n\n            sage: from surface_dynamics import *\n            sage: p = iet.Permutation('a b c d', 'd c b a')\n            sage: F = iet.IETFamily(p, [(2,3,0,0), (0,1,1,1)]) # optional - pplpy\n            sage: repr(F)  # indirect doctest # optional - pplpy\n            'top 0 1 2 3\\nbot 3 2 1 0\\n0 1 1 1\\n2 3 0 0'\n        ";
#if CYTHON_COMPILING_IN_CPYTHON
struct wrapperbase __pyx_wrapperbase_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_10__repr__;
#endif
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_11__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_10__repr__(((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_8__repr___2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "surface_dynamics/interval_exchanges/iet_family.pyx":286
 *         cdef size_t i, j
 *         s = []
 *         s.append('top ' + ' '.join(str(self.perm[i]) for i in range(self.dim)))             # <<<<<<<<<<<<<<
 *         s.append('bot ' + ' '.join(str(self.perm[self.dim+i]) for i in range(self.dim)))
 *         for i in range(self.length):
 */

static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_8__repr___genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr *)__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr(__pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 286, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_8__repr___2generator1, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_repr___locals_genexpr, __pyx_n_s_surface_dynamics_interval_exchan_2); if (unlikely(!gen)) __PYX_ERR(0, 286, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.__repr__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_8__repr___2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 286, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 286, __pyx_L1_error) }
  __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->dim;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_cur_scope->__pyx_v_i = __pyx_t_3;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 286, __pyx_L1_error) }
    __pyx_t_4 = __Pyx_PyInt_From_int((__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->perm[__pyx_cur_scope->__pyx_v_i])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 286, __pyx_L1_error)
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_8__repr___5generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "surface_dynamics/interval_exchanges/iet_family.pyx":287
 *         s = []
 *         s.append('top ' + ' '.join(str(self.perm[i]) for i in range(self.dim)))
 *         s.append('bot ' + ' '.join(str(self.perm[self.dim+i]) for i in range(self.dim)))             # <<<<<<<<<<<<<<
 *         for i in range(self.length):
 *             s.append(' '.join(str(self.ray_coefficient(i, j)) for j in range(self.dim)))
 */

static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_8__repr___3genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr *)__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr(__pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 287, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_8__repr___5generator2, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_repr___locals_genexpr, __pyx_n_s_surface_dynamics_interval_exchan_2); if (unlikely(!gen)) __PYX_ERR(0, 287, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.__repr__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_8__repr___5generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr *__pyx_cur_scope = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 287, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 287, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->dim); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 287, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 287, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 287, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 287, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 287, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_4(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 287, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_i);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_i, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 287, __pyx_L1_error) }
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 287, __pyx_L1_error) }
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->dim); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = PyNumber_Add(__pyx_t_2, __pyx_cur_scope->__pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyIndex_AsSsize_t(__pyx_t_5); if (unlikely((__pyx_t_6 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 287, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyInt_From_int((__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->perm[__pyx_t_6])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 287, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_8__repr___8generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "surface_dynamics/interval_exchanges/iet_family.pyx":289
 *         s.append('bot ' + ' '.join(str(self.perm[self.dim+i]) for i in range(self.dim)))
 *         for i in range(self.length):
 *             s.append(' '.join(str(self.ray_coefficient(i, j)) for j in range(self.dim)))             # <<<<<<<<<<<<<<
 *         return '\n'.join(s)
 * 
 */

static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_8__repr___6genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr *)__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr(__pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 289, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__ *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_8__repr___8generator3, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_repr___locals_genexpr, __pyx_n_s_surface_dynamics_interval_exchan_2); if (unlikely(!gen)) __PYX_ERR(0, 289, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.__repr__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_8__repr___8generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr *__pyx_cur_scope = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 289, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 289, __pyx_L1_error) }
  __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->dim;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_cur_scope->__pyx_v_j = __pyx_t_3;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 289, __pyx_L1_error) }
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self), __pyx_n_s_ray_coefficient); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_FromSize_t(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_i); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_FromSize_t(__pyx_cur_scope->__pyx_v_j); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 289, __pyx_L1_error)
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "surface_dynamics/interval_exchanges/iet_family.pyx":274
 *         return [F([self.ray_coefficient(i, j) for j in range(self.dim)]) for i in range(self.length)]
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         r"""
 *         TESTS::
 */

static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_10__repr__(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self) {
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__ *__pyx_cur_scope;
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);
  __pyx_cur_scope = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__ *)__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__(__pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 274, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":285
 *         """
 *         cdef size_t i, j
 *         s = []             # <<<<<<<<<<<<<<
 *         s.append('top ' + ' '.join(str(self.perm[i]) for i in range(self.dim)))
 *         s.append('bot ' + ' '.join(str(self.perm[self.dim+i]) for i in range(self.dim)))
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_s = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":286
 *         cdef size_t i, j
 *         s = []
 *         s.append('top ' + ' '.join(str(self.perm[i]) for i in range(self.dim)))             # <<<<<<<<<<<<<<
 *         s.append('bot ' + ' '.join(str(self.perm[self.dim+i]) for i in range(self.dim)))
 *         for i in range(self.length):
 */
  __pyx_t_1 = __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_8__repr___genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyString_Join(__pyx_kp_s__5, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_kp_s_top, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_s, __pyx_t_1); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":287
 *         s = []
 *         s.append('top ' + ' '.join(str(self.perm[i]) for i in range(self.dim)))
 *         s.append('bot ' + ' '.join(str(self.perm[self.dim+i]) for i in range(self.dim)))             # <<<<<<<<<<<<<<
 *         for i in range(self.length):
 *             s.append(' '.join(str(self.ray_coefficient(i, j)) for j in range(self.dim)))
 */
  __pyx_t_1 = __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_8__repr___3genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyString_Join(__pyx_kp_s__5, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_kp_s_bot, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_s, __pyx_t_1); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":288
 *         s.append('top ' + ' '.join(str(self.perm[i]) for i in range(self.dim)))
 *         s.append('bot ' + ' '.join(str(self.perm[self.dim+i]) for i in range(self.dim)))
 *         for i in range(self.length):             # <<<<<<<<<<<<<<
 *             s.append(' '.join(str(self.ray_coefficient(i, j)) for j in range(self.dim)))
 *         return '\n'.join(s)
 */
  __pyx_t_4 = __pyx_cur_scope->__pyx_v_self->length;
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_cur_scope->__pyx_v_i = __pyx_t_6;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":289
 *         s.append('bot ' + ' '.join(str(self.perm[self.dim+i]) for i in range(self.dim)))
 *         for i in range(self.length):
 *             s.append(' '.join(str(self.ray_coefficient(i, j)) for j in range(self.dim)))             # <<<<<<<<<<<<<<
 *         return '\n'.join(s)
 * 
 */
    __pyx_t_1 = __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_8__repr___6genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyString_Join(__pyx_kp_s__5, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 289, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_s, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 289, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":290
 *         for i in range(self.length):
 *             s.append(' '.join(str(self.ray_coefficient(i, j)) for j in range(self.dim)))
 *         return '\n'.join(s)             # <<<<<<<<<<<<<<
 * 
 *     def _debug_info(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyString_Join(__pyx_kp_s__6, __pyx_v_s); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":274
 *         return [F([self.ray_coefficient(i, j) for j in range(self.dim)]) for i in range(self.length)]
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         r"""
 *         TESTS::
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "surface_dynamics/interval_exchanges/iet_family.pyx":292
 *         return '\n'.join(s)
 * 
 *     def _debug_info(self):             # <<<<<<<<<<<<<<
 *         import sys
 *         sys.stdout.write('dim      %lu\n' % self.dim)
 */

/* Python wrapper */
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_13_debug_info(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_13_debug_info(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_debug_info (wrapper)", 0);
  __pyx_r = __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_12_debug_info(((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_12_debug_info(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self) {
  PyObject *__pyx_v_sys = NULL;
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  size_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_debug_info", 0);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":293
 * 
 *     def _debug_info(self):
 *         import sys             # <<<<<<<<<<<<<<
 *         sys.stdout.write('dim      %lu\n' % self.dim)
 *         sys.stdout.write('length   %lu\n' % self.length)
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_sys, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_sys = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":294
 *     def _debug_info(self):
 *         import sys
 *         sys.stdout.write('dim      %lu\n' % self.dim)             # <<<<<<<<<<<<<<
 *         sys.stdout.write('length   %lu\n' % self.length)
 *         sys.stdout.write('alloc    %lu\n' % self.alloc)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_stdout); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_self->dim); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_dim_lu, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":295
 *         import sys
 *         sys.stdout.write('dim      %lu\n' % self.dim)
 *         sys.stdout.write('length   %lu\n' % self.length)             # <<<<<<<<<<<<<<
 *         sys.stdout.write('alloc    %lu\n' % self.alloc)
 *         sys.stdout.flush()
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_stdout); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_write); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_self->length); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_length_lu, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":296
 *         sys.stdout.write('dim      %lu\n' % self.dim)
 *         sys.stdout.write('length   %lu\n' % self.length)
 *         sys.stdout.write('alloc    %lu\n' % self.alloc)             # <<<<<<<<<<<<<<
 *         sys.stdout.flush()
 *         sys.stdout.write('entries  %lu\n' % <size_t> self.entries)
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_stdout); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_write); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_FromSize_t(__pyx_v_self->alloc); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_alloc_lu, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":297
 *         sys.stdout.write('length   %lu\n' % self.length)
 *         sys.stdout.write('alloc    %lu\n' % self.alloc)
 *         sys.stdout.flush()             # <<<<<<<<<<<<<<
 *         sys.stdout.write('entries  %lu\n' % <size_t> self.entries)
 *         sys.stdout.flush()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_stdout); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_flush); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":298
 *         sys.stdout.write('alloc    %lu\n' % self.alloc)
 *         sys.stdout.flush()
 *         sys.stdout.write('entries  %lu\n' % <size_t> self.entries)             # <<<<<<<<<<<<<<
 *         sys.stdout.flush()
 *         sys.stdout.write('rows     %lu\n' % <size_t> self.rows)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_stdout); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_write); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_FromSize_t(((size_t)__pyx_v_self->entries)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_entries_lu, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":299
 *         sys.stdout.flush()
 *         sys.stdout.write('entries  %lu\n' % <size_t> self.entries)
 *         sys.stdout.flush()             # <<<<<<<<<<<<<<
 *         sys.stdout.write('rows     %lu\n' % <size_t> self.rows)
 *         sys.stdout.flush()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_stdout); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_flush); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":300
 *         sys.stdout.write('entries  %lu\n' % <size_t> self.entries)
 *         sys.stdout.flush()
 *         sys.stdout.write('rows     %lu\n' % <size_t> self.rows)             # <<<<<<<<<<<<<<
 *         sys.stdout.flush()
 *         cdef size_t i
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_stdout); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_write); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyInt_FromSize_t(((size_t)__pyx_v_self->rows)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_rows_lu, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":301
 *         sys.stdout.flush()
 *         sys.stdout.write('rows     %lu\n' % <size_t> self.rows)
 *         sys.stdout.flush()             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         for i in range(self.alloc):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_stdout); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_flush); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":303
 *         sys.stdout.flush()
 *         cdef size_t i
 *         for i in range(self.alloc):             # <<<<<<<<<<<<<<
 *             sys.stdout.write('rows[%d]  %lu\n' %(i, <size_t> (self.rows[i] - self.entries)))
 *             sys.stdout.flush()
 */
  __pyx_t_5 = __pyx_v_self->alloc;
  __pyx_t_6 = __pyx_t_5;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":304
 *         cdef size_t i
 *         for i in range(self.alloc):
 *             sys.stdout.write('rows[%d]  %lu\n' %(i, <size_t> (self.rows[i] - self.entries)))             # <<<<<<<<<<<<<<
 *             sys.stdout.flush()
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_stdout); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 304, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_write); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 304, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 304, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyInt_FromSize_t(((size_t)((__pyx_v_self->rows[__pyx_v_i]) - __pyx_v_self->entries))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 304, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 304, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_4);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_rows_d_lu, __pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 304, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_8, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 304, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":305
 *         for i in range(self.alloc):
 *             sys.stdout.write('rows[%d]  %lu\n' %(i, <size_t> (self.rows[i] - self.entries)))
 *             sys.stdout.flush()             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sys, __pyx_n_s_stdout); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 305, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_flush); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 305, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 305, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":292
 *         return '\n'.join(s)
 * 
 *     def _debug_info(self):             # <<<<<<<<<<<<<<
 *         import sys
 *         sys.stdout.write('dim      %lu\n' % self.dim)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily._debug_info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_sys);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "surface_dynamics/interval_exchanges/iet_family.pyx":308
 * 
 * 
 *     cdef void fit_length(self, size_t length):             # <<<<<<<<<<<<<<
 *         r"""
 *         Manage allocation
 */

static void __pyx_f_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_fit_length(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self, size_t __pyx_v_length) {
  size_t __pyx_v_i;
  mpz_t *__pyx_v_old_entries;
  ptrdiff_t __pyx_v_offset;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  mpz_t *__pyx_t_5;
  size_t __pyx_t_6;
  __Pyx_RefNannySetupContext("fit_length", 0);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":318
 *         cdef mpz_t * old_entries
 *         cdef ptrdiff_t offset
 *         if length > self.alloc:             # <<<<<<<<<<<<<<
 *             length = max(length, 2 * self.alloc)
 * 
 */
  __pyx_t_1 = ((__pyx_v_length > __pyx_v_self->alloc) != 0);
  if (__pyx_t_1) {

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":319
 *         cdef ptrdiff_t offset
 *         if length > self.alloc:
 *             length = max(length, 2 * self.alloc)             # <<<<<<<<<<<<<<
 * 
 *             old_entries = self.entries
 */
    __pyx_t_2 = (2 * __pyx_v_self->alloc);
    __pyx_t_3 = __pyx_v_length;
    if (((__pyx_t_2 > __pyx_t_3) != 0)) {
      __pyx_t_4 = __pyx_t_2;
    } else {
      __pyx_t_4 = __pyx_t_3;
    }
    __pyx_v_length = __pyx_t_4;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":321
 *             length = max(length, 2 * self.alloc)
 * 
 *             old_entries = self.entries             # <<<<<<<<<<<<<<
 *             self.entries = <mpz_t *> realloc(self.entries, self.dim * length * sizeof(mpz_t))
 *             offset = self.entries - old_entries
 */
    __pyx_t_5 = __pyx_v_self->entries;
    __pyx_v_old_entries = __pyx_t_5;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":322
 * 
 *             old_entries = self.entries
 *             self.entries = <mpz_t *> realloc(self.entries, self.dim * length * sizeof(mpz_t))             # <<<<<<<<<<<<<<
 *             offset = self.entries - old_entries
 * 
 */
    __pyx_v_self->entries = ((mpz_t *)realloc(__pyx_v_self->entries, ((__pyx_v_self->dim * __pyx_v_length) * (sizeof(mpz_t)))));

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":323
 *             old_entries = self.entries
 *             self.entries = <mpz_t *> realloc(self.entries, self.dim * length * sizeof(mpz_t))
 *             offset = self.entries - old_entries             # <<<<<<<<<<<<<<
 * 
 *             for i in range(self.dim * self.alloc, self.dim * length):
 */
    __pyx_v_offset = (__pyx_v_self->entries - __pyx_v_old_entries);

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":325
 *             offset = self.entries - old_entries
 * 
 *             for i in range(self.dim * self.alloc, self.dim * length):             # <<<<<<<<<<<<<<
 *                 mpz_init(self.entries[i])
 * 
 */
    __pyx_t_4 = (__pyx_v_self->dim * __pyx_v_length);
    __pyx_t_2 = __pyx_t_4;
    for (__pyx_t_3 = (__pyx_v_self->dim * __pyx_v_self->alloc); __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
      __pyx_v_i = __pyx_t_3;

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":326
 * 
 *             for i in range(self.dim * self.alloc, self.dim * length):
 *                 mpz_init(self.entries[i])             # <<<<<<<<<<<<<<
 * 
 *             self.rows = <mpz_t **> realloc(self.rows, length * sizeof(mpz_t *))
 */
      mpz_init((__pyx_v_self->entries[__pyx_v_i]));
    }

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":328
 *                 mpz_init(self.entries[i])
 * 
 *             self.rows = <mpz_t **> realloc(self.rows, length * sizeof(mpz_t *))             # <<<<<<<<<<<<<<
 *             if offset:
 *                 for i in range(self.alloc):
 */
    __pyx_v_self->rows = ((mpz_t **)realloc(__pyx_v_self->rows, (__pyx_v_length * (sizeof(mpz_t *)))));

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":329
 * 
 *             self.rows = <mpz_t **> realloc(self.rows, length * sizeof(mpz_t *))
 *             if offset:             # <<<<<<<<<<<<<<
 *                 for i in range(self.alloc):
 *                     self.rows[i] += offset
 */
    __pyx_t_1 = (__pyx_v_offset != 0);
    if (__pyx_t_1) {

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":330
 *             self.rows = <mpz_t **> realloc(self.rows, length * sizeof(mpz_t *))
 *             if offset:
 *                 for i in range(self.alloc):             # <<<<<<<<<<<<<<
 *                     self.rows[i] += offset
 *             for i in range(self.alloc, length):
 */
      __pyx_t_4 = __pyx_v_self->alloc;
      __pyx_t_2 = __pyx_t_4;
      for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
        __pyx_v_i = __pyx_t_3;

        /* "surface_dynamics/interval_exchanges/iet_family.pyx":331
 *             if offset:
 *                 for i in range(self.alloc):
 *                     self.rows[i] += offset             # <<<<<<<<<<<<<<
 *             for i in range(self.alloc, length):
 *                 self.rows[i] = self.entries + (i * self.dim)
 */
        __pyx_t_6 = __pyx_v_i;
        (__pyx_v_self->rows[__pyx_t_6]) = ((__pyx_v_self->rows[__pyx_t_6]) + __pyx_v_offset);
      }

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":329
 * 
 *             self.rows = <mpz_t **> realloc(self.rows, length * sizeof(mpz_t *))
 *             if offset:             # <<<<<<<<<<<<<<
 *                 for i in range(self.alloc):
 *                     self.rows[i] += offset
 */
    }

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":332
 *                 for i in range(self.alloc):
 *                     self.rows[i] += offset
 *             for i in range(self.alloc, length):             # <<<<<<<<<<<<<<
 *                 self.rows[i] = self.entries + (i * self.dim)
 *             self.alloc = length
 */
    __pyx_t_4 = __pyx_v_length;
    __pyx_t_2 = __pyx_t_4;
    for (__pyx_t_3 = __pyx_v_self->alloc; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
      __pyx_v_i = __pyx_t_3;

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":333
 *                     self.rows[i] += offset
 *             for i in range(self.alloc, length):
 *                 self.rows[i] = self.entries + (i * self.dim)             # <<<<<<<<<<<<<<
 *             self.alloc = length
 * 
 */
      (__pyx_v_self->rows[__pyx_v_i]) = (__pyx_v_self->entries + (__pyx_v_i * __pyx_v_self->dim));
    }

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":334
 *             for i in range(self.alloc, length):
 *                 self.rows[i] = self.entries + (i * self.dim)
 *             self.alloc = length             # <<<<<<<<<<<<<<
 * 
 *     cdef void fill_rays(self, const PPL_Generator_System& gs):
 */
    __pyx_v_self->alloc = __pyx_v_length;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":318
 *         cdef mpz_t * old_entries
 *         cdef ptrdiff_t offset
 *         if length > self.alloc:             # <<<<<<<<<<<<<<
 *             length = max(length, 2 * self.alloc)
 * 
 */
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":308
 * 
 * 
 *     cdef void fit_length(self, size_t length):             # <<<<<<<<<<<<<<
 *         r"""
 *         Manage allocation
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "surface_dynamics/interval_exchanges/iet_family.pyx":336
 *             self.alloc = length
 * 
 *     cdef void fill_rays(self, const PPL_Generator_System& gs):             # <<<<<<<<<<<<<<
 *         r"""
 *         Set the attributes related to the cone
 */

static void __pyx_f_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_fill_rays(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self, Parma_Polyhedra_Library::Generator_System const &__pyx_v_gs) {
  Parma_Polyhedra_Library::Generator_System_const_iterator *__pyx_v_gsi_ptr;
  size_t __pyx_v_j;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  __Pyx_RefNannySetupContext("fill_rays", 0);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":340
 *         Set the attributes related to the cone
 *         """
 *         cdef PPL_gs_iterator * gsi_ptr = new PPL_gs_iterator(gs.begin())             # <<<<<<<<<<<<<<
 *         cdef size_t j
 *         while gsi_ptr[0] != gs.end():
 */
  __pyx_v_gsi_ptr = new Parma_Polyhedra_Library::Generator_System_const_iterator(__pyx_v_gs.begin());

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":342
 *         cdef PPL_gs_iterator * gsi_ptr = new PPL_gs_iterator(gs.begin())
 *         cdef size_t j
 *         while gsi_ptr[0] != gs.end():             # <<<<<<<<<<<<<<
 *             if deref(gsi_ptr[0]).is_ray():
 *                 self.fit_length(self.length + 1)
 */
  while (1) {
    __pyx_t_1 = (((__pyx_v_gsi_ptr[0]) != __pyx_v_gs.end()) != 0);
    if (!__pyx_t_1) break;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":343
 *         cdef size_t j
 *         while gsi_ptr[0] != gs.end():
 *             if deref(gsi_ptr[0]).is_ray():             # <<<<<<<<<<<<<<
 *                 self.fit_length(self.length + 1)
 *                 for j in range(self.dim):
 */
    __pyx_t_1 = ((*(__pyx_v_gsi_ptr[0])).is_ray() != 0);
    if (__pyx_t_1) {

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":344
 *         while gsi_ptr[0] != gs.end():
 *             if deref(gsi_ptr[0]).is_ray():
 *                 self.fit_length(self.length + 1)             # <<<<<<<<<<<<<<
 *                 for j in range(self.dim):
 *                     mpz_set(self.rows[self.length][j],
 */
      ((struct __pyx_vtabstruct_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)__pyx_v_self->__pyx_vtab)->fit_length(__pyx_v_self, (__pyx_v_self->length + 1));

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":345
 *             if deref(gsi_ptr[0]).is_ray():
 *                 self.fit_length(self.length + 1)
 *                 for j in range(self.dim):             # <<<<<<<<<<<<<<
 *                     mpz_set(self.rows[self.length][j],
 *                             deref(gsi_ptr[0]).coefficient(PPL_Variable(j)).get_mpz_t())
 */
      __pyx_t_2 = __pyx_v_self->dim;
      __pyx_t_3 = __pyx_t_2;
      for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
        __pyx_v_j = __pyx_t_4;

        /* "surface_dynamics/interval_exchanges/iet_family.pyx":346
 *                 self.fit_length(self.length + 1)
 *                 for j in range(self.dim):
 *                     mpz_set(self.rows[self.length][j],             # <<<<<<<<<<<<<<
 *                             deref(gsi_ptr[0]).coefficient(PPL_Variable(j)).get_mpz_t())
 *                 self.length += 1
 */
        mpz_set(((__pyx_v_self->rows[__pyx_v_self->length])[__pyx_v_j]), (*(__pyx_v_gsi_ptr[0])).coefficient(Parma_Polyhedra_Library::Variable(__pyx_v_j)).get_mpz_t());
      }

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":348
 *                     mpz_set(self.rows[self.length][j],
 *                             deref(gsi_ptr[0]).coefficient(PPL_Variable(j)).get_mpz_t())
 *                 self.length += 1             # <<<<<<<<<<<<<<
 *             gsi_ptr[0].inc(1)
 *         del gsi_ptr
 */
      __pyx_v_self->length = (__pyx_v_self->length + 1);

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":343
 *         cdef size_t j
 *         while gsi_ptr[0] != gs.end():
 *             if deref(gsi_ptr[0]).is_ray():             # <<<<<<<<<<<<<<
 *                 self.fit_length(self.length + 1)
 *                 for j in range(self.dim):
 */
    }

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":349
 *                             deref(gsi_ptr[0]).coefficient(PPL_Variable(j)).get_mpz_t())
 *                 self.length += 1
 *             gsi_ptr[0].inc(1)             # <<<<<<<<<<<<<<
 *         del gsi_ptr
 *         qsort(self.rows, self.length, sizeof(mpz_t *), & compare)
 */
    (void)((__pyx_v_gsi_ptr[0]).operator++(1));
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":350
 *                 self.length += 1
 *             gsi_ptr[0].inc(1)
 *         del gsi_ptr             # <<<<<<<<<<<<<<
 *         qsort(self.rows, self.length, sizeof(mpz_t *), & compare)
 * 
 */
  delete __pyx_v_gsi_ptr;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":351
 *             gsi_ptr[0].inc(1)
 *         del gsi_ptr
 *         qsort(self.rows, self.length, sizeof(mpz_t *), & compare)             # <<<<<<<<<<<<<<
 * 
 *     def permutation(self):
 */
  qsort(__pyx_v_self->rows, __pyx_v_self->length, (sizeof(mpz_t *)), (&__pyx_f_16surface_dynamics_18interval_exchanges_10iet_family_compare));

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":336
 *             self.alloc = length
 * 
 *     cdef void fill_rays(self, const PPL_Generator_System& gs):             # <<<<<<<<<<<<<<
 *         r"""
 *         Set the attributes related to the cone
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "surface_dynamics/interval_exchanges/iet_family.pyx":353
 *         qsort(self.rows, self.length, sizeof(mpz_t *), & compare)
 * 
 *     def permutation(self):             # <<<<<<<<<<<<<<
 *         r"""
 *         Return the permutation as a list of integers
 */

/* Python wrapper */
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_15permutation(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_14permutation[] = "\n        Return the permutation as a list of integers\n        ";
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_15permutation(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("permutation (wrapper)", 0);
  __pyx_r = __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_14permutation(((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_14permutation(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self) {
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("permutation", 0);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":358
 *         """
 *         cdef size_t i
 *         return [[self.perm[i] for i in range(self.dim)], [self.perm[self.dim+i] for i in range(self.dim)]]             # <<<<<<<<<<<<<<
 * 
 *     def __hash__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_v_self->dim;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;
    __pyx_t_5 = __Pyx_PyInt_From_int((__pyx_v_self->perm[__pyx_v_i])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __pyx_v_self->dim;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;
    __pyx_t_6 = __Pyx_PyInt_From_int((__pyx_v_self->perm[(__pyx_v_self->dim + __pyx_v_i)])); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __pyx_t_6 = PyList_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyList_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_5);
  PyList_SET_ITEM(__pyx_t_6, 1, __pyx_t_5);
  __pyx_t_1 = 0;
  __pyx_t_5 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":353
 *         qsort(self.rows, self.length, sizeof(mpz_t *), & compare)
 * 
 *     def permutation(self):             # <<<<<<<<<<<<<<
 *         r"""
 *         Return the permutation as a list of integers
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.permutation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "surface_dynamics/interval_exchanges/iet_family.pyx":360
 *         return [[self.perm[i] for i in range(self.dim)], [self.perm[self.dim+i] for i in range(self.dim)]]
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         r"""
 *         EXAMPLES::
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_17__hash__(PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_16__hash__[] = "\n        EXAMPLES::\n\n            sage: from surface_dynamics import *\n            sage: p = iet.Permutation('a b c d', 'd c b a')\n            sage: F = iet.IETFamily(p, Polyhedron(rays=(ZZ**4).basis())) # optional - pplpy\n            sage: hash(F) # optional - pplpy\n            4696861030007372737\n        ";
#if CYTHON_COMPILING_IN_CPYTHON
struct wrapperbase __pyx_wrapperbase_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_16__hash__;
#endif
static Py_hash_t __pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_17__hash__(PyObject *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_16__hash__(((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_16__hash__(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self) {
  Py_hash_t __pyx_v_x;
  Py_hash_t __pyx_v_y;
  Py_hash_t __pyx_v_mult;
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":373
 *         cdef size_t i, j
 * 
 *         mult = 3             # <<<<<<<<<<<<<<
 *         x = 0x345678
 *         for i in range(self.dim):
 */
  __pyx_v_mult = 3;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":374
 * 
 *         mult = 3
 *         x = 0x345678             # <<<<<<<<<<<<<<
 *         for i in range(self.dim):
 *             x = (x ^ (self.perm[i])) * mult
 */
  __pyx_v_x = 0x345678;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":375
 *         mult = 3
 *         x = 0x345678
 *         for i in range(self.dim):             # <<<<<<<<<<<<<<
 *             x = (x ^ (self.perm[i])) * mult
 *             mult += 1345
 */
  __pyx_t_1 = __pyx_v_self->dim;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":376
 *         x = 0x345678
 *         for i in range(self.dim):
 *             x = (x ^ (self.perm[i])) * mult             # <<<<<<<<<<<<<<
 *             mult += 1345
 *             x += 35
 */
    __pyx_v_x = ((__pyx_v_x ^ (__pyx_v_self->perm[__pyx_v_i])) * __pyx_v_mult);

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":377
 *         for i in range(self.dim):
 *             x = (x ^ (self.perm[i])) * mult
 *             mult += 1345             # <<<<<<<<<<<<<<
 *             x += 35
 * 
 */
    __pyx_v_mult = (__pyx_v_mult + 0x541);

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":378
 *             x = (x ^ (self.perm[i])) * mult
 *             mult += 1345
 *             x += 35             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.length):
 */
    __pyx_v_x = (__pyx_v_x + 35);
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":380
 *             x += 35
 * 
 *         for i in range(self.length):             # <<<<<<<<<<<<<<
 *             for j in range(self.dim):
 *                 y = mpz_pythonhash(self.rows[i][j])
 */
  __pyx_t_1 = __pyx_v_self->length;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":381
 * 
 *         for i in range(self.length):
 *             for j in range(self.dim):             # <<<<<<<<<<<<<<
 *                 y = mpz_pythonhash(self.rows[i][j])
 *                 x = (x ^ y) * mult
 */
    __pyx_t_4 = __pyx_v_self->dim;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_j = __pyx_t_6;

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":382
 *         for i in range(self.length):
 *             for j in range(self.dim):
 *                 y = mpz_pythonhash(self.rows[i][j])             # <<<<<<<<<<<<<<
 *                 x = (x ^ y) * mult
 *                 mult += 82520
 */
      __pyx_v_y = __pyx_f_16surface_dynamics_18interval_exchanges_10iet_family_mpz_pythonhash(((__pyx_v_self->rows[__pyx_v_i])[__pyx_v_j]));

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":383
 *             for j in range(self.dim):
 *                 y = mpz_pythonhash(self.rows[i][j])
 *                 x = (x ^ y) * mult             # <<<<<<<<<<<<<<
 *                 mult += 82520
 *             x += 97531
 */
      __pyx_v_x = ((__pyx_v_x ^ __pyx_v_y) * __pyx_v_mult);

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":384
 *                 y = mpz_pythonhash(self.rows[i][j])
 *                 x = (x ^ y) * mult
 *                 mult += 82520             # <<<<<<<<<<<<<<
 *             x += 97531
 * 
 */
      __pyx_v_mult = (__pyx_v_mult + 0x14258);
    }

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":385
 *                 x = (x ^ y) * mult
 *                 mult += 82520
 *             x += 97531             # <<<<<<<<<<<<<<
 * 
 *         return x
 */
    __pyx_v_x = (__pyx_v_x + 0x17CFB);
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":387
 *             x += 97531
 * 
 *         return x             # <<<<<<<<<<<<<<
 * 
 *     def __richcmp__(IETFamily self, IETFamily other, int op):
 */
  __pyx_r = __pyx_v_x;
  goto __pyx_L0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":360
 *         return [[self.perm[i] for i in range(self.dim)], [self.perm[self.dim+i] for i in range(self.dim)]]
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         r"""
 *         EXAMPLES::
 */

  /* function exit code */
  __pyx_L0:;
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "surface_dynamics/interval_exchanges/iet_family.pyx":389
 *         return x
 * 
 *     def __richcmp__(IETFamily self, IETFamily other, int op):             # <<<<<<<<<<<<<<
 *         r"""
 *         TESTS::
 */

/* Python wrapper */
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_19__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op); /*proto*/
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_19__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_v_op) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily, 1, "other", 0))) __PYX_ERR(0, 389, __pyx_L1_error)
  __pyx_r = __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_18__richcmp__(((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)__pyx_v_self), ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)__pyx_v_other), ((int)__pyx_v_op));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_18__richcmp__(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self, struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_other, int __pyx_v_op) {
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":426
 *             True
 *         """
 *         if op != Py_EQ and op != Py_NE:             # <<<<<<<<<<<<<<
 *             raise TypeError('only equality and difference tests available')
 *         if self.dim != other.dim or self.length != other.length:
 */
  __pyx_t_2 = ((__pyx_v_op != Py_EQ) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_op != Py_NE) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":427
 *         """
 *         if op != Py_EQ and op != Py_NE:
 *             raise TypeError('only equality and difference tests available')             # <<<<<<<<<<<<<<
 *         if self.dim != other.dim or self.length != other.length:
 *             return op == Py_NE
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 427, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 427, __pyx_L1_error)

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":426
 *             True
 *         """
 *         if op != Py_EQ and op != Py_NE:             # <<<<<<<<<<<<<<
 *             raise TypeError('only equality and difference tests available')
 *         if self.dim != other.dim or self.length != other.length:
 */
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":428
 *         if op != Py_EQ and op != Py_NE:
 *             raise TypeError('only equality and difference tests available')
 *         if self.dim != other.dim or self.length != other.length:             # <<<<<<<<<<<<<<
 *             return op == Py_NE
 * 
 */
  __pyx_t_2 = ((__pyx_v_self->dim != __pyx_v_other->dim) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_self->length != __pyx_v_other->length) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_1) {

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":429
 *             raise TypeError('only equality and difference tests available')
 *         if self.dim != other.dim or self.length != other.length:
 *             return op == Py_NE             # <<<<<<<<<<<<<<
 * 
 *         cdef size_t i,j
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyBool_FromLong((__pyx_v_op == Py_NE)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":428
 *         if op != Py_EQ and op != Py_NE:
 *             raise TypeError('only equality and difference tests available')
 *         if self.dim != other.dim or self.length != other.length:             # <<<<<<<<<<<<<<
 *             return op == Py_NE
 * 
 */
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":432
 * 
 *         cdef size_t i,j
 *         for i in range(2 * self.dim):             # <<<<<<<<<<<<<<
 *             if self.perm[i] != other.perm[i]:
 *                 return op == Py_NE
 */
  __pyx_t_4 = (2 * __pyx_v_self->dim);
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":433
 *         cdef size_t i,j
 *         for i in range(2 * self.dim):
 *             if self.perm[i] != other.perm[i]:             # <<<<<<<<<<<<<<
 *                 return op == Py_NE
 *         for i in range(self.length):
 */
    __pyx_t_1 = (((__pyx_v_self->perm[__pyx_v_i]) != (__pyx_v_other->perm[__pyx_v_i])) != 0);
    if (__pyx_t_1) {

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":434
 *         for i in range(2 * self.dim):
 *             if self.perm[i] != other.perm[i]:
 *                 return op == Py_NE             # <<<<<<<<<<<<<<
 *         for i in range(self.length):
 *             for j in range(self.dim):
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyBool_FromLong((__pyx_v_op == Py_NE)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 434, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":433
 *         cdef size_t i,j
 *         for i in range(2 * self.dim):
 *             if self.perm[i] != other.perm[i]:             # <<<<<<<<<<<<<<
 *                 return op == Py_NE
 *         for i in range(self.length):
 */
    }
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":435
 *             if self.perm[i] != other.perm[i]:
 *                 return op == Py_NE
 *         for i in range(self.length):             # <<<<<<<<<<<<<<
 *             for j in range(self.dim):
 *                 if mpz_cmp(self.rows[i][j], other.rows[i][j]):
 */
  __pyx_t_4 = __pyx_v_self->length;
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":436
 *                 return op == Py_NE
 *         for i in range(self.length):
 *             for j in range(self.dim):             # <<<<<<<<<<<<<<
 *                 if mpz_cmp(self.rows[i][j], other.rows[i][j]):
 *                     return op == Py_NE
 */
    __pyx_t_7 = __pyx_v_self->dim;
    __pyx_t_8 = __pyx_t_7;
    for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
      __pyx_v_j = __pyx_t_9;

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":437
 *         for i in range(self.length):
 *             for j in range(self.dim):
 *                 if mpz_cmp(self.rows[i][j], other.rows[i][j]):             # <<<<<<<<<<<<<<
 *                     return op == Py_NE
 * 
 */
      __pyx_t_1 = (mpz_cmp(((__pyx_v_self->rows[__pyx_v_i])[__pyx_v_j]), ((__pyx_v_other->rows[__pyx_v_i])[__pyx_v_j])) != 0);
      if (__pyx_t_1) {

        /* "surface_dynamics/interval_exchanges/iet_family.pyx":438
 *             for j in range(self.dim):
 *                 if mpz_cmp(self.rows[i][j], other.rows[i][j]):
 *                     return op == Py_NE             # <<<<<<<<<<<<<<
 * 
 *         return op == Py_EQ
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyBool_FromLong((__pyx_v_op == Py_NE)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 438, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_r = __pyx_t_3;
        __pyx_t_3 = 0;
        goto __pyx_L0;

        /* "surface_dynamics/interval_exchanges/iet_family.pyx":437
 *         for i in range(self.length):
 *             for j in range(self.dim):
 *                 if mpz_cmp(self.rows[i][j], other.rows[i][j]):             # <<<<<<<<<<<<<<
 *                     return op == Py_NE
 * 
 */
      }
    }
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":440
 *                     return op == Py_NE
 * 
 *         return op == Py_EQ             # <<<<<<<<<<<<<<
 * 
 *     # this is completely useless and checks many times the same thing
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyBool_FromLong((__pyx_v_op == Py_EQ)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":389
 *         return x
 * 
 *     def __richcmp__(IETFamily self, IETFamily other, int op):             # <<<<<<<<<<<<<<
 *         r"""
 *         TESTS::
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "surface_dynamics/interval_exchanges/iet_family.pyx":444
 *     # this is completely useless and checks many times the same thing
 *     # we should just look at left/right Rauzy induction
 *     def has_connection(self):             # <<<<<<<<<<<<<<
 *         r"""
 *         Check whether this family has a connection.
 */

/* Python wrapper */
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_21has_connection(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_20has_connection[] = "\n        Check whether this family has a connection.\n\n        EXAMPLES::\n\n            sage: from surface_dynamics import *\n            sage: p = iet.Permutation('A B C D E', 'E D C B A')\n            sage: C0 = Polyhedron(rays=(ZZ**5).basis())\n            sage: iet.IETFamily(p, C0).has_connection() # optional - pplpy\n            False\n\n            sage: C = Polyhedron(rays=[(1,2,3,2,3),(1,1,0,0,2)])\n            sage: iet.IETFamily(p, C).has_connection() # optional - pplpy\n            True\n\n            sage: C = Polyhedron(rays=[(1,1,1,1,4),(2,0,1,2,5)])\n            sage: iet.IETFamily(p, C).has_connection() # optional - pplpy\n            True\n\n            sage: C = Polyhedron(rays=[(1,0,0,1,0),(1,0,0,0,1),(0,1,0,1,0),(0,1,0,0,1)])\n            sage: iet.IETFamily(p, C).has_connection() # optional - pplpy\n            True\n\n            sage: C = Polyhedron(rays=[(1,2,2,2,2)])\n            sage: iet.IETFamily(p, C).has_connection() # optional - pplpy\n            False\n        ";
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_21has_connection(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("has_connection (wrapper)", 0);
  __pyx_r = __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_20has_connection(((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_20has_connection(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self) {
  int __pyx_v_ans;
  mpz_t *__pyx_v_x;
  mpz_t *__pyx_v_y;
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  size_t __pyx_v_k;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  void *__pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("has_connection", 0);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":474
 *         cdef bint ans
 * 
 *         cdef mpz_t * x = <mpz_t *> check_malloc(2 * self.length * sizeof(mpz_t))             # <<<<<<<<<<<<<<
 *         cdef mpz_t * y = x + self.length
 *         cdef size_t i, j, k
 */
  __pyx_t_1 = __pyx_f_9cysignals_6memory_check_malloc(((2 * __pyx_v_self->length) * (sizeof(mpz_t)))); if (unlikely(__pyx_t_1 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 474, __pyx_L1_error)
  __pyx_v_x = ((mpz_t *)__pyx_t_1);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":475
 * 
 *         cdef mpz_t * x = <mpz_t *> check_malloc(2 * self.length * sizeof(mpz_t))
 *         cdef mpz_t * y = x + self.length             # <<<<<<<<<<<<<<
 *         cdef size_t i, j, k
 *         for i in range(self.length):
 */
  __pyx_v_y = (__pyx_v_x + __pyx_v_self->length);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":477
 *         cdef mpz_t * y = x + self.length
 *         cdef size_t i, j, k
 *         for i in range(self.length):             # <<<<<<<<<<<<<<
 *             mpz_init(x[i])
 *             mpz_set(x[i], self.rows[i][self.perm[0]])
 */
  __pyx_t_2 = __pyx_v_self->length;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":478
 *         cdef size_t i, j, k
 *         for i in range(self.length):
 *             mpz_init(x[i])             # <<<<<<<<<<<<<<
 *             mpz_set(x[i], self.rows[i][self.perm[0]])
 *             mpz_init(y[i])
 */
    mpz_init((__pyx_v_x[__pyx_v_i]));

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":479
 *         for i in range(self.length):
 *             mpz_init(x[i])
 *             mpz_set(x[i], self.rows[i][self.perm[0]])             # <<<<<<<<<<<<<<
 *             mpz_init(y[i])
 *             mpz_set(y[i], self.rows[i][self.perm[self.dim]])
 */
    mpz_set((__pyx_v_x[__pyx_v_i]), ((__pyx_v_self->rows[__pyx_v_i])[(__pyx_v_self->perm[0])]));

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":480
 *             mpz_init(x[i])
 *             mpz_set(x[i], self.rows[i][self.perm[0]])
 *             mpz_init(y[i])             # <<<<<<<<<<<<<<
 *             mpz_set(y[i], self.rows[i][self.perm[self.dim]])
 * 
 */
    mpz_init((__pyx_v_y[__pyx_v_i]));

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":481
 *             mpz_set(x[i], self.rows[i][self.perm[0]])
 *             mpz_init(y[i])
 *             mpz_set(y[i], self.rows[i][self.perm[self.dim]])             # <<<<<<<<<<<<<<
 * 
 *         j = 1
 */
    mpz_set((__pyx_v_y[__pyx_v_i]), ((__pyx_v_self->rows[__pyx_v_i])[(__pyx_v_self->perm[__pyx_v_self->dim])]));
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":483
 *             mpz_set(y[i], self.rows[i][self.perm[self.dim]])
 * 
 *         j = 1             # <<<<<<<<<<<<<<
 *         for i in range(1, self.dim):
 *             while j < self.dim and mpz_vec_any_smaller(y, x, self.length):
 */
  __pyx_v_j = 1;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":484
 * 
 *         j = 1
 *         for i in range(1, self.dim):             # <<<<<<<<<<<<<<
 *             while j < self.dim and mpz_vec_any_smaller(y, x, self.length):
 *                 for k in range(self.length):
 */
  __pyx_t_2 = __pyx_v_self->dim;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 1; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":485
 *         j = 1
 *         for i in range(1, self.dim):
 *             while j < self.dim and mpz_vec_any_smaller(y, x, self.length):             # <<<<<<<<<<<<<<
 *                 for k in range(self.length):
 *                     mpz_add(y[k], y[k], self.rows[k][self.perm[self.dim+j]])
 */
    while (1) {
      __pyx_t_6 = ((__pyx_v_j < __pyx_v_self->dim) != 0);
      if (__pyx_t_6) {
      } else {
        __pyx_t_5 = __pyx_t_6;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_6 = (__pyx_f_16surface_dynamics_18interval_exchanges_10iet_family_mpz_vec_any_smaller(__pyx_v_y, __pyx_v_x, __pyx_v_self->length) != 0);
      __pyx_t_5 = __pyx_t_6;
      __pyx_L9_bool_binop_done:;
      if (!__pyx_t_5) break;

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":486
 *         for i in range(1, self.dim):
 *             while j < self.dim and mpz_vec_any_smaller(y, x, self.length):
 *                 for k in range(self.length):             # <<<<<<<<<<<<<<
 *                     mpz_add(y[k], y[k], self.rows[k][self.perm[self.dim+j]])
 *                 j += 1
 */
      __pyx_t_7 = __pyx_v_self->length;
      __pyx_t_8 = __pyx_t_7;
      for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
        __pyx_v_k = __pyx_t_9;

        /* "surface_dynamics/interval_exchanges/iet_family.pyx":487
 *             while j < self.dim and mpz_vec_any_smaller(y, x, self.length):
 *                 for k in range(self.length):
 *                     mpz_add(y[k], y[k], self.rows[k][self.perm[self.dim+j]])             # <<<<<<<<<<<<<<
 *                 j += 1
 *             if mpz_vec_equal(x, y, self.length):
 */
        mpz_add((__pyx_v_y[__pyx_v_k]), (__pyx_v_y[__pyx_v_k]), ((__pyx_v_self->rows[__pyx_v_k])[(__pyx_v_self->perm[(__pyx_v_self->dim + __pyx_v_j)])]));
      }

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":488
 *                 for k in range(self.length):
 *                     mpz_add(y[k], y[k], self.rows[k][self.perm[self.dim+j]])
 *                 j += 1             # <<<<<<<<<<<<<<
 *             if mpz_vec_equal(x, y, self.length):
 *                 ans = True
 */
      __pyx_v_j = (__pyx_v_j + 1);
    }

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":489
 *                     mpz_add(y[k], y[k], self.rows[k][self.perm[self.dim+j]])
 *                 j += 1
 *             if mpz_vec_equal(x, y, self.length):             # <<<<<<<<<<<<<<
 *                 ans = True
 *                 break
 */
    __pyx_t_5 = (__pyx_f_16surface_dynamics_18interval_exchanges_10iet_family_mpz_vec_equal(__pyx_v_x, __pyx_v_y, __pyx_v_self->length) != 0);
    if (__pyx_t_5) {

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":490
 *                 j += 1
 *             if mpz_vec_equal(x, y, self.length):
 *                 ans = True             # <<<<<<<<<<<<<<
 *                 break
 *             for k in range(self.length):
 */
      __pyx_v_ans = 1;

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":491
 *             if mpz_vec_equal(x, y, self.length):
 *                 ans = True
 *                 break             # <<<<<<<<<<<<<<
 *             for k in range(self.length):
 *                 mpz_add(x[k], x[k], self.rows[k][self.perm[i]])
 */
      goto __pyx_L6_break;

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":489
 *                     mpz_add(y[k], y[k], self.rows[k][self.perm[self.dim+j]])
 *                 j += 1
 *             if mpz_vec_equal(x, y, self.length):             # <<<<<<<<<<<<<<
 *                 ans = True
 *                 break
 */
    }

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":492
 *                 ans = True
 *                 break
 *             for k in range(self.length):             # <<<<<<<<<<<<<<
 *                 mpz_add(x[k], x[k], self.rows[k][self.perm[i]])
 *         else:
 */
    __pyx_t_7 = __pyx_v_self->length;
    __pyx_t_8 = __pyx_t_7;
    for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
      __pyx_v_k = __pyx_t_9;

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":493
 *                 break
 *             for k in range(self.length):
 *                 mpz_add(x[k], x[k], self.rows[k][self.perm[i]])             # <<<<<<<<<<<<<<
 *         else:
 *             ans = False
 */
      mpz_add((__pyx_v_x[__pyx_v_k]), (__pyx_v_x[__pyx_v_k]), ((__pyx_v_self->rows[__pyx_v_k])[(__pyx_v_self->perm[__pyx_v_i])]));
    }
  }
  /*else*/ {

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":495
 *                 mpz_add(x[k], x[k], self.rows[k][self.perm[i]])
 *         else:
 *             ans = False             # <<<<<<<<<<<<<<
 * 
 *         for i in range(2 * self.length):
 */
    __pyx_v_ans = 0;
  }
  __pyx_L6_break:;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":497
 *             ans = False
 * 
 *         for i in range(2 * self.length):             # <<<<<<<<<<<<<<
 *             mpz_clear(x[i])
 *         sig_free(x)
 */
  __pyx_t_2 = (2 * __pyx_v_self->length);
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":498
 * 
 *         for i in range(2 * self.length):
 *             mpz_clear(x[i])             # <<<<<<<<<<<<<<
 *         sig_free(x)
 * 
 */
    mpz_clear((__pyx_v_x[__pyx_v_i]));
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":499
 *         for i in range(2 * self.length):
 *             mpz_clear(x[i])
 *         sig_free(x)             # <<<<<<<<<<<<<<
 * 
 *         return ans
 */
  sig_free(__pyx_v_x);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":501
 *         sig_free(x)
 * 
 *         return ans             # <<<<<<<<<<<<<<
 * 
 *     def children(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_10 = __Pyx_PyBool_FromLong(__pyx_v_ans); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":444
 *     # this is completely useless and checks many times the same thing
 *     # we should just look at left/right Rauzy induction
 *     def has_connection(self):             # <<<<<<<<<<<<<<
 *         r"""
 *         Check whether this family has a connection.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.has_connection", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "surface_dynamics/interval_exchanges/iet_family.pyx":503
 *         return ans
 * 
 *     def children(self):             # <<<<<<<<<<<<<<
 *         r"""
 *         Compute the children of the pair ``(p, C)`` composed of a permutation and a
 */

/* Python wrapper */
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_23children(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_22children[] = "\n        Compute the children of the pair ``(p, C)`` composed of a permutation and a\n        slice of a Rauzy simplex.\n\n        EXAMPLES::\n\n            sage: from surface_dynamics import *\n\n            sage: p = iet.Permutation('a b c d', 'd c b a')\n            sage: F = iet.IETFamily(p, Polyhedron(rays=(ZZ**4).basis())) # optional - pplpy\n            sage: top, bot = F.children() # optional - pplpy\n            sage: top[0] # optional - pplpy\n            't'\n            sage: top[1] # optional - pplpy\n            top 0 1 2 3\n            bot 3 0 2 1\n            0 0 0 1\n            0 0 1 0\n            0 1 0 0\n            1 0 0 0\n            sage: bot[0] # optional - pplpy\n            'b'\n            sage: bot[1] # optional - pplpy\n            top 0 3 1 2\n            bot 3 2 1 0\n            0 0 0 1\n            0 0 1 0\n            0 1 0 0\n            1 0 0 0\n\n            sage: p = iet.Permutation('a b c d e f', 'e b f c a d')\n            sage: F = iet.IETFamily(p, Polyhedron(rays=[(1,1,1,1,1,2),(0,1,0,1,0,1)])) # optional - pplpy\n            sage: c = F.children() # optional - pplpy\n            sage: len(c) # optional - pplpy\n            1\n            sage: c[0][0] # optional - pplpy\n            't'\n            sage: c[0][1] # optional - pplpy\n            top 0 1 2 3 4 5\n            bot 4 1 5 3 2 0\n            0 1 0 1 0 0\n            1 1 1 1 1 1\n\n            sage: F = iet.IETFamily(p, Polyhedron(rays=[(0,0,0,1,0,0),(1,0,0,0,0,0),(1,1,1,1,1,1)])) # optional - pplpy\n            sage: c = F.children() # optional - pplpy\n            sage: len(c) # optional - pplpy\n            1\n            sage: c[0][0] # optional - pplpy\n            'b'\n            sage: c[0][1] # optional - pplpy\n            top 0 1 2 3 5 4\n            bot 4 1 5 2 0 3\n            0 0 0 1 0 0\n            1 0 0 0 0 0\n            1 1 1 0 1 1\n        ";
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_23children(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("children (wrapper)", 0);
  __pyx_r = __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_22children(((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_22children(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self) {
  int __pyx_v_itop;
  int __pyx_v_ibot;
  PyObject *__pyx_v_ans = 0;
  size_t __pyx_v_d;
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_FF = 0;
  struct __pyx_obj_3ppl_10polyhedron_C_Polyhedron *__pyx_v_Ctop = 0;
  struct __pyx_obj_3ppl_10polyhedron_C_Polyhedron *__pyx_v_ineq = 0;
  struct __pyx_obj_3ppl_10polyhedron_C_Polyhedron *__pyx_v_Cbot = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  void *__pyx_t_12;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("children", 0);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":560
 *             1 1 1 0 1 1
 *         """
 *         cdef int itop = self.perm[self.dim - 1]             # <<<<<<<<<<<<<<
 *         cdef int ibot = self.perm[2*self.dim - 1]
 *         cdef list ans = []
 */
  __pyx_v_itop = (__pyx_v_self->perm[(__pyx_v_self->dim - 1)]);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":561
 *         """
 *         cdef int itop = self.perm[self.dim - 1]
 *         cdef int ibot = self.perm[2*self.dim - 1]             # <<<<<<<<<<<<<<
 *         cdef list ans = []
 *         cdef size_t d = self.poly.affine_dimension()
 */
  __pyx_v_ibot = (__pyx_v_self->perm[((2 * __pyx_v_self->dim) - 1)]);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":562
 *         cdef int itop = self.perm[self.dim - 1]
 *         cdef int ibot = self.perm[2*self.dim - 1]
 *         cdef list ans = []             # <<<<<<<<<<<<<<
 *         cdef size_t d = self.poly.affine_dimension()
 *         cdef IETFamily FF
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 562, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ans = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":563
 *         cdef int ibot = self.perm[2*self.dim - 1]
 *         cdef list ans = []
 *         cdef size_t d = self.poly.affine_dimension()             # <<<<<<<<<<<<<<
 *         cdef IETFamily FF
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->poly), __pyx_n_s_affine_dimension); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyInt_As_size_t(__pyx_t_1); if (unlikely((__pyx_t_4 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_d = __pyx_t_4;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":566
 *         cdef IETFamily FF
 * 
 *         cdef C_Polyhedron Ctop = C_Polyhedron(self.poly)             # <<<<<<<<<<<<<<
 *         cdef C_Polyhedron ineq = C_Polyhedron(Variable(itop) >= Variable(ibot))
 *         ineq.add_space_dimensions_and_embed(self.dim - max(ibot, itop) - 1)
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_3ppl_10polyhedron_C_Polyhedron), ((PyObject *)__pyx_v_self->poly)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_Ctop = ((struct __pyx_obj_3ppl_10polyhedron_C_Polyhedron *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":567
 * 
 *         cdef C_Polyhedron Ctop = C_Polyhedron(self.poly)
 *         cdef C_Polyhedron ineq = C_Polyhedron(Variable(itop) >= Variable(ibot))             # <<<<<<<<<<<<<<
 *         ineq.add_space_dimensions_and_embed(self.dim - max(ibot, itop) - 1)
 *         Ctop.intersection_assign(ineq)
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_itop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_3ppl_14linear_algebra_Variable), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_ibot); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_3ppl_14linear_algebra_Variable), __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_GE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_3ppl_10polyhedron_C_Polyhedron), __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_ineq = ((struct __pyx_obj_3ppl_10polyhedron_C_Polyhedron *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":568
 *         cdef C_Polyhedron Ctop = C_Polyhedron(self.poly)
 *         cdef C_Polyhedron ineq = C_Polyhedron(Variable(itop) >= Variable(ibot))
 *         ineq.add_space_dimensions_and_embed(self.dim - max(ibot, itop) - 1)             # <<<<<<<<<<<<<<
 *         Ctop.intersection_assign(ineq)
 *         if Ctop.affine_dimension() == d:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_ineq), __pyx_n_s_add_space_dimensions_and_embed); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __pyx_v_itop;
  __pyx_t_6 = __pyx_v_ibot;
  if (((__pyx_t_5 > __pyx_t_6) != 0)) {
    __pyx_t_7 = __pyx_t_5;
  } else {
    __pyx_t_7 = __pyx_t_6;
  }
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(((__pyx_v_self->dim - __pyx_t_7) - 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_8, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":569
 *         cdef C_Polyhedron ineq = C_Polyhedron(Variable(itop) >= Variable(ibot))
 *         ineq.add_space_dimensions_and_embed(self.dim - max(ibot, itop) - 1)
 *         Ctop.intersection_assign(ineq)             # <<<<<<<<<<<<<<
 *         if Ctop.affine_dimension() == d:
 *             Ctop.affine_image(Variable(itop), Variable(itop) - Variable(ibot))
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_Ctop), __pyx_n_s_intersection_assign); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, ((PyObject *)__pyx_v_ineq)) : __Pyx_PyObject_CallOneArg(__pyx_t_1, ((PyObject *)__pyx_v_ineq));
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":570
 *         ineq.add_space_dimensions_and_embed(self.dim - max(ibot, itop) - 1)
 *         Ctop.intersection_assign(ineq)
 *         if Ctop.affine_dimension() == d:             # <<<<<<<<<<<<<<
 *             Ctop.affine_image(Variable(itop), Variable(itop) - Variable(ibot))
 *             FF = IETFamily.__new__(IETFamily)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_Ctop), __pyx_n_s_affine_dimension); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_d); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_9) {

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":571
 *         Ctop.intersection_assign(ineq)
 *         if Ctop.affine_dimension() == d:
 *             Ctop.affine_image(Variable(itop), Variable(itop) - Variable(ibot))             # <<<<<<<<<<<<<<
 *             FF = IETFamily.__new__(IETFamily)
 *             FF.dim = self.dim
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_Ctop), __pyx_n_s_affine_image); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_itop); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_3ppl_14linear_algebra_Variable), __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_itop); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_3ppl_14linear_algebra_Variable), __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_ibot); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_11 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_3ppl_14linear_algebra_Variable), __pyx_t_3); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_Subtract(__pyx_t_10, __pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_11, __pyx_t_8, __pyx_t_3};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 571, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_11, __pyx_t_8, __pyx_t_3};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 571, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 571, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_11) {
        __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_11); __pyx_t_11 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_7, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_7, __pyx_t_3);
      __pyx_t_8 = 0;
      __pyx_t_3 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_10, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 571, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":572
 *         if Ctop.affine_dimension() == d:
 *             Ctop.affine_image(Variable(itop), Variable(itop) - Variable(ibot))
 *             FF = IETFamily.__new__(IETFamily)             # <<<<<<<<<<<<<<
 *             FF.dim = self.dim
 *             FF.perm = <int *> check_malloc(2 * self.dim * sizeof(int))
 */
    __pyx_t_2 = ((PyObject *)__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily(((PyTypeObject *)__pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 572, __pyx_L1_error)
    __Pyx_GOTREF(((PyObject *)__pyx_t_2));
    __pyx_v_FF = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":573
 *             Ctop.affine_image(Variable(itop), Variable(itop) - Variable(ibot))
 *             FF = IETFamily.__new__(IETFamily)
 *             FF.dim = self.dim             # <<<<<<<<<<<<<<
 *             FF.perm = <int *> check_malloc(2 * self.dim * sizeof(int))
 *             memcpy(FF.perm, self.perm, 2 * sizeof(int) * FF.dim)
 */
    __pyx_t_4 = __pyx_v_self->dim;
    __pyx_v_FF->dim = __pyx_t_4;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":574
 *             FF = IETFamily.__new__(IETFamily)
 *             FF.dim = self.dim
 *             FF.perm = <int *> check_malloc(2 * self.dim * sizeof(int))             # <<<<<<<<<<<<<<
 *             memcpy(FF.perm, self.perm, 2 * sizeof(int) * FF.dim)
 *             rauzy_top(FF.perm, FF.dim)
 */
    __pyx_t_12 = __pyx_f_9cysignals_6memory_check_malloc(((2 * __pyx_v_self->dim) * (sizeof(int)))); if (unlikely(__pyx_t_12 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 574, __pyx_L1_error)
    __pyx_v_FF->perm = ((int *)__pyx_t_12);

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":575
 *             FF.dim = self.dim
 *             FF.perm = <int *> check_malloc(2 * self.dim * sizeof(int))
 *             memcpy(FF.perm, self.perm, 2 * sizeof(int) * FF.dim)             # <<<<<<<<<<<<<<
 *             rauzy_top(FF.perm, FF.dim)
 *             FF.fill_rays(Ctop.thisptr.minimized_generators())
 */
    (void)(memcpy(__pyx_v_FF->perm, __pyx_v_self->perm, ((2 * (sizeof(int))) * __pyx_v_FF->dim)));

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":576
 *             FF.perm = <int *> check_malloc(2 * self.dim * sizeof(int))
 *             memcpy(FF.perm, self.perm, 2 * sizeof(int) * FF.dim)
 *             rauzy_top(FF.perm, FF.dim)             # <<<<<<<<<<<<<<
 *             FF.fill_rays(Ctop.thisptr.minimized_generators())
 *             FF.poly = Ctop
 */
    __pyx_f_16surface_dynamics_18interval_exchanges_10iet_family_rauzy_top(__pyx_v_FF->perm, __pyx_v_FF->dim);

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":577
 *             memcpy(FF.perm, self.perm, 2 * sizeof(int) * FF.dim)
 *             rauzy_top(FF.perm, FF.dim)
 *             FF.fill_rays(Ctop.thisptr.minimized_generators())             # <<<<<<<<<<<<<<
 *             FF.poly = Ctop
 *             ans.append(('t', FF))
 */
    ((struct __pyx_vtabstruct_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)__pyx_v_FF->__pyx_vtab)->fill_rays(__pyx_v_FF, __pyx_v_Ctop->__pyx_base.thisptr->minimized_generators());

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":578
 *             rauzy_top(FF.perm, FF.dim)
 *             FF.fill_rays(Ctop.thisptr.minimized_generators())
 *             FF.poly = Ctop             # <<<<<<<<<<<<<<
 *             ans.append(('t', FF))
 * 
 */
    __Pyx_INCREF(((PyObject *)__pyx_v_Ctop));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_Ctop));
    __Pyx_GOTREF(__pyx_v_FF->poly);
    __Pyx_DECREF(((PyObject *)__pyx_v_FF->poly));
    __pyx_v_FF->poly = __pyx_v_Ctop;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":579
 *             FF.fill_rays(Ctop.thisptr.minimized_generators())
 *             FF.poly = Ctop
 *             ans.append(('t', FF))             # <<<<<<<<<<<<<<
 * 
 *         cdef C_Polyhedron Cbot = C_Polyhedron(self.poly)
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_t);
    __Pyx_GIVEREF(__pyx_n_s_t);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_t);
    __Pyx_INCREF(((PyObject *)__pyx_v_FF));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_FF));
    PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_FF));
    __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_ans, __pyx_t_2); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":570
 *         ineq.add_space_dimensions_and_embed(self.dim - max(ibot, itop) - 1)
 *         Ctop.intersection_assign(ineq)
 *         if Ctop.affine_dimension() == d:             # <<<<<<<<<<<<<<
 *             Ctop.affine_image(Variable(itop), Variable(itop) - Variable(ibot))
 *             FF = IETFamily.__new__(IETFamily)
 */
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":581
 *             ans.append(('t', FF))
 * 
 *         cdef C_Polyhedron Cbot = C_Polyhedron(self.poly)             # <<<<<<<<<<<<<<
 *         ineq = C_Polyhedron(Variable(itop) <= Variable(ibot))
 *         ineq.add_space_dimensions_and_embed(self.dim - max(ibot, itop) - 1)
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_3ppl_10polyhedron_C_Polyhedron), ((PyObject *)__pyx_v_self->poly)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_Cbot = ((struct __pyx_obj_3ppl_10polyhedron_C_Polyhedron *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":582
 * 
 *         cdef C_Polyhedron Cbot = C_Polyhedron(self.poly)
 *         ineq = C_Polyhedron(Variable(itop) <= Variable(ibot))             # <<<<<<<<<<<<<<
 *         ineq.add_space_dimensions_and_embed(self.dim - max(ibot, itop) - 1)
 *         Cbot.intersection_assign(ineq)
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_itop); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_3ppl_14linear_algebra_Variable), __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_ibot); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_10 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_3ppl_14linear_algebra_Variable), __pyx_t_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_10, Py_LE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_3ppl_10polyhedron_C_Polyhedron), __pyx_t_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_ineq, ((struct __pyx_obj_3ppl_10polyhedron_C_Polyhedron *)__pyx_t_10));
  __pyx_t_10 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":583
 *         cdef C_Polyhedron Cbot = C_Polyhedron(self.poly)
 *         ineq = C_Polyhedron(Variable(itop) <= Variable(ibot))
 *         ineq.add_space_dimensions_and_embed(self.dim - max(ibot, itop) - 1)             # <<<<<<<<<<<<<<
 *         Cbot.intersection_assign(ineq)
 *         if Cbot.affine_dimension() == d:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_ineq), __pyx_n_s_add_space_dimensions_and_embed); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __pyx_v_itop;
  __pyx_t_5 = __pyx_v_ibot;
  if (((__pyx_t_7 > __pyx_t_5) != 0)) {
    __pyx_t_6 = __pyx_t_7;
  } else {
    __pyx_t_6 = __pyx_t_5;
  }
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(((__pyx_v_self->dim - __pyx_t_6) - 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_10 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":584
 *         ineq = C_Polyhedron(Variable(itop) <= Variable(ibot))
 *         ineq.add_space_dimensions_and_embed(self.dim - max(ibot, itop) - 1)
 *         Cbot.intersection_assign(ineq)             # <<<<<<<<<<<<<<
 *         if Cbot.affine_dimension() == d:
 *             Cbot.affine_image(Variable(ibot), Variable(ibot) - Variable(itop))
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_Cbot), __pyx_n_s_intersection_assign); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_10 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, ((PyObject *)__pyx_v_ineq)) : __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)__pyx_v_ineq));
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":585
 *         ineq.add_space_dimensions_and_embed(self.dim - max(ibot, itop) - 1)
 *         Cbot.intersection_assign(ineq)
 *         if Cbot.affine_dimension() == d:             # <<<<<<<<<<<<<<
 *             Cbot.affine_image(Variable(ibot), Variable(ibot) - Variable(itop))
 *             FF = IETFamily.__new__(IETFamily)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_Cbot), __pyx_n_s_affine_dimension); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_10 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_d); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_10, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 585, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 585, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_9) {

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":586
 *         Cbot.intersection_assign(ineq)
 *         if Cbot.affine_dimension() == d:
 *             Cbot.affine_image(Variable(ibot), Variable(ibot) - Variable(itop))             # <<<<<<<<<<<<<<
 *             FF = IETFamily.__new__(IETFamily)
 *             FF.dim = self.dim
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_Cbot), __pyx_n_s_affine_image); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_ibot); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_3ppl_14linear_algebra_Variable), __pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_ibot); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_8 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_3ppl_14linear_algebra_Variable), __pyx_t_10); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_itop); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_3ppl_14linear_algebra_Variable), __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = PyNumber_Subtract(__pyx_t_8, __pyx_t_11); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_11, __pyx_t_3, __pyx_t_10};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 586, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[3] = {__pyx_t_11, __pyx_t_3, __pyx_t_10};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 586, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 586, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_11) {
        __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_11); __pyx_t_11 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_6, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_6, __pyx_t_10);
      __pyx_t_3 = 0;
      __pyx_t_10 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 586, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":587
 *         if Cbot.affine_dimension() == d:
 *             Cbot.affine_image(Variable(ibot), Variable(ibot) - Variable(itop))
 *             FF = IETFamily.__new__(IETFamily)             # <<<<<<<<<<<<<<
 *             FF.dim = self.dim
 *             FF.perm = <int *> check_malloc(2 * self.dim * sizeof(int))
 */
    __pyx_t_1 = ((PyObject *)__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily(((PyTypeObject *)__pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 587, __pyx_L1_error)
    __Pyx_GOTREF(((PyObject *)__pyx_t_1));
    __Pyx_XDECREF_SET(__pyx_v_FF, ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":588
 *             Cbot.affine_image(Variable(ibot), Variable(ibot) - Variable(itop))
 *             FF = IETFamily.__new__(IETFamily)
 *             FF.dim = self.dim             # <<<<<<<<<<<<<<
 *             FF.perm = <int *> check_malloc(2 * self.dim * sizeof(int))
 *             memcpy(FF.perm, self.perm, 2 * sizeof(int) * FF.dim)
 */
    __pyx_t_4 = __pyx_v_self->dim;
    __pyx_v_FF->dim = __pyx_t_4;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":589
 *             FF = IETFamily.__new__(IETFamily)
 *             FF.dim = self.dim
 *             FF.perm = <int *> check_malloc(2 * self.dim * sizeof(int))             # <<<<<<<<<<<<<<
 *             memcpy(FF.perm, self.perm, 2 * sizeof(int) * FF.dim)
 *             rauzy_bot(FF.perm, FF.dim)
 */
    __pyx_t_12 = __pyx_f_9cysignals_6memory_check_malloc(((2 * __pyx_v_self->dim) * (sizeof(int)))); if (unlikely(__pyx_t_12 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 589, __pyx_L1_error)
    __pyx_v_FF->perm = ((int *)__pyx_t_12);

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":590
 *             FF.dim = self.dim
 *             FF.perm = <int *> check_malloc(2 * self.dim * sizeof(int))
 *             memcpy(FF.perm, self.perm, 2 * sizeof(int) * FF.dim)             # <<<<<<<<<<<<<<
 *             rauzy_bot(FF.perm, FF.dim)
 *             FF.fill_rays(Cbot.thisptr.minimized_generators())
 */
    (void)(memcpy(__pyx_v_FF->perm, __pyx_v_self->perm, ((2 * (sizeof(int))) * __pyx_v_FF->dim)));

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":591
 *             FF.perm = <int *> check_malloc(2 * self.dim * sizeof(int))
 *             memcpy(FF.perm, self.perm, 2 * sizeof(int) * FF.dim)
 *             rauzy_bot(FF.perm, FF.dim)             # <<<<<<<<<<<<<<
 *             FF.fill_rays(Cbot.thisptr.minimized_generators())
 *             FF.poly = Cbot
 */
    __pyx_f_16surface_dynamics_18interval_exchanges_10iet_family_rauzy_bot(__pyx_v_FF->perm, __pyx_v_FF->dim);

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":592
 *             memcpy(FF.perm, self.perm, 2 * sizeof(int) * FF.dim)
 *             rauzy_bot(FF.perm, FF.dim)
 *             FF.fill_rays(Cbot.thisptr.minimized_generators())             # <<<<<<<<<<<<<<
 *             FF.poly = Cbot
 *             ans.append(('b', FF))
 */
    ((struct __pyx_vtabstruct_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)__pyx_v_FF->__pyx_vtab)->fill_rays(__pyx_v_FF, __pyx_v_Cbot->__pyx_base.thisptr->minimized_generators());

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":593
 *             rauzy_bot(FF.perm, FF.dim)
 *             FF.fill_rays(Cbot.thisptr.minimized_generators())
 *             FF.poly = Cbot             # <<<<<<<<<<<<<<
 *             ans.append(('b', FF))
 * 
 */
    __Pyx_INCREF(((PyObject *)__pyx_v_Cbot));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_Cbot));
    __Pyx_GOTREF(__pyx_v_FF->poly);
    __Pyx_DECREF(((PyObject *)__pyx_v_FF->poly));
    __pyx_v_FF->poly = __pyx_v_Cbot;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":594
 *             FF.fill_rays(Cbot.thisptr.minimized_generators())
 *             FF.poly = Cbot
 *             ans.append(('b', FF))             # <<<<<<<<<<<<<<
 * 
 *         return ans
 */
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 594, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_b);
    __Pyx_GIVEREF(__pyx_n_s_b);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_b);
    __Pyx_INCREF(((PyObject *)__pyx_v_FF));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_FF));
    PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_FF));
    __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_ans, __pyx_t_1); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 594, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":585
 *         ineq.add_space_dimensions_and_embed(self.dim - max(ibot, itop) - 1)
 *         Cbot.intersection_assign(ineq)
 *         if Cbot.affine_dimension() == d:             # <<<<<<<<<<<<<<
 *             Cbot.affine_image(Variable(ibot), Variable(ibot) - Variable(itop))
 *             FF = IETFamily.__new__(IETFamily)
 */
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":596
 *             ans.append(('b', FF))
 * 
 *         return ans             # <<<<<<<<<<<<<<
 * 
 *     def tree(self, max_depth=5, verbose=False):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ans);
  __pyx_r = __pyx_v_ans;
  goto __pyx_L0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":503
 *         return ans
 * 
 *     def children(self):             # <<<<<<<<<<<<<<
 *         r"""
 *         Compute the children of the pair ``(p, C)`` composed of a permutation and a
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.children", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ans);
  __Pyx_XDECREF((PyObject *)__pyx_v_FF);
  __Pyx_XDECREF((PyObject *)__pyx_v_Ctop);
  __Pyx_XDECREF((PyObject *)__pyx_v_ineq);
  __Pyx_XDECREF((PyObject *)__pyx_v_Cbot);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_26generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "surface_dynamics/interval_exchanges/iet_family.pyx":598
 *         return ans
 * 
 *     def tree(self, max_depth=5, verbose=False):             # <<<<<<<<<<<<<<
 *         r"""
 *         The tree should be smarter in several ways:
 */

/* Python wrapper */
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_25tree(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_24tree[] = "\n        The tree should be smarter in several ways:\n\n        - once a saddle connection is found it should check whether it is stable and\n          whether it corresponds to a splitting, a cylinder decomposition, etc\n\n        - implement a C version!\n\n        - understand what the parabolic are doing!\n\n        EXAMPLES::\n\n            sage: from surface_dynamics import *\n            sage: p = iet.Permutation([0,1,2,3,4,5],[5,4,3,2,1,0])\n            sage: rays = [[5, 1, 0, 0, 3, 8], [2, 1, 0, 3, 0, 5], [1, 0, 1, 2, 0, 3], [3, 0, 1, 0, 2, 5]]\n            sage: F = iet.IETFamily(p, rays) # optional - pplpy\n        ";
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_25tree(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_max_depth = 0;
  PyObject *__pyx_v_verbose = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("tree (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_max_depth,&__pyx_n_s_verbose,0};
    PyObject* values[2] = {0,0};
    values[0] = ((PyObject *)__pyx_int_5);
    values[1] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_depth);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_verbose);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "tree") < 0)) __PYX_ERR(0, 598, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_max_depth = values[0];
    __pyx_v_verbose = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("tree", 0, 0, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 598, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.tree", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_24tree(((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)__pyx_v_self), __pyx_v_max_depth, __pyx_v_verbose);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_24tree(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self, PyObject *__pyx_v_max_depth, PyObject *__pyx_v_verbose) {
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("tree", 0);
  __pyx_cur_scope = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree *)__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree(__pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 598, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_max_depth = __pyx_v_max_depth;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_max_depth);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_max_depth);
  __pyx_cur_scope->__pyx_v_verbose = __pyx_v_verbose;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_verbose);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_verbose);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_26generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_tree, __pyx_n_s_IETFamily_tree, __pyx_n_s_surface_dynamics_interval_exchan_2); if (unlikely(!gen)) __PYX_ERR(0, 598, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.tree", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_26generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree *__pyx_cur_scope = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("tree", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L19_resume_from_yield;
    case 2: goto __pyx_L20_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 598, __pyx_L1_error)

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":616
 *             sage: F = iet.IETFamily(p, rays) # optional - pplpy
 *         """
 *         s = ''             # <<<<<<<<<<<<<<
 *         f = self
 *         branch = [[(s, f)]]
 */
  __Pyx_INCREF(__pyx_kp_s__8);
  __Pyx_GIVEREF(__pyx_kp_s__8);
  __pyx_cur_scope->__pyx_v_s = __pyx_kp_s__8;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":617
 *         """
 *         s = ''
 *         f = self             # <<<<<<<<<<<<<<
 *         branch = [[(s, f)]]
 *         seen = set([f])
 */
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_cur_scope->__pyx_v_self));
  __pyx_cur_scope->__pyx_v_f = ((PyObject *)__pyx_cur_scope->__pyx_v_self);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":618
 *         s = ''
 *         f = self
 *         branch = [[(s, f)]]             # <<<<<<<<<<<<<<
 *         seen = set([f])
 *         while True:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_s);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_s);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_cur_scope->__pyx_v_s);
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_f);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_f);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_cur_scope->__pyx_v_f);
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_branch = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":619
 *         f = self
 *         branch = [[(s, f)]]
 *         seen = set([f])             # <<<<<<<<<<<<<<
 *         while True:
 *             if verbose:
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PySet_Add(__pyx_t_1, __pyx_cur_scope->__pyx_v_f) < 0) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_seen = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":620
 *         branch = [[(s, f)]]
 *         seen = set([f])
 *         while True:             # <<<<<<<<<<<<<<
 *             if verbose:
 *                 print("branch:")
 */
  while (1) {

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":621
 *         seen = set([f])
 *         while True:
 *             if verbose:             # <<<<<<<<<<<<<<
 *                 print("branch:")
 *                 for ss,ff in branch[-1]:
 */
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_verbose); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 621, __pyx_L1_error)
    if (__pyx_t_3) {

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":622
 *         while True:
 *             if verbose:
 *                 print("branch:")             # <<<<<<<<<<<<<<
 *                 for ss,ff in branch[-1]:
 *                     print(ss)
 */
      if (__Pyx_PrintOne(0, __pyx_kp_s_branch) < 0) __PYX_ERR(0, 622, __pyx_L1_error)

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":623
 *             if verbose:
 *                 print("branch:")
 *                 for ss,ff in branch[-1]:             # <<<<<<<<<<<<<<
 *                     print(ss)
 *                     print(ff)
 */
      __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_cur_scope->__pyx_v_branch, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 623, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
        __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
        __pyx_t_5 = NULL;
      } else {
        __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 623, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 623, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      for (;;) {
        if (likely(!__pyx_t_5)) {
          if (likely(PyList_CheckExact(__pyx_t_2))) {
            if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 623, __pyx_L1_error)
            #else
            __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 623, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            #endif
          } else {
            if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 623, __pyx_L1_error)
            #else
            __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 623, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            #endif
          }
        } else {
          __pyx_t_1 = __pyx_t_5(__pyx_t_2);
          if (unlikely(!__pyx_t_1)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 623, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_1);
        }
        if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
          PyObject* sequence = __pyx_t_1;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 623, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_7 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          #else
          __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 623, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_7 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 623, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          #endif
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_8 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 623, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_9 = Py_TYPE(__pyx_t_8)->tp_iternext;
          index = 0; __pyx_t_6 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_6)) goto __pyx_L9_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_6);
          index = 1; __pyx_t_7 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_7)) goto __pyx_L9_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_7);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_8), 2) < 0) __PYX_ERR(0, 623, __pyx_L1_error)
          __pyx_t_9 = NULL;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          goto __pyx_L10_unpacking_done;
          __pyx_L9_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_9 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 623, __pyx_L1_error)
          __pyx_L10_unpacking_done:;
        }
        __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_ss);
        __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_ss, __pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_6);
        __pyx_t_6 = 0;
        __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_ff);
        __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_ff, __pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_7);
        __pyx_t_7 = 0;

        /* "surface_dynamics/interval_exchanges/iet_family.pyx":624
 *                 print("branch:")
 *                 for ss,ff in branch[-1]:
 *                     print(ss)             # <<<<<<<<<<<<<<
 *                     print(ff)
 *                     print()
 */
        if (__Pyx_PrintOne(0, __pyx_cur_scope->__pyx_v_ss) < 0) __PYX_ERR(0, 624, __pyx_L1_error)

        /* "surface_dynamics/interval_exchanges/iet_family.pyx":625
 *                 for ss,ff in branch[-1]:
 *                     print(ss)
 *                     print(ff)             # <<<<<<<<<<<<<<
 *                     print()
 *             while len(branch) < max_depth:
 */
        if (__Pyx_PrintOne(0, __pyx_cur_scope->__pyx_v_ff) < 0) __PYX_ERR(0, 625, __pyx_L1_error)

        /* "surface_dynamics/interval_exchanges/iet_family.pyx":626
 *                     print(ss)
 *                     print(ff)
 *                     print()             # <<<<<<<<<<<<<<
 *             while len(branch) < max_depth:
 *                 branch.append([])
 */
        if (__Pyx_PrintOne(0, __pyx_empty_tuple) < 0) __PYX_ERR(0, 626, __pyx_L1_error)

        /* "surface_dynamics/interval_exchanges/iet_family.pyx":623
 *             if verbose:
 *                 print("branch:")
 *                 for ss,ff in branch[-1]:             # <<<<<<<<<<<<<<
 *                     print(ss)
 *                     print(ff)
 */
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":621
 *         seen = set([f])
 *         while True:
 *             if verbose:             # <<<<<<<<<<<<<<
 *                 print("branch:")
 *                 for ss,ff in branch[-1]:
 */
    }

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":627
 *                     print(ff)
 *                     print()
 *             while len(branch) < max_depth:             # <<<<<<<<<<<<<<
 *                 branch.append([])
 *                 for ss, ff in f.children():
 */
    while (1) {
      __pyx_t_4 = PyList_GET_SIZE(__pyx_cur_scope->__pyx_v_branch); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 627, __pyx_L1_error)
      __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 627, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = PyObject_RichCompare(__pyx_t_2, __pyx_cur_scope->__pyx_v_max_depth, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 627, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 627, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (!__pyx_t_3) break;

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":628
 *                     print()
 *             while len(branch) < max_depth:
 *                 branch.append([])             # <<<<<<<<<<<<<<
 *                 for ss, ff in f.children():
 *                     if verbose:
 */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 628, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_cur_scope->__pyx_v_branch, __pyx_t_1); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 628, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":629
 *             while len(branch) < max_depth:
 *                 branch.append([])
 *                 for ss, ff in f.children():             # <<<<<<<<<<<<<<
 *                     if verbose:
 *                         print("looking children", ss)
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_f, __pyx_n_s_children); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 629, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 629, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
        __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
        __pyx_t_5 = NULL;
      } else {
        __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 629, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 629, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      for (;;) {
        if (likely(!__pyx_t_5)) {
          if (likely(PyList_CheckExact(__pyx_t_2))) {
            if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 629, __pyx_L1_error)
            #else
            __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 629, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            #endif
          } else {
            if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 629, __pyx_L1_error)
            #else
            __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 629, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            #endif
          }
        } else {
          __pyx_t_1 = __pyx_t_5(__pyx_t_2);
          if (unlikely(!__pyx_t_1)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 629, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_1);
        }
        if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
          PyObject* sequence = __pyx_t_1;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 629, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_6);
          #else
          __pyx_t_7 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 629, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 629, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_8 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 629, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_9 = Py_TYPE(__pyx_t_8)->tp_iternext;
          index = 0; __pyx_t_7 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_7)) goto __pyx_L15_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_7);
          index = 1; __pyx_t_6 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_6)) goto __pyx_L15_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_6);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_8), 2) < 0) __PYX_ERR(0, 629, __pyx_L1_error)
          __pyx_t_9 = NULL;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          goto __pyx_L16_unpacking_done;
          __pyx_L15_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_9 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 629, __pyx_L1_error)
          __pyx_L16_unpacking_done:;
        }
        __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_ss);
        __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_ss, __pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_7);
        __pyx_t_7 = 0;
        __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_ff);
        __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_ff, __pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_6);
        __pyx_t_6 = 0;

        /* "surface_dynamics/interval_exchanges/iet_family.pyx":630
 *                 branch.append([])
 *                 for ss, ff in f.children():
 *                     if verbose:             # <<<<<<<<<<<<<<
 *                         print("looking children", ss)
 *                         print("cone", ff)
 */
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_cur_scope->__pyx_v_verbose); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 630, __pyx_L1_error)
        if (__pyx_t_3) {

          /* "surface_dynamics/interval_exchanges/iet_family.pyx":631
 *                 for ss, ff in f.children():
 *                     if verbose:
 *                         print("looking children", ss)             # <<<<<<<<<<<<<<
 *                         print("cone", ff)
 *                     # check for saddles among the children
 */
          __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 631, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_kp_s_looking_children);
          __Pyx_GIVEREF(__pyx_kp_s_looking_children);
          PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_s_looking_children);
          __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ss);
          __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ss);
          PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_cur_scope->__pyx_v_ss);
          if (__Pyx_PrintOne(0, __pyx_t_1) < 0) __PYX_ERR(0, 631, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "surface_dynamics/interval_exchanges/iet_family.pyx":632
 *                     if verbose:
 *                         print("looking children", ss)
 *                         print("cone", ff)             # <<<<<<<<<<<<<<
 *                     # check for saddles among the children
 *                     if ff.has_connection():
 */
          __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 632, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_n_s_cone);
          __Pyx_GIVEREF(__pyx_n_s_cone);
          PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_cone);
          __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ff);
          __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ff);
          PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_cur_scope->__pyx_v_ff);
          if (__Pyx_PrintOne(0, __pyx_t_1) < 0) __PYX_ERR(0, 632, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "surface_dynamics/interval_exchanges/iet_family.pyx":630
 *                 branch.append([])
 *                 for ss, ff in f.children():
 *                     if verbose:             # <<<<<<<<<<<<<<
 *                         print("looking children", ss)
 *                         print("cone", ff)
 */
        }

        /* "surface_dynamics/interval_exchanges/iet_family.pyx":634
 *                         print("cone", ff)
 *                     # check for saddles among the children
 *                     if ff.has_connection():             # <<<<<<<<<<<<<<
 *                         yield 'saddle', s+ss, ff
 * 
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_ff, __pyx_n_s_has_connection); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 634, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_6, function);
          }
        }
        __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 634, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 634, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (__pyx_t_3) {

          /* "surface_dynamics/interval_exchanges/iet_family.pyx":635
 *                     # check for saddles among the children
 *                     if ff.has_connection():
 *                         yield 'saddle', s+ss, ff             # <<<<<<<<<<<<<<
 * 
 *                     # check for auto-simlarity
 */
          __pyx_t_1 = PyNumber_Add(__pyx_cur_scope->__pyx_v_s, __pyx_cur_scope->__pyx_v_ss); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 635, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_INCREF(__pyx_n_s_saddle);
          __Pyx_GIVEREF(__pyx_n_s_saddle);
          PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_n_s_saddle);
          __Pyx_GIVEREF(__pyx_t_1);
          PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
          __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ff);
          __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ff);
          PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_cur_scope->__pyx_v_ff);
          __pyx_t_1 = 0;
          __pyx_r = __pyx_t_6;
          __pyx_t_6 = 0;
          __Pyx_XGIVEREF(__pyx_t_2);
          __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
          __pyx_cur_scope->__pyx_t_1 = __pyx_t_4;
          __pyx_cur_scope->__pyx_t_2 = __pyx_t_5;
          __Pyx_XGIVEREF(__pyx_r);
          __Pyx_RefNannyFinishContext();
          __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
          /* return from generator, yielding value */
          __pyx_generator->resume_label = 1;
          return __pyx_r;
          __pyx_L19_resume_from_yield:;
          __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
          __pyx_cur_scope->__pyx_t_0 = 0;
          __Pyx_XGOTREF(__pyx_t_2);
          __pyx_t_4 = __pyx_cur_scope->__pyx_t_1;
          __pyx_t_5 = __pyx_cur_scope->__pyx_t_2;
          if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 635, __pyx_L1_error)

          /* "surface_dynamics/interval_exchanges/iet_family.pyx":634
 *                         print("cone", ff)
 *                     # check for saddles among the children
 *                     if ff.has_connection():             # <<<<<<<<<<<<<<
 *                         yield 'saddle', s+ss, ff
 * 
 */
          goto __pyx_L18;
        }

        /* "surface_dynamics/interval_exchanges/iet_family.pyx":638
 * 
 *                     # check for auto-simlarity
 *                     elif ff in seen:             # <<<<<<<<<<<<<<
 *                         yield 'autosim', s+ss, ff
 * 
 */
        __pyx_t_3 = (__Pyx_PySet_ContainsTF(__pyx_cur_scope->__pyx_v_ff, __pyx_cur_scope->__pyx_v_seen, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 638, __pyx_L1_error)
        __pyx_t_11 = (__pyx_t_3 != 0);
        if (__pyx_t_11) {

          /* "surface_dynamics/interval_exchanges/iet_family.pyx":639
 *                     # check for auto-simlarity
 *                     elif ff in seen:
 *                         yield 'autosim', s+ss, ff             # <<<<<<<<<<<<<<
 * 
 *                     else:
 */
          __pyx_t_6 = PyNumber_Add(__pyx_cur_scope->__pyx_v_s, __pyx_cur_scope->__pyx_v_ss); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 639, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 639, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_n_s_autosim);
          __Pyx_GIVEREF(__pyx_n_s_autosim);
          PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_autosim);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_6);
          __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ff);
          __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ff);
          PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_cur_scope->__pyx_v_ff);
          __pyx_t_6 = 0;
          __pyx_r = __pyx_t_1;
          __pyx_t_1 = 0;
          __Pyx_XGIVEREF(__pyx_t_2);
          __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
          __pyx_cur_scope->__pyx_t_1 = __pyx_t_4;
          __pyx_cur_scope->__pyx_t_2 = __pyx_t_5;
          __Pyx_XGIVEREF(__pyx_r);
          __Pyx_RefNannyFinishContext();
          __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
          /* return from generator, yielding value */
          __pyx_generator->resume_label = 2;
          return __pyx_r;
          __pyx_L20_resume_from_yield:;
          __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
          __pyx_cur_scope->__pyx_t_0 = 0;
          __Pyx_XGOTREF(__pyx_t_2);
          __pyx_t_4 = __pyx_cur_scope->__pyx_t_1;
          __pyx_t_5 = __pyx_cur_scope->__pyx_t_2;
          if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 639, __pyx_L1_error)

          /* "surface_dynamics/interval_exchanges/iet_family.pyx":638
 * 
 *                     # check for auto-simlarity
 *                     elif ff in seen:             # <<<<<<<<<<<<<<
 *                         yield 'autosim', s+ss, ff
 * 
 */
          goto __pyx_L18;
        }

        /* "surface_dynamics/interval_exchanges/iet_family.pyx":642
 * 
 *                     else:
 *                         branch[-1].append((s+ss, ff))             # <<<<<<<<<<<<<<
 * 
 *                 if not branch[-1]:
 */
        /*else*/ {
          __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_cur_scope->__pyx_v_branch, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 642, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_6 = PyNumber_Add(__pyx_cur_scope->__pyx_v_s, __pyx_cur_scope->__pyx_v_ss); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 642, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 642, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6);
          __Pyx_INCREF(__pyx_cur_scope->__pyx_v_ff);
          __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_ff);
          PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_cur_scope->__pyx_v_ff);
          __pyx_t_6 = 0;
          __pyx_t_10 = __Pyx_PyObject_Append(__pyx_t_1, __pyx_t_7); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 642, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __pyx_L18:;

        /* "surface_dynamics/interval_exchanges/iet_family.pyx":629
 *             while len(branch) < max_depth:
 *                 branch.append([])
 *                 for ss, ff in f.children():             # <<<<<<<<<<<<<<
 *                     if verbose:
 *                         print("looking children", ss)
 */
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":644
 *                         branch[-1].append((s+ss, ff))
 * 
 *                 if not branch[-1]:             # <<<<<<<<<<<<<<
 *                     branch.pop(-1)
 *                     break
 */
      __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_cur_scope->__pyx_v_branch, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 644, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 644, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_3 = ((!__pyx_t_11) != 0);
      if (__pyx_t_3) {

        /* "surface_dynamics/interval_exchanges/iet_family.pyx":645
 * 
 *                 if not branch[-1]:
 *                     branch.pop(-1)             # <<<<<<<<<<<<<<
 *                     break
 *                 else:
 */
        __pyx_t_2 = __Pyx_PyList_PopIndex(__pyx_cur_scope->__pyx_v_branch, __pyx_int_neg_1, -1L, 1, Py_ssize_t, PyInt_FromSsize_t); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 645, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "surface_dynamics/interval_exchanges/iet_family.pyx":646
 *                 if not branch[-1]:
 *                     branch.pop(-1)
 *                     break             # <<<<<<<<<<<<<<
 *                 else:
 *                     s, f = branch[-1][-1]
 */
        goto __pyx_L12_break;

        /* "surface_dynamics/interval_exchanges/iet_family.pyx":644
 *                         branch[-1].append((s+ss, ff))
 * 
 *                 if not branch[-1]:             # <<<<<<<<<<<<<<
 *                     branch.pop(-1)
 *                     break
 */
      }

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":648
 *                     break
 *                 else:
 *                     s, f = branch[-1][-1]             # <<<<<<<<<<<<<<
 *                     assert f not in seen
 *                     seen.add(f)
 */
      /*else*/ {
        __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_cur_scope->__pyx_v_branch, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 648, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_2, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 648, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
          PyObject* sequence = __pyx_t_7;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 648, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_1);
          #else
          __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 648, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 648, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_6 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 648, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_9 = Py_TYPE(__pyx_t_6)->tp_iternext;
          index = 0; __pyx_t_2 = __pyx_t_9(__pyx_t_6); if (unlikely(!__pyx_t_2)) goto __pyx_L22_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_2);
          index = 1; __pyx_t_1 = __pyx_t_9(__pyx_t_6); if (unlikely(!__pyx_t_1)) goto __pyx_L22_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_1);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_6), 2) < 0) __PYX_ERR(0, 648, __pyx_L1_error)
          __pyx_t_9 = NULL;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          goto __pyx_L23_unpacking_done;
          __pyx_L22_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_9 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 648, __pyx_L1_error)
          __pyx_L23_unpacking_done:;
        }
        __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_s);
        __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_s, __pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_f);
        __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_f, __pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_1);
        __pyx_t_1 = 0;

        /* "surface_dynamics/interval_exchanges/iet_family.pyx":649
 *                 else:
 *                     s, f = branch[-1][-1]
 *                     assert f not in seen             # <<<<<<<<<<<<<<
 *                     seen.add(f)
 * 
 */
        #ifndef CYTHON_WITHOUT_ASSERTIONS
        if (unlikely(!Py_OptimizeFlag)) {
          __pyx_t_3 = (__Pyx_PySet_ContainsTF(__pyx_cur_scope->__pyx_v_f, __pyx_cur_scope->__pyx_v_seen, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 649, __pyx_L1_error)
          if (unlikely(!(__pyx_t_3 != 0))) {
            PyErr_SetNone(PyExc_AssertionError);
            __PYX_ERR(0, 649, __pyx_L1_error)
          }
        }
        #endif

        /* "surface_dynamics/interval_exchanges/iet_family.pyx":650
 *                     s, f = branch[-1][-1]
 *                     assert f not in seen
 *                     seen.add(f)             # <<<<<<<<<<<<<<
 * 
 *             # backtrack
 */
        __pyx_t_10 = PySet_Add(__pyx_cur_scope->__pyx_v_seen, __pyx_cur_scope->__pyx_v_f); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 650, __pyx_L1_error)
      }
    }
    __pyx_L12_break:;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":653
 * 
 *             # backtrack
 *             while branch and len(branch[-1]) == 1:             # <<<<<<<<<<<<<<
 *                 s, f = branch.pop()[0]
 *                 if f not in seen:
 */
    while (1) {
      __pyx_t_11 = (PyList_GET_SIZE(__pyx_cur_scope->__pyx_v_branch) != 0);
      if (__pyx_t_11) {
      } else {
        __pyx_t_3 = __pyx_t_11;
        goto __pyx_L26_bool_binop_done;
      }
      __pyx_t_7 = __Pyx_GetItemInt_List(__pyx_cur_scope->__pyx_v_branch, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 653, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_4 = PyObject_Length(__pyx_t_7); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 653, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_11 = ((__pyx_t_4 == 1) != 0);
      __pyx_t_3 = __pyx_t_11;
      __pyx_L26_bool_binop_done:;
      if (!__pyx_t_3) break;

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":654
 *             # backtrack
 *             while branch and len(branch[-1]) == 1:
 *                 s, f = branch.pop()[0]             # <<<<<<<<<<<<<<
 *                 if f not in seen:
 *                     raise RuntimeError("s = {}\nf =\n{}".format(s, f))
 */
      __pyx_t_7 = __Pyx_PyList_Pop(__pyx_cur_scope->__pyx_v_branch); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 654, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_7, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 654, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
        PyObject* sequence = __pyx_t_1;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 654, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_2);
        #else
        __pyx_t_7 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 654, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 654, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_6 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 654, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_9 = Py_TYPE(__pyx_t_6)->tp_iternext;
        index = 0; __pyx_t_7 = __pyx_t_9(__pyx_t_6); if (unlikely(!__pyx_t_7)) goto __pyx_L28_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_7);
        index = 1; __pyx_t_2 = __pyx_t_9(__pyx_t_6); if (unlikely(!__pyx_t_2)) goto __pyx_L28_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_2);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_6), 2) < 0) __PYX_ERR(0, 654, __pyx_L1_error)
        __pyx_t_9 = NULL;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L29_unpacking_done;
        __pyx_L28_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_9 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 654, __pyx_L1_error)
        __pyx_L29_unpacking_done:;
      }
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_s);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_s, __pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_f);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_f, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      __pyx_t_2 = 0;

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":655
 *             while branch and len(branch[-1]) == 1:
 *                 s, f = branch.pop()[0]
 *                 if f not in seen:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError("s = {}\nf =\n{}".format(s, f))
 *                 seen.remove(f)
 */
      __pyx_t_3 = (__Pyx_PySet_ContainsTF(__pyx_cur_scope->__pyx_v_f, __pyx_cur_scope->__pyx_v_seen, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 655, __pyx_L1_error)
      __pyx_t_11 = (__pyx_t_3 != 0);
      if (unlikely(__pyx_t_11)) {

        /* "surface_dynamics/interval_exchanges/iet_family.pyx":656
 *                 s, f = branch.pop()[0]
 *                 if f not in seen:
 *                     raise RuntimeError("s = {}\nf =\n{}".format(s, f))             # <<<<<<<<<<<<<<
 *                 seen.remove(f)
 * 
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_s_f, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 656, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = NULL;
        __pyx_t_12 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_12 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_cur_scope->__pyx_v_s, __pyx_cur_scope->__pyx_v_f};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 656, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
          PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_cur_scope->__pyx_v_s, __pyx_cur_scope->__pyx_v_f};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 656, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(2+__pyx_t_12); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 656, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_7) {
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7); __pyx_t_7 = NULL;
          }
          __Pyx_INCREF(__pyx_cur_scope->__pyx_v_s);
          __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_s);
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_12, __pyx_cur_scope->__pyx_v_s);
          __Pyx_INCREF(__pyx_cur_scope->__pyx_v_f);
          __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_f);
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_12, __pyx_cur_scope->__pyx_v_f);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 656, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 656, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_Raise(__pyx_t_2, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __PYX_ERR(0, 656, __pyx_L1_error)

        /* "surface_dynamics/interval_exchanges/iet_family.pyx":655
 *             while branch and len(branch[-1]) == 1:
 *                 s, f = branch.pop()[0]
 *                 if f not in seen:             # <<<<<<<<<<<<<<
 *                     raise RuntimeError("s = {}\nf =\n{}".format(s, f))
 *                 seen.remove(f)
 */
      }

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":657
 *                 if f not in seen:
 *                     raise RuntimeError("s = {}\nf =\n{}".format(s, f))
 *                 seen.remove(f)             # <<<<<<<<<<<<<<
 * 
 *             if not branch:
 */
      __pyx_t_10 = __Pyx_PySet_Remove(__pyx_cur_scope->__pyx_v_seen, __pyx_cur_scope->__pyx_v_f); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 657, __pyx_L1_error)
    }

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":659
 *                 seen.remove(f)
 * 
 *             if not branch:             # <<<<<<<<<<<<<<
 *                 return
 * 
 */
    __pyx_t_11 = (PyList_GET_SIZE(__pyx_cur_scope->__pyx_v_branch) != 0);
    __pyx_t_3 = ((!__pyx_t_11) != 0);
    if (__pyx_t_3) {

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":660
 * 
 *             if not branch:
 *                 return             # <<<<<<<<<<<<<<
 * 
 *             s, f = branch[-1].pop(-1)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = NULL;
      goto __pyx_L0;

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":659
 *                 seen.remove(f)
 * 
 *             if not branch:             # <<<<<<<<<<<<<<
 *                 return
 * 
 */
    }

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":662
 *                 return
 * 
 *             s, f = branch[-1].pop(-1)             # <<<<<<<<<<<<<<
 *             seen.remove(f)
 *             s, f = branch[-1][-1]
 */
    __pyx_t_2 = __Pyx_GetItemInt_List(__pyx_cur_scope->__pyx_v_branch, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 662, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_PopIndex(__pyx_t_2, __pyx_int_neg_1, -1L, 1, Py_ssize_t, PyInt_FromSsize_t); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 662, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 662, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 662, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 662, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 662, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_9 = Py_TYPE(__pyx_t_7)->tp_iternext;
      index = 0; __pyx_t_2 = __pyx_t_9(__pyx_t_7); if (unlikely(!__pyx_t_2)) goto __pyx_L32_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      index = 1; __pyx_t_6 = __pyx_t_9(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L32_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_7), 2) < 0) __PYX_ERR(0, 662, __pyx_L1_error)
      __pyx_t_9 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L33_unpacking_done;
      __pyx_L32_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_9 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 662, __pyx_L1_error)
      __pyx_L33_unpacking_done:;
    }
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_s);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_s, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_f);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_f, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __pyx_t_6 = 0;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":663
 * 
 *             s, f = branch[-1].pop(-1)
 *             seen.remove(f)             # <<<<<<<<<<<<<<
 *             s, f = branch[-1][-1]
 *             seen.add(f)
 */
    __pyx_t_10 = __Pyx_PySet_Remove(__pyx_cur_scope->__pyx_v_seen, __pyx_cur_scope->__pyx_v_f); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 663, __pyx_L1_error)

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":664
 *             s, f = branch[-1].pop(-1)
 *             seen.remove(f)
 *             s, f = branch[-1][-1]             # <<<<<<<<<<<<<<
 *             seen.add(f)
 * 
 */
    __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_cur_scope->__pyx_v_branch, -1L, long, 1, __Pyx_PyInt_From_long, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 664, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_1, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 664, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
      PyObject* sequence = __pyx_t_6;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 664, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      #else
      __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 664, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 664, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      #endif
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 664, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_9 = Py_TYPE(__pyx_t_7)->tp_iternext;
      index = 0; __pyx_t_1 = __pyx_t_9(__pyx_t_7); if (unlikely(!__pyx_t_1)) goto __pyx_L34_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 1; __pyx_t_2 = __pyx_t_9(__pyx_t_7); if (unlikely(!__pyx_t_2)) goto __pyx_L34_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_2);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_7), 2) < 0) __PYX_ERR(0, 664, __pyx_L1_error)
      __pyx_t_9 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L35_unpacking_done;
      __pyx_L34_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_9 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 664, __pyx_L1_error)
      __pyx_L35_unpacking_done:;
    }
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_s);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_s, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_f);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_f, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":665
 *             seen.remove(f)
 *             s, f = branch[-1][-1]
 *             seen.add(f)             # <<<<<<<<<<<<<<
 * 
 *     def random_element(self, ring, *args, **kwds):
 */
    __pyx_t_10 = PySet_Add(__pyx_cur_scope->__pyx_v_seen, __pyx_cur_scope->__pyx_v_f); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 665, __pyx_L1_error)
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":598
 *         return ans
 * 
 *     def tree(self, max_depth=5, verbose=False):             # <<<<<<<<<<<<<<
 *         r"""
 *         The tree should be smarter in several ways:
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("tree", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "surface_dynamics/interval_exchanges/iet_family.pyx":667
 *             seen.add(f)
 * 
 *     def random_element(self, ring, *args, **kwds):             # <<<<<<<<<<<<<<
 *         r"""
 *         Return a random point in the cone
 */

/* Python wrapper */
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_28random_element(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_27random_element[] = "\n        Return a random point in the cone\n\n        EXAMPLES::\n\n            sage: from surface_dynamics import *\n            sage: from surface_dynamics.misc.linalg import deformation_space\n            sage: q = iet.Permutation([0,1,2,3,4,5],[5,3,2,1,0,4])\n            sage: rays = [[0, 0, 0, 1, 1, 0], [3, 1, 0, 1, 0, 2], [5, 0, 1, 2, 0, 3]]\n            sage: F = iet.IETFamily(q, rays) # optional - pplpy\n            sage: F # optional - pplpy\n            top 0 1 2 3 4 5\n            bot 5 3 2 1 0 4\n            0 0 0 1 1 0\n            3 1 0 1 0 2\n            5 0 1 2 0 3\n            sage: x = polygen(ZZ)\n            sage: K.<a> = NumberField(x^10 - 3, embedding=AA(3)**(1/10))\n            sage: T = F.random_element(K) # optional - pplpy\n            sage: V = deformation_space(T.lengths()) # optional - pplpy\n            sage: (QQ**6).subspace(F.rays()) == V # optional - pplpy\n            True\n        ";
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_28random_element(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_ring = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwds = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("random_element (wrapper)", 0);
  __pyx_v_kwds = PyDict_New(); if (unlikely(!__pyx_v_kwds)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwds);
  if (PyTuple_GET_SIZE(__pyx_args) > 1) {
    __pyx_v_args = PyTuple_GetSlice(__pyx_args, 1, PyTuple_GET_SIZE(__pyx_args));
    if (unlikely(!__pyx_v_args)) {
      __Pyx_DECREF(__pyx_v_kwds); __pyx_v_kwds = 0;
      __Pyx_RefNannyFinishContext();
      return NULL;
    }
    __Pyx_GOTREF(__pyx_v_args);
  } else {
    __pyx_v_args = __pyx_empty_tuple; __Pyx_INCREF(__pyx_empty_tuple);
  }
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_ring,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        default:
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ring)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t used_pos_args = (pos_args < 1) ? pos_args : 1;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwds, values, used_pos_args, "random_element") < 0)) __PYX_ERR(0, 667, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) < 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_ring = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("random_element", 0, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 667, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_DECREF(__pyx_v_kwds); __pyx_v_kwds = 0;
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.random_element", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_27random_element(((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)__pyx_v_self), __pyx_v_ring, __pyx_v_args, __pyx_v_kwds);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwds);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_14random_element_2generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "surface_dynamics/interval_exchanges/iet_family.pyx":703
 *                 c = ring.random_element(*args, **kwds)
 *             coeffs.append(c.abs())
 *         lengths = [sum(coeffs[i] * self.ray_coefficient(i, j) for i in range(self.length)) for j in range(self.dim)]             # <<<<<<<<<<<<<<
 * 
 *         return IET(p, lengths)
 */

static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_14random_element_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr *)__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr(__pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 703, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_14random_element_2generator4, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_random_element_locals_genexpr, __pyx_n_s_surface_dynamics_interval_exchan_2); if (unlikely(!gen)) __PYX_ERR(0, 703, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.random_element.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_14random_element_2generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr *__pyx_cur_scope = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 703, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 703, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self->length); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 703, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 703, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 703, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 703, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 703, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 703, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_4(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 703, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_i);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_i, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_coeffs)) { __Pyx_RaiseClosureNameError("coeffs"); __PYX_ERR(0, 703, __pyx_L1_error) }
    if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_coeffs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 703, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_coeffs, __pyx_cur_scope->__pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 703, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 703, __pyx_L1_error) }
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self), __pyx_n_s_ray_coefficient); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 703, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_FromSize_t(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_j); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 703, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_cur_scope->__pyx_v_i, __pyx_t_7};
      __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 703, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_cur_scope->__pyx_v_i, __pyx_t_7};
      __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 703, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 703, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_i);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_cur_scope->__pyx_v_i);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_10, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 703, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyNumber_Multiply(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 703, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 703, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "surface_dynamics/interval_exchanges/iet_family.pyx":667
 *             seen.add(f)
 * 
 *     def random_element(self, ring, *args, **kwds):             # <<<<<<<<<<<<<<
 *         r"""
 *         Return a random point in the cone
 */

static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_27random_element(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self, PyObject *__pyx_v_ring, PyObject *__pyx_v_args, PyObject *__pyx_v_kwds) {
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element *__pyx_cur_scope;
  PyObject *__pyx_v_Permutation = NULL;
  PyObject *__pyx_v_IET = NULL;
  PyObject *__pyx_v_top = NULL;
  PyObject *__pyx_v_bot = NULL;
  PyObject *__pyx_v_p = NULL;
  CYTHON_UNUSED size_t __pyx_v__;
  PyObject *__pyx_v_c = NULL;
  PyObject *__pyx_v_lengths = NULL;
  size_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("random_element", 0);
  __pyx_cur_scope = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element *)__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element(__pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 667, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":691
 *             True
 *         """
 *         from constructors import Permutation, IET             # <<<<<<<<<<<<<<
 * 
 *         top = [self.perm[i] for i in range(self.dim)]
 */
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_Permutation);
  __Pyx_GIVEREF(__pyx_n_s_Permutation);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_Permutation);
  __Pyx_INCREF(__pyx_n_s_IET);
  __Pyx_GIVEREF(__pyx_n_s_IET);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_IET);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_constructors, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Permutation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_Permutation = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_IET); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_IET = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":693
 *         from constructors import Permutation, IET
 * 
 *         top = [self.perm[i] for i in range(self.dim)]             # <<<<<<<<<<<<<<
 *         bot = [self.perm[i] for i in range(self.dim,2*self.dim)]
 *         p = Permutation(top, bot, alphabet=range(self.dim))
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 693, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_cur_scope->__pyx_v_self->dim;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;
    __pyx_t_1 = __Pyx_PyInt_From_int((__pyx_cur_scope->__pyx_v_self->perm[__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 693, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 693, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_v_top = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":694
 * 
 *         top = [self.perm[i] for i in range(self.dim)]
 *         bot = [self.perm[i] for i in range(self.dim,2*self.dim)]             # <<<<<<<<<<<<<<
 *         p = Permutation(top, bot, alphabet=range(self.dim))
 * 
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = (2 * __pyx_cur_scope->__pyx_v_self->dim);
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = __pyx_cur_scope->__pyx_v_self->dim; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;
    __pyx_t_1 = __Pyx_PyInt_From_int((__pyx_cur_scope->__pyx_v_self->perm[__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 694, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_v_bot = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":695
 *         top = [self.perm[i] for i in range(self.dim)]
 *         bot = [self.perm[i] for i in range(self.dim,2*self.dim)]
 *         p = Permutation(top, bot, alphabet=range(self.dim))             # <<<<<<<<<<<<<<
 * 
 *         coeffs = []
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_top);
  __Pyx_GIVEREF(__pyx_v_top);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_top);
  __Pyx_INCREF(__pyx_v_bot);
  __Pyx_GIVEREF(__pyx_v_bot);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_bot);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyInt_FromSize_t(__pyx_cur_scope->__pyx_v_self->dim); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_alphabet, __pyx_t_7) < 0) __PYX_ERR(0, 695, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_v_Permutation, __pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_p = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":697
 *         p = Permutation(top, bot, alphabet=range(self.dim))
 * 
 *         coeffs = []             # <<<<<<<<<<<<<<
 *         for _ in range(self.length):
 *             c = ring.random_element(*args, **kwds)
 */
  __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_cur_scope->__pyx_v_coeffs = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":698
 * 
 *         coeffs = []
 *         for _ in range(self.length):             # <<<<<<<<<<<<<<
 *             c = ring.random_element(*args, **kwds)
 *             while c.is_zero():
 */
  __pyx_t_3 = __pyx_cur_scope->__pyx_v_self->length;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v__ = __pyx_t_5;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":699
 *         coeffs = []
 *         for _ in range(self.length):
 *             c = ring.random_element(*args, **kwds)             # <<<<<<<<<<<<<<
 *             while c.is_zero():
 *                 c = ring.random_element(*args, **kwds)
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_ring, __pyx_n_s_random_element); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_v_args, __pyx_v_kwds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF_SET(__pyx_v_c, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":700
 *         for _ in range(self.length):
 *             c = ring.random_element(*args, **kwds)
 *             while c.is_zero():             # <<<<<<<<<<<<<<
 *                 c = ring.random_element(*args, **kwds)
 *             coeffs.append(c.abs())
 */
    while (1) {
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_c, __pyx_n_s_is_zero); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 700, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_2 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 700, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 700, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (!__pyx_t_8) break;

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":701
 *             c = ring.random_element(*args, **kwds)
 *             while c.is_zero():
 *                 c = ring.random_element(*args, **kwds)             # <<<<<<<<<<<<<<
 *             coeffs.append(c.abs())
 *         lengths = [sum(coeffs[i] * self.ray_coefficient(i, j) for i in range(self.length)) for j in range(self.dim)]
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_ring, __pyx_n_s_random_element); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 701, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_v_args, __pyx_v_kwds); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 701, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF_SET(__pyx_v_c, __pyx_t_7);
      __pyx_t_7 = 0;
    }

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":702
 *             while c.is_zero():
 *                 c = ring.random_element(*args, **kwds)
 *             coeffs.append(c.abs())             # <<<<<<<<<<<<<<
 *         lengths = [sum(coeffs[i] * self.ray_coefficient(i, j) for i in range(self.length)) for j in range(self.dim)]
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_c, __pyx_n_s_abs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 702, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_7 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 702, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_9 = __Pyx_PyList_Append(__pyx_cur_scope->__pyx_v_coeffs, __pyx_t_7); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 702, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":703
 *                 c = ring.random_element(*args, **kwds)
 *             coeffs.append(c.abs())
 *         lengths = [sum(coeffs[i] * self.ray_coefficient(i, j) for i in range(self.length)) for j in range(self.dim)]             # <<<<<<<<<<<<<<
 * 
 *         return IET(p, lengths)
 */
  __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = __pyx_cur_scope->__pyx_v_self->dim;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_cur_scope->__pyx_v_j = __pyx_t_5;
    __pyx_t_1 = __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_14random_element_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 703, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 703, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_7, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 703, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_lengths = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":705
 *         lengths = [sum(coeffs[i] * self.ray_coefficient(i, j) for i in range(self.length)) for j in range(self.dim)]
 * 
 *         return IET(p, lengths)             # <<<<<<<<<<<<<<
 * 
 *     # TODO: find a better name and output!!!
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_IET);
  __pyx_t_2 = __pyx_v_IET; __pyx_t_1 = NULL;
  __pyx_t_10 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_10 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_p, __pyx_v_lengths};
    __pyx_t_7 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 705, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_7);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_p, __pyx_v_lengths};
    __pyx_t_7 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 705, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_7);
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 705, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_1) {
      __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1); __pyx_t_1 = NULL;
    }
    __Pyx_INCREF(__pyx_v_p);
    __Pyx_GIVEREF(__pyx_v_p);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_10, __pyx_v_p);
    __Pyx_INCREF(__pyx_v_lengths);
    __Pyx_GIVEREF(__pyx_v_lengths);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_10, __pyx_v_lengths);
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 705, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":667
 *             seen.add(f)
 * 
 *     def random_element(self, ring, *args, **kwds):             # <<<<<<<<<<<<<<
 *         r"""
 *         Return a random point in the cone
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.random_element", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_Permutation);
  __Pyx_XDECREF(__pyx_v_IET);
  __Pyx_XDECREF(__pyx_v_top);
  __Pyx_XDECREF(__pyx_v_bot);
  __Pyx_XDECREF(__pyx_v_p);
  __Pyx_XDECREF(__pyx_v_c);
  __Pyx_XDECREF(__pyx_v_lengths);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "surface_dynamics/interval_exchanges/iet_family.pyx":708
 * 
 *     # TODO: find a better name and output!!!
 *     def random_element_statistics(self, ring, num_exp=100, num_iterations=200, *args, **kwds):             # <<<<<<<<<<<<<<
 *         r"""
 *         INPUT:
 */

/* Python wrapper */
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_30random_element_statistics(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_29random_element_statistics[] = "\n        INPUT:\n\n        - ``ring`` -- the ring in which is taken the lengths (usually a number field)\n\n        - ``num_exp`` -- number of experiments\n\n        - ``num_iterations`` -- the number of Zorich move to be performed on the\n          interval exchange transformations to find a saddle connection\n\n        EXAMPLES::\n\n            sage: from surface_dynamics import *\n            sage: q = iet.Permutation([0,1,2,3,4,5],[5,3,2,1,0,4])\n            sage: rays = [[0, 0, 0, 1, 1, 0], [3, 1, 0, 1, 0, 2], [5, 0, 1, 2, 0, 3]]\n            sage: F = iet.IETFamily(q, rays)  # optional - pplpy\n            sage: F.random_element_statistics(K)  # optional - pplpy\n            100 saddles found on 100 random point\n\n        A conjectural counterexample to Dynnikov-Skripshenko conjecture::\n\n            sage: path = 'bbbbbtbttbttbbttbtttbbbbttbttbtbbtbbtbbtt'\n            sage: p = iet.Permutation('a b c d e f g h i', 'i h g f e d c b a')\n            sage: R = p.rauzy_diagram()\n            sage: T = R.path(p, *path).self_similar_iet()\n            sage: T.sah_arnoux_fathi_invariant()\n            (0, 0, 0, 0, 0, 0)\n            sage: F = iet.IETFamily(T)  # optional - pplpy\n            sage: F  # optional - pplpy\n            top 0 1 2 3 4 5 6 7 8\n            bot 8 7 6 5 4 3 2 1 0\n            71 67 70 107 0 0 19 2 0\n            95 63 21 44 35 0 0 0 43\n            105 94 88 147 0 0 0 15 19\n            253 310 0 243 34 101 68 0 0\n            425 395 404 611 14 0 129 0 0\n            524 703 0 628 0 245 51 68 0\n            619 462 0 229 238 63 0 0 272\n            701 910 0 823 0 308 0 119 51\n            sage: K.<cbrt3> = NumberField(x^3 - 3, embedding=AA(3)**(1/3))\n            sage: F.random_element_statistics(K, num_exp=20, num_iterations=200)  # optional - pplpy # random\n            ... saddles found on 20 random points\n        ";
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_30random_element_statistics(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_ring = 0;
  PyObject *__pyx_v_num_exp = 0;
  PyObject *__pyx_v_num_iterations = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwds = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("random_element_statistics (wrapper)", 0);
  __pyx_v_kwds = PyDict_New(); if (unlikely(!__pyx_v_kwds)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwds);
  if (PyTuple_GET_SIZE(__pyx_args) > 3) {
    __pyx_v_args = PyTuple_GetSlice(__pyx_args, 3, PyTuple_GET_SIZE(__pyx_args));
    if (unlikely(!__pyx_v_args)) {
      __Pyx_DECREF(__pyx_v_kwds); __pyx_v_kwds = 0;
      __Pyx_RefNannyFinishContext();
      return NULL;
    }
    __Pyx_GOTREF(__pyx_v_args);
  } else {
    __pyx_v_args = __pyx_empty_tuple; __Pyx_INCREF(__pyx_empty_tuple);
  }
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_ring,&__pyx_n_s_num_exp,&__pyx_n_s_num_iterations,0};
    PyObject* values[3] = {0,0,0};
    values[1] = ((PyObject *)__pyx_int_100);
    values[2] = ((PyObject *)__pyx_int_200);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        default:
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ring)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_num_exp);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_num_iterations);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t used_pos_args = (pos_args < 3) ? pos_args : 3;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwds, values, used_pos_args, "random_element_statistics") < 0)) __PYX_ERR(0, 708, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        default:
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        case  0:
        goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_ring = values[0];
    __pyx_v_num_exp = values[1];
    __pyx_v_num_iterations = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("random_element_statistics", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 708, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_DECREF(__pyx_v_kwds); __pyx_v_kwds = 0;
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.random_element_statistics", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_29random_element_statistics(((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)__pyx_v_self), __pyx_v_ring, __pyx_v_num_exp, __pyx_v_num_iterations, __pyx_v_args, __pyx_v_kwds);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwds);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_29random_element_statistics(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self, PyObject *__pyx_v_ring, PyObject *__pyx_v_num_exp, PyObject *__pyx_v_num_iterations, PyObject *__pyx_v_args, PyObject *__pyx_v_kwds) {
  PyObject *__pyx_v_num_saddles = NULL;
  PyObject *__pyx_v_bads = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_T = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("random_element_statistics", 0);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":752
 *             ... saddles found on 20 random points
 *         """
 *         num_saddles = 0             # <<<<<<<<<<<<<<
 *         bads = []
 *         for i in range(num_exp):
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_v_num_saddles = __pyx_int_0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":753
 *         """
 *         num_saddles = 0
 *         bads = []             # <<<<<<<<<<<<<<
 *         for i in range(num_exp):
 *             T = self.random_element(ring, *args, **kwds)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_bads = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":754
 *         num_saddles = 0
 *         bads = []
 *         for i in range(num_exp):             # <<<<<<<<<<<<<<
 *             T = self.random_element(ring, *args, **kwds)
 *             try:
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_v_num_exp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 754, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 754, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 754, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 754, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 754, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 754, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 754, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":755
 *         bads = []
 *         for i in range(num_exp):
 *             T = self.random_element(ring, *args, **kwds)             # <<<<<<<<<<<<<<
 *             try:
 *                 T.zorich_move(iterations=num_iterations)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_random_element); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 755, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 755, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_ring);
    __Pyx_GIVEREF(__pyx_v_ring);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_ring);
    __pyx_t_6 = PyNumber_Add(__pyx_t_5, __pyx_v_args); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 755, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, __pyx_v_kwds); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 755, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_T, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":756
 *         for i in range(num_exp):
 *             T = self.random_element(ring, *args, **kwds)
 *             try:             # <<<<<<<<<<<<<<
 *                 T.zorich_move(iterations=num_iterations)
 *             except ValueError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      /*try:*/ {

        /* "surface_dynamics/interval_exchanges/iet_family.pyx":757
 *             T = self.random_element(ring, *args, **kwds)
 *             try:
 *                 T.zorich_move(iterations=num_iterations)             # <<<<<<<<<<<<<<
 *             except ValueError:
 *                 num_saddles += 1
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_T, __pyx_n_s_zorich_move); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 757, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 757, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_iterations, __pyx_v_num_iterations) < 0) __PYX_ERR(0, 757, __pyx_L5_error)
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 757, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "surface_dynamics/interval_exchanges/iet_family.pyx":756
 *         for i in range(num_exp):
 *             T = self.random_element(ring, *args, **kwds)
 *             try:             # <<<<<<<<<<<<<<
 *                 T.zorich_move(iterations=num_iterations)
 *             except ValueError:
 */
      }

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":761
 *                 num_saddles += 1
 *             else:
 *                 bads.append(T)             # <<<<<<<<<<<<<<
 *         print('%d saddles found on %d random points' % (num_saddles, num_exp))
 * 
 */
      /*else:*/ {
        __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_bads, __pyx_v_T); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 761, __pyx_L7_except_error)
      }
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L12_try_end;
      __pyx_L5_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":758
 *             try:
 *                 T.zorich_move(iterations=num_iterations)
 *             except ValueError:             # <<<<<<<<<<<<<<
 *                 num_saddles += 1
 *             else:
 */
      __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
      if (__pyx_t_11) {
        __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.random_element_statistics", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_6, &__pyx_t_5) < 0) __PYX_ERR(0, 758, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GOTREF(__pyx_t_5);

        /* "surface_dynamics/interval_exchanges/iet_family.pyx":759
 *                 T.zorich_move(iterations=num_iterations)
 *             except ValueError:
 *                 num_saddles += 1             # <<<<<<<<<<<<<<
 *             else:
 *                 bads.append(T)
 */
        __pyx_t_12 = __Pyx_PyInt_AddObjC(__pyx_v_num_saddles, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 759, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF_SET(__pyx_v_num_saddles, __pyx_t_12);
        __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L6_exception_handled;
      }
      goto __pyx_L7_except_error;
      __pyx_L7_except_error:;

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":756
 *         for i in range(num_exp):
 *             T = self.random_element(ring, *args, **kwds)
 *             try:             # <<<<<<<<<<<<<<
 *                 T.zorich_move(iterations=num_iterations)
 *             except ValueError:
 */
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      goto __pyx_L1_error;
      __pyx_L6_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      __pyx_L12_try_end:;
    }

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":754
 *         num_saddles = 0
 *         bads = []
 *         for i in range(num_exp):             # <<<<<<<<<<<<<<
 *             T = self.random_element(ring, *args, **kwds)
 *             try:
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":762
 *             else:
 *                 bads.append(T)
 *         print('%d saddles found on %d random points' % (num_saddles, num_exp))             # <<<<<<<<<<<<<<
 * 
 *     def random_integer_point_statistics(self, num_exp=100):
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_num_saddles);
  __Pyx_GIVEREF(__pyx_v_num_saddles);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_num_saddles);
  __Pyx_INCREF(__pyx_v_num_exp);
  __Pyx_GIVEREF(__pyx_v_num_exp);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_num_exp);
  __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_d_saddles_found_on_d_random_poi, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_PrintOne(0, __pyx_t_5) < 0) __PYX_ERR(0, 762, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":708
 * 
 *     # TODO: find a better name and output!!!
 *     def random_element_statistics(self, ring, num_exp=100, num_iterations=200, *args, **kwds):             # <<<<<<<<<<<<<<
 *         r"""
 *         INPUT:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.random_element_statistics", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_num_saddles);
  __Pyx_XDECREF(__pyx_v_bads);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_T);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "surface_dynamics/interval_exchanges/iet_family.pyx":764
 *         print('%d saddles found on %d random points' % (num_saddles, num_exp))
 * 
 *     def random_integer_point_statistics(self, num_exp=100):             # <<<<<<<<<<<<<<
 *         r"""
 *         OUTPUT: a dictionary whose keys are integers ``i`` and values are the
 */

/* Python wrapper */
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_32random_integer_point_statistics(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_31random_integer_point_statistics[] = "\n        OUTPUT: a dictionary whose keys are integers ``i`` and values are the\n        number of integer points found so that the number of cylinders is\n        ``i``.\n        ";
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_32random_integer_point_statistics(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_num_exp = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("random_integer_point_statistics (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_num_exp,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)__pyx_int_100);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_num_exp);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "random_integer_point_statistics") < 0)) __PYX_ERR(0, 764, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_num_exp = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("random_integer_point_statistics", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 764, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.random_integer_point_statistics", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_31random_integer_point_statistics(((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)__pyx_v_self), __pyx_v_num_exp);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_31random_integer_point_statistics_2generator5(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "surface_dynamics/interval_exchanges/iet_family.pyx":777
 *         cdef dict cyls = {}
 *         for _ in range(num_exp):
 *             lengths = sum(ZZ.random_element(0,100000000) * r for r in self.rays())             # <<<<<<<<<<<<<<
 *             ncyls = iet_interface.number_of_cylinders(top, bot, lengths)
 *             if ncyls not in cyls:
 */

static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_31random_integer_point_statistics_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr *)__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr(__pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 777, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_31random_integer_point_statistics_2generator5, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_random_integer_point_statistics, __pyx_n_s_surface_dynamics_interval_exchan_2); if (unlikely(!gen)) __PYX_ERR(0, 777, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.random_integer_point_statistics.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_31random_integer_point_statistics_2generator5(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr *__pyx_cur_scope = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 777, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 777, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_outer_scope->__pyx_v_self), __pyx_n_s_rays); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 777, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 777, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 777, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 777, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 777, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 777, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_r);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_r, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ZZ)) { __Pyx_RaiseClosureNameError("ZZ"); __PYX_ERR(0, 777, __pyx_L1_error) }
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ZZ, __pyx_n_s_random_element); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Multiply(__pyx_t_3, __pyx_cur_scope->__pyx_v_r); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_4;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_5;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_5 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 777, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "surface_dynamics/interval_exchanges/iet_family.pyx":764
 *         print('%d saddles found on %d random points' % (num_saddles, num_exp))
 * 
 *     def random_integer_point_statistics(self, num_exp=100):             # <<<<<<<<<<<<<<
 *         r"""
 *         OUTPUT: a dictionary whose keys are integers ``i`` and values are the
 */

static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_31random_integer_point_statistics(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self, PyObject *__pyx_v_num_exp) {
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics *__pyx_cur_scope;
  PyObject *__pyx_v_iet_interface = NULL;
  PyObject *__pyx_v_top = NULL;
  PyObject *__pyx_v_bot = NULL;
  CYTHON_UNUSED Py_ssize_t __pyx_v_dim;
  PyObject *__pyx_v_cyls = 0;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_v_lengths = NULL;
  PyObject *__pyx_v_ncyls = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("random_integer_point_statistics", 0);
  __pyx_cur_scope = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics *)__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics(__pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 764, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":771
 *         """
 *         # TODO: make iet_interface part of flatsurf
 *         import iet_interface             # <<<<<<<<<<<<<<
 *         from sage.rings.integer_ring import ZZ
 *         top, bot = self.permutation()
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_iet_interface, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_iet_interface = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":772
 *         # TODO: make iet_interface part of flatsurf
 *         import iet_interface
 *         from sage.rings.integer_ring import ZZ             # <<<<<<<<<<<<<<
 *         top, bot = self.permutation()
 *         dim = len(top)
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_ZZ);
  __Pyx_GIVEREF(__pyx_n_s_ZZ);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_ZZ);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_sage_rings_integer_ring, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_ZZ); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_ZZ = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":773
 *         import iet_interface
 *         from sage.rings.integer_ring import ZZ
 *         top, bot = self.permutation()             # <<<<<<<<<<<<<<
 *         dim = len(top)
 *         cdef dict cyls = {}
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_permutation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
    PyObject* sequence = __pyx_t_2;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 773, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 773, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 773, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 773, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext;
    index = 0; __pyx_t_1 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_1)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_1);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(0, 773, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 773, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_top = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_bot = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":774
 *         from sage.rings.integer_ring import ZZ
 *         top, bot = self.permutation()
 *         dim = len(top)             # <<<<<<<<<<<<<<
 *         cdef dict cyls = {}
 *         for _ in range(num_exp):
 */
  __pyx_t_6 = PyObject_Length(__pyx_v_top); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 774, __pyx_L1_error)
  __pyx_v_dim = __pyx_t_6;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":775
 *         top, bot = self.permutation()
 *         dim = len(top)
 *         cdef dict cyls = {}             # <<<<<<<<<<<<<<
 *         for _ in range(num_exp):
 *             lengths = sum(ZZ.random_element(0,100000000) * r for r in self.rays())
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_cyls = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":776
 *         dim = len(top)
 *         cdef dict cyls = {}
 *         for _ in range(num_exp):             # <<<<<<<<<<<<<<
 *             lengths = sum(ZZ.random_element(0,100000000) * r for r in self.rays())
 *             ncyls = iet_interface.number_of_cylinders(top, bot, lengths)
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_v_num_exp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 776, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 776, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 776, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 776, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 776, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 776, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_7(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 776, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v__, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":777
 *         cdef dict cyls = {}
 *         for _ in range(num_exp):
 *             lengths = sum(ZZ.random_element(0,100000000) * r for r in self.rays())             # <<<<<<<<<<<<<<
 *             ncyls = iet_interface.number_of_cylinders(top, bot, lengths)
 *             if ncyls not in cyls:
 */
    __pyx_t_2 = __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_31random_integer_point_statistics_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_lengths, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":778
 *         for _ in range(num_exp):
 *             lengths = sum(ZZ.random_element(0,100000000) * r for r in self.rays())
 *             ncyls = iet_interface.number_of_cylinders(top, bot, lengths)             # <<<<<<<<<<<<<<
 *             if ncyls not in cyls:
 *                 cyls[ncyls] = 0
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_iet_interface, __pyx_n_s_number_of_cylinders); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 778, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_top, __pyx_v_bot, __pyx_v_lengths};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 3+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 778, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_v_top, __pyx_v_bot, __pyx_v_lengths};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 3+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 778, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_9 = PyTuple_New(3+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 778, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_v_top);
      __Pyx_GIVEREF(__pyx_v_top);
      PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_v_top);
      __Pyx_INCREF(__pyx_v_bot);
      __Pyx_GIVEREF(__pyx_v_bot);
      PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_bot);
      __Pyx_INCREF(__pyx_v_lengths);
      __Pyx_GIVEREF(__pyx_v_lengths);
      PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_8, __pyx_v_lengths);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 778, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_ncyls, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":779
 *             lengths = sum(ZZ.random_element(0,100000000) * r for r in self.rays())
 *             ncyls = iet_interface.number_of_cylinders(top, bot, lengths)
 *             if ncyls not in cyls:             # <<<<<<<<<<<<<<
 *                 cyls[ncyls] = 0
 *             cyls[ncyls] += 1
 */
    __pyx_t_10 = (__Pyx_PyDict_ContainsTF(__pyx_v_ncyls, __pyx_v_cyls, Py_NE)); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 779, __pyx_L1_error)
    __pyx_t_11 = (__pyx_t_10 != 0);
    if (__pyx_t_11) {

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":780
 *             ncyls = iet_interface.number_of_cylinders(top, bot, lengths)
 *             if ncyls not in cyls:
 *                 cyls[ncyls] = 0             # <<<<<<<<<<<<<<
 *             cyls[ncyls] += 1
 *         return cyls
 */
      if (unlikely(PyDict_SetItem(__pyx_v_cyls, __pyx_v_ncyls, __pyx_int_0) < 0)) __PYX_ERR(0, 780, __pyx_L1_error)

      /* "surface_dynamics/interval_exchanges/iet_family.pyx":779
 *             lengths = sum(ZZ.random_element(0,100000000) * r for r in self.rays())
 *             ncyls = iet_interface.number_of_cylinders(top, bot, lengths)
 *             if ncyls not in cyls:             # <<<<<<<<<<<<<<
 *                 cyls[ncyls] = 0
 *             cyls[ncyls] += 1
 */
    }

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":781
 *             if ncyls not in cyls:
 *                 cyls[ncyls] = 0
 *             cyls[ncyls] += 1             # <<<<<<<<<<<<<<
 *         return cyls
 * 
 */
    __Pyx_INCREF(__pyx_v_ncyls);
    __pyx_t_1 = __pyx_v_ncyls;
    __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_cyls, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 781, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = __Pyx_PyInt_AddObjC(__pyx_t_2, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 781, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(PyDict_SetItem(__pyx_v_cyls, __pyx_t_1, __pyx_t_9) < 0)) __PYX_ERR(0, 781, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "surface_dynamics/interval_exchanges/iet_family.pyx":776
 *         dim = len(top)
 *         cdef dict cyls = {}
 *         for _ in range(num_exp):             # <<<<<<<<<<<<<<
 *             lengths = sum(ZZ.random_element(0,100000000) * r for r in self.rays())
 *             ncyls = iet_interface.number_of_cylinders(top, bot, lengths)
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":782
 *                 cyls[ncyls] = 0
 *             cyls[ncyls] += 1
 *         return cyls             # <<<<<<<<<<<<<<
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_cyls);
  __pyx_r = __pyx_v_cyls;
  goto __pyx_L0;

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":764
 *         print('%d saddles found on %d random points' % (num_saddles, num_exp))
 * 
 *     def random_integer_point_statistics(self, num_exp=100):             # <<<<<<<<<<<<<<
 *         r"""
 *         OUTPUT: a dictionary whose keys are integers ``i`` and values are the
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.random_integer_point_statistics", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_iet_interface);
  __Pyx_XDECREF(__pyx_v_top);
  __Pyx_XDECREF(__pyx_v_bot);
  __Pyx_XDECREF(__pyx_v_cyls);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XDECREF(__pyx_v_lengths);
  __Pyx_XDECREF(__pyx_v_ncyls);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_34__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_34__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_33__reduce_cython__(((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_33__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_36__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_36__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_35__setstate_cython__(((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_35__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("surface_dynamics.interval_exchanges.iet_family.IETFamily.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "signals.pxd":57
 * # (PyErr_Occurred() is non-NULL). To Cython, it will look like
 * # cython_check_exception() actually raised the exception.
 * cdef inline void cython_check_exception() nogil except *:             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

static CYTHON_INLINE void __pyx_f_9cysignals_7signals_cython_check_exception(void) {

  /* function exit code */
}

/* "signals.pxd":63
 * cdef void verify_exc_value()
 * 
 * cdef inline PyObject* sig_occurred():             # <<<<<<<<<<<<<<
 *     """
 *     Borrowed reference to the exception which is currently being
 */

static CYTHON_INLINE PyObject *__pyx_f_9cysignals_7signals_sig_occurred(void) {
  PyObject *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("sig_occurred", 0);

  /* "signals.pxd":77
 *     exception is stored somewhere.
 *     """
 *     if unlikely(cysigs.exc_value is not NULL):             # <<<<<<<<<<<<<<
 *         verify_exc_value()
 *     return cysigs.exc_value
 */
  __pyx_t_1 = (unlikely((cysigs.exc_value != NULL)) != 0);
  if (__pyx_t_1) {

    /* "signals.pxd":78
 *     """
 *     if unlikely(cysigs.exc_value is not NULL):
 *         verify_exc_value()             # <<<<<<<<<<<<<<
 *     return cysigs.exc_value
 * 
 */
    __pyx_f_9cysignals_7signals_verify_exc_value();

    /* "signals.pxd":77
 *     exception is stored somewhere.
 *     """
 *     if unlikely(cysigs.exc_value is not NULL):             # <<<<<<<<<<<<<<
 *         verify_exc_value()
 *     return cysigs.exc_value
 */
  }

  /* "signals.pxd":79
 *     if unlikely(cysigs.exc_value is not NULL):
 *         verify_exc_value()
 *     return cysigs.exc_value             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = cysigs.exc_value;
  goto __pyx_L0;

  /* "signals.pxd":63
 * cdef void verify_exc_value()
 * 
 * cdef inline PyObject* sig_occurred():             # <<<<<<<<<<<<<<
 *     """
 *     Borrowed reference to the exception which is currently being
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "signals.pxd":93
 * 
 * 
 * cdef inline void __generate_declarations():             # <<<<<<<<<<<<<<
 *     cysigs
 *     _sig_on_interrupt_received
 */

static CYTHON_INLINE void __pyx_f_9cysignals_7signals___generate_declarations(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__generate_declarations", 0);

  /* "signals.pxd":94
 * 
 * cdef inline void __generate_declarations():
 *     cysigs             # <<<<<<<<<<<<<<
 *     _sig_on_interrupt_received
 *     _sig_on_recover
 */
  (void)(cysigs);

  /* "signals.pxd":95
 * cdef inline void __generate_declarations():
 *     cysigs
 *     _sig_on_interrupt_received             # <<<<<<<<<<<<<<
 *     _sig_on_recover
 *     _sig_off_warning
 */
  (void)(_sig_on_interrupt_received);

  /* "signals.pxd":96
 *     cysigs
 *     _sig_on_interrupt_received
 *     _sig_on_recover             # <<<<<<<<<<<<<<
 *     _sig_off_warning
 *     print_backtrace
 */
  (void)(_sig_on_recover);

  /* "signals.pxd":97
 *     _sig_on_interrupt_received
 *     _sig_on_recover
 *     _sig_off_warning             # <<<<<<<<<<<<<<
 *     print_backtrace
 */
  (void)(_sig_off_warning);

  /* "signals.pxd":98
 *     _sig_on_recover
 *     _sig_off_warning
 *     print_backtrace             # <<<<<<<<<<<<<<
 */
  (void)(print_backtrace);

  /* "signals.pxd":93
 * 
 * 
 * cdef inline void __generate_declarations():             # <<<<<<<<<<<<<<
 *     cysigs
 *     _sig_on_interrupt_received
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "cysignals/memory.pxd":36
 * 
 * 
 * cdef inline void* sig_malloc "sig_malloc"(size_t n) nogil:             # <<<<<<<<<<<<<<
 *     sig_block()
 *     cdef void* ret = malloc(n)
 */

static CYTHON_INLINE void *sig_malloc(size_t __pyx_v_n) {
  void *__pyx_v_ret;
  void *__pyx_r;

  /* "cysignals/memory.pxd":37
 * 
 * cdef inline void* sig_malloc "sig_malloc"(size_t n) nogil:
 *     sig_block()             # <<<<<<<<<<<<<<
 *     cdef void* ret = malloc(n)
 *     sig_unblock()
 */
  sig_block();

  /* "cysignals/memory.pxd":38
 * cdef inline void* sig_malloc "sig_malloc"(size_t n) nogil:
 *     sig_block()
 *     cdef void* ret = malloc(n)             # <<<<<<<<<<<<<<
 *     sig_unblock()
 *     return ret
 */
  __pyx_v_ret = malloc(__pyx_v_n);

  /* "cysignals/memory.pxd":39
 *     sig_block()
 *     cdef void* ret = malloc(n)
 *     sig_unblock()             # <<<<<<<<<<<<<<
 *     return ret
 * 
 */
  sig_unblock();

  /* "cysignals/memory.pxd":40
 *     cdef void* ret = malloc(n)
 *     sig_unblock()
 *     return ret             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "cysignals/memory.pxd":36
 * 
 * 
 * cdef inline void* sig_malloc "sig_malloc"(size_t n) nogil:             # <<<<<<<<<<<<<<
 *     sig_block()
 *     cdef void* ret = malloc(n)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cysignals/memory.pxd":43
 * 
 * 
 * cdef inline void* sig_realloc "sig_realloc"(void* ptr, size_t size) nogil:             # <<<<<<<<<<<<<<
 *     sig_block()
 *     cdef void* ret = realloc(ptr, size)
 */

static CYTHON_INLINE void *sig_realloc(void *__pyx_v_ptr, size_t __pyx_v_size) {
  void *__pyx_v_ret;
  void *__pyx_r;

  /* "cysignals/memory.pxd":44
 * 
 * cdef inline void* sig_realloc "sig_realloc"(void* ptr, size_t size) nogil:
 *     sig_block()             # <<<<<<<<<<<<<<
 *     cdef void* ret = realloc(ptr, size)
 *     sig_unblock()
 */
  sig_block();

  /* "cysignals/memory.pxd":45
 * cdef inline void* sig_realloc "sig_realloc"(void* ptr, size_t size) nogil:
 *     sig_block()
 *     cdef void* ret = realloc(ptr, size)             # <<<<<<<<<<<<<<
 *     sig_unblock()
 *     return ret
 */
  __pyx_v_ret = realloc(__pyx_v_ptr, __pyx_v_size);

  /* "cysignals/memory.pxd":46
 *     sig_block()
 *     cdef void* ret = realloc(ptr, size)
 *     sig_unblock()             # <<<<<<<<<<<<<<
 *     return ret
 * 
 */
  sig_unblock();

  /* "cysignals/memory.pxd":47
 *     cdef void* ret = realloc(ptr, size)
 *     sig_unblock()
 *     return ret             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "cysignals/memory.pxd":43
 * 
 * 
 * cdef inline void* sig_realloc "sig_realloc"(void* ptr, size_t size) nogil:             # <<<<<<<<<<<<<<
 *     sig_block()
 *     cdef void* ret = realloc(ptr, size)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cysignals/memory.pxd":50
 * 
 * 
 * cdef inline void* sig_calloc "sig_calloc"(size_t nmemb, size_t size) nogil:             # <<<<<<<<<<<<<<
 *     sig_block()
 *     cdef void* ret = calloc(nmemb, size)
 */

static CYTHON_INLINE void *sig_calloc(size_t __pyx_v_nmemb, size_t __pyx_v_size) {
  void *__pyx_v_ret;
  void *__pyx_r;

  /* "cysignals/memory.pxd":51
 * 
 * cdef inline void* sig_calloc "sig_calloc"(size_t nmemb, size_t size) nogil:
 *     sig_block()             # <<<<<<<<<<<<<<
 *     cdef void* ret = calloc(nmemb, size)
 *     sig_unblock()
 */
  sig_block();

  /* "cysignals/memory.pxd":52
 * cdef inline void* sig_calloc "sig_calloc"(size_t nmemb, size_t size) nogil:
 *     sig_block()
 *     cdef void* ret = calloc(nmemb, size)             # <<<<<<<<<<<<<<
 *     sig_unblock()
 *     return ret
 */
  __pyx_v_ret = calloc(__pyx_v_nmemb, __pyx_v_size);

  /* "cysignals/memory.pxd":53
 *     sig_block()
 *     cdef void* ret = calloc(nmemb, size)
 *     sig_unblock()             # <<<<<<<<<<<<<<
 *     return ret
 * 
 */
  sig_unblock();

  /* "cysignals/memory.pxd":54
 *     cdef void* ret = calloc(nmemb, size)
 *     sig_unblock()
 *     return ret             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "cysignals/memory.pxd":50
 * 
 * 
 * cdef inline void* sig_calloc "sig_calloc"(size_t nmemb, size_t size) nogil:             # <<<<<<<<<<<<<<
 *     sig_block()
 *     cdef void* ret = calloc(nmemb, size)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cysignals/memory.pxd":57
 * 
 * 
 * cdef inline void sig_free "sig_free"(void* ptr) nogil:             # <<<<<<<<<<<<<<
 *     sig_block()
 *     free(ptr)
 */

static CYTHON_INLINE void sig_free(void *__pyx_v_ptr) {

  /* "cysignals/memory.pxd":58
 * 
 * cdef inline void sig_free "sig_free"(void* ptr) nogil:
 *     sig_block()             # <<<<<<<<<<<<<<
 *     free(ptr)
 *     sig_unblock()
 */
  sig_block();

  /* "cysignals/memory.pxd":59
 * cdef inline void sig_free "sig_free"(void* ptr) nogil:
 *     sig_block()
 *     free(ptr)             # <<<<<<<<<<<<<<
 *     sig_unblock()
 * 
 */
  free(__pyx_v_ptr);

  /* "cysignals/memory.pxd":60
 *     sig_block()
 *     free(ptr)
 *     sig_unblock()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  sig_unblock();

  /* "cysignals/memory.pxd":57
 * 
 * 
 * cdef inline void sig_free "sig_free"(void* ptr) nogil:             # <<<<<<<<<<<<<<
 *     sig_block()
 *     free(ptr)
 */

  /* function exit code */
}

/* "cysignals/memory.pxd":64
 * 
 * @cython.cdivision(True)
 * cdef inline size_t mul_overflowcheck(size_t a, size_t b) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Return a*b, checking for overflow. Assume that a > 0.
 */

static CYTHON_INLINE size_t __pyx_f_9cysignals_6memory_mul_overflowcheck(size_t __pyx_v_a, size_t __pyx_v_b) {
  size_t __pyx_v_MUL_NO_OVERFLOW;
  size_t __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "cysignals/memory.pxd":71
 *     """
 *     # If a and b both less than MUL_NO_OVERFLOW, no overflow can occur
 *     cdef size_t MUL_NO_OVERFLOW = ((<size_t>1) << (4*sizeof(size_t)))             # <<<<<<<<<<<<<<
 *     if a >= MUL_NO_OVERFLOW or b >= MUL_NO_OVERFLOW:
 *         if unlikely(b > (<size_t>-1) // a):
 */
  __pyx_v_MUL_NO_OVERFLOW = (((size_t)1) << (4 * (sizeof(size_t))));

  /* "cysignals/memory.pxd":72
 *     # If a and b both less than MUL_NO_OVERFLOW, no overflow can occur
 *     cdef size_t MUL_NO_OVERFLOW = ((<size_t>1) << (4*sizeof(size_t)))
 *     if a >= MUL_NO_OVERFLOW or b >= MUL_NO_OVERFLOW:             # <<<<<<<<<<<<<<
 *         if unlikely(b > (<size_t>-1) // a):
 *             return <size_t>(-1)
 */
  __pyx_t_2 = ((__pyx_v_a >= __pyx_v_MUL_NO_OVERFLOW) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_b >= __pyx_v_MUL_NO_OVERFLOW) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "cysignals/memory.pxd":73
 *     cdef size_t MUL_NO_OVERFLOW = ((<size_t>1) << (4*sizeof(size_t)))
 *     if a >= MUL_NO_OVERFLOW or b >= MUL_NO_OVERFLOW:
 *         if unlikely(b > (<size_t>-1) // a):             # <<<<<<<<<<<<<<
 *             return <size_t>(-1)
 *     return a*b
 */
    __pyx_t_1 = (unlikely((__pyx_v_b > (((size_t)-1L) / __pyx_v_a))) != 0);
    if (__pyx_t_1) {

      /* "cysignals/memory.pxd":74
 *     if a >= MUL_NO_OVERFLOW or b >= MUL_NO_OVERFLOW:
 *         if unlikely(b > (<size_t>-1) // a):
 *             return <size_t>(-1)             # <<<<<<<<<<<<<<
 *     return a*b
 * 
 */
      __pyx_r = ((size_t)-1L);
      goto __pyx_L0;

      /* "cysignals/memory.pxd":73
 *     cdef size_t MUL_NO_OVERFLOW = ((<size_t>1) << (4*sizeof(size_t)))
 *     if a >= MUL_NO_OVERFLOW or b >= MUL_NO_OVERFLOW:
 *         if unlikely(b > (<size_t>-1) // a):             # <<<<<<<<<<<<<<
 *             return <size_t>(-1)
 *     return a*b
 */
    }

    /* "cysignals/memory.pxd":72
 *     # If a and b both less than MUL_NO_OVERFLOW, no overflow can occur
 *     cdef size_t MUL_NO_OVERFLOW = ((<size_t>1) << (4*sizeof(size_t)))
 *     if a >= MUL_NO_OVERFLOW or b >= MUL_NO_OVERFLOW:             # <<<<<<<<<<<<<<
 *         if unlikely(b > (<size_t>-1) // a):
 *             return <size_t>(-1)
 */
  }

  /* "cysignals/memory.pxd":75
 *         if unlikely(b > (<size_t>-1) // a):
 *             return <size_t>(-1)
 *     return a*b             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_a * __pyx_v_b);
  goto __pyx_L0;

  /* "cysignals/memory.pxd":64
 * 
 * @cython.cdivision(True)
 * cdef inline size_t mul_overflowcheck(size_t a, size_t b) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Return a*b, checking for overflow. Assume that a > 0.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cysignals/memory.pxd":78
 * 
 * 
 * cdef inline void* check_allocarray(size_t nmemb, size_t size) except? NULL:             # <<<<<<<<<<<<<<
 *     """
 *     Allocate memory for ``nmemb`` elements of size ``size``.
 */

static CYTHON_INLINE void *__pyx_f_9cysignals_6memory_check_allocarray(size_t __pyx_v_nmemb, size_t __pyx_v_size) {
  size_t __pyx_v_n;
  void *__pyx_v_ret;
  void *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("check_allocarray", 0);

  /* "cysignals/memory.pxd":82
 *     Allocate memory for ``nmemb`` elements of size ``size``.
 *     """
 *     if nmemb == 0:             # <<<<<<<<<<<<<<
 *         return NULL
 *     cdef size_t n = mul_overflowcheck(nmemb, size)
 */
  __pyx_t_1 = ((__pyx_v_nmemb == 0) != 0);
  if (__pyx_t_1) {

    /* "cysignals/memory.pxd":83
 *     """
 *     if nmemb == 0:
 *         return NULL             # <<<<<<<<<<<<<<
 *     cdef size_t n = mul_overflowcheck(nmemb, size)
 *     cdef void* ret = sig_malloc(n)
 */
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "cysignals/memory.pxd":82
 *     Allocate memory for ``nmemb`` elements of size ``size``.
 *     """
 *     if nmemb == 0:             # <<<<<<<<<<<<<<
 *         return NULL
 *     cdef size_t n = mul_overflowcheck(nmemb, size)
 */
  }

  /* "cysignals/memory.pxd":84
 *     if nmemb == 0:
 *         return NULL
 *     cdef size_t n = mul_overflowcheck(nmemb, size)             # <<<<<<<<<<<<<<
 *     cdef void* ret = sig_malloc(n)
 *     if unlikely(ret == NULL):
 */
  __pyx_v_n = __pyx_f_9cysignals_6memory_mul_overflowcheck(__pyx_v_nmemb, __pyx_v_size);

  /* "cysignals/memory.pxd":85
 *         return NULL
 *     cdef size_t n = mul_overflowcheck(nmemb, size)
 *     cdef void* ret = sig_malloc(n)             # <<<<<<<<<<<<<<
 *     if unlikely(ret == NULL):
 *         raise MemoryError("failed to allocate %s * %s bytes" % (nmemb, size))
 */
  __pyx_v_ret = sig_malloc(__pyx_v_n);

  /* "cysignals/memory.pxd":86
 *     cdef size_t n = mul_overflowcheck(nmemb, size)
 *     cdef void* ret = sig_malloc(n)
 *     if unlikely(ret == NULL):             # <<<<<<<<<<<<<<
 *         raise MemoryError("failed to allocate %s * %s bytes" % (nmemb, size))
 *     return ret
 */
  __pyx_t_1 = (unlikely((__pyx_v_ret == NULL)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "cysignals/memory.pxd":87
 *     cdef void* ret = sig_malloc(n)
 *     if unlikely(ret == NULL):
 *         raise MemoryError("failed to allocate %s * %s bytes" % (nmemb, size))             # <<<<<<<<<<<<<<
 *     return ret
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_nmemb); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_failed_to_allocate_s_s_bytes, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_MemoryError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(2, 87, __pyx_L1_error)

    /* "cysignals/memory.pxd":86
 *     cdef size_t n = mul_overflowcheck(nmemb, size)
 *     cdef void* ret = sig_malloc(n)
 *     if unlikely(ret == NULL):             # <<<<<<<<<<<<<<
 *         raise MemoryError("failed to allocate %s * %s bytes" % (nmemb, size))
 *     return ret
 */
  }

  /* "cysignals/memory.pxd":88
 *     if unlikely(ret == NULL):
 *         raise MemoryError("failed to allocate %s * %s bytes" % (nmemb, size))
 *     return ret             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "cysignals/memory.pxd":78
 * 
 * 
 * cdef inline void* check_allocarray(size_t nmemb, size_t size) except? NULL:             # <<<<<<<<<<<<<<
 *     """
 *     Allocate memory for ``nmemb`` elements of size ``size``.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cysignals.memory.check_allocarray", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cysignals/memory.pxd":91
 * 
 * 
 * cdef inline void* check_reallocarray(void* ptr, size_t nmemb, size_t size) except? NULL:             # <<<<<<<<<<<<<<
 *     """
 *     Re-allocate memory at ``ptr`` to hold ``nmemb`` elements of size
 */

static CYTHON_INLINE void *__pyx_f_9cysignals_6memory_check_reallocarray(void *__pyx_v_ptr, size_t __pyx_v_nmemb, size_t __pyx_v_size) {
  size_t __pyx_v_n;
  void *__pyx_v_ret;
  void *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("check_reallocarray", 0);

  /* "cysignals/memory.pxd":99
 *     When ``nmemb`` equals 0, then free the memory at ``ptr``.
 *     """
 *     if nmemb == 0:             # <<<<<<<<<<<<<<
 *         sig_free(ptr)
 *         return NULL
 */
  __pyx_t_1 = ((__pyx_v_nmemb == 0) != 0);
  if (__pyx_t_1) {

    /* "cysignals/memory.pxd":100
 *     """
 *     if nmemb == 0:
 *         sig_free(ptr)             # <<<<<<<<<<<<<<
 *         return NULL
 *     cdef size_t n = mul_overflowcheck(nmemb, size)
 */
    sig_free(__pyx_v_ptr);

    /* "cysignals/memory.pxd":101
 *     if nmemb == 0:
 *         sig_free(ptr)
 *         return NULL             # <<<<<<<<<<<<<<
 *     cdef size_t n = mul_overflowcheck(nmemb, size)
 *     cdef void* ret = sig_realloc(ptr, n)
 */
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "cysignals/memory.pxd":99
 *     When ``nmemb`` equals 0, then free the memory at ``ptr``.
 *     """
 *     if nmemb == 0:             # <<<<<<<<<<<<<<
 *         sig_free(ptr)
 *         return NULL
 */
  }

  /* "cysignals/memory.pxd":102
 *         sig_free(ptr)
 *         return NULL
 *     cdef size_t n = mul_overflowcheck(nmemb, size)             # <<<<<<<<<<<<<<
 *     cdef void* ret = sig_realloc(ptr, n)
 *     if unlikely(ret == NULL):
 */
  __pyx_v_n = __pyx_f_9cysignals_6memory_mul_overflowcheck(__pyx_v_nmemb, __pyx_v_size);

  /* "cysignals/memory.pxd":103
 *         return NULL
 *     cdef size_t n = mul_overflowcheck(nmemb, size)
 *     cdef void* ret = sig_realloc(ptr, n)             # <<<<<<<<<<<<<<
 *     if unlikely(ret == NULL):
 *         raise MemoryError("failed to allocate %s * %s bytes" % (nmemb, size))
 */
  __pyx_v_ret = sig_realloc(__pyx_v_ptr, __pyx_v_n);

  /* "cysignals/memory.pxd":104
 *     cdef size_t n = mul_overflowcheck(nmemb, size)
 *     cdef void* ret = sig_realloc(ptr, n)
 *     if unlikely(ret == NULL):             # <<<<<<<<<<<<<<
 *         raise MemoryError("failed to allocate %s * %s bytes" % (nmemb, size))
 *     return ret
 */
  __pyx_t_1 = (unlikely((__pyx_v_ret == NULL)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "cysignals/memory.pxd":105
 *     cdef void* ret = sig_realloc(ptr, n)
 *     if unlikely(ret == NULL):
 *         raise MemoryError("failed to allocate %s * %s bytes" % (nmemb, size))             # <<<<<<<<<<<<<<
 *     return ret
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_nmemb); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_failed_to_allocate_s_s_bytes, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_MemoryError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(2, 105, __pyx_L1_error)

    /* "cysignals/memory.pxd":104
 *     cdef size_t n = mul_overflowcheck(nmemb, size)
 *     cdef void* ret = sig_realloc(ptr, n)
 *     if unlikely(ret == NULL):             # <<<<<<<<<<<<<<
 *         raise MemoryError("failed to allocate %s * %s bytes" % (nmemb, size))
 *     return ret
 */
  }

  /* "cysignals/memory.pxd":106
 *     if unlikely(ret == NULL):
 *         raise MemoryError("failed to allocate %s * %s bytes" % (nmemb, size))
 *     return ret             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "cysignals/memory.pxd":91
 * 
 * 
 * cdef inline void* check_reallocarray(void* ptr, size_t nmemb, size_t size) except? NULL:             # <<<<<<<<<<<<<<
 *     """
 *     Re-allocate memory at ``ptr`` to hold ``nmemb`` elements of size
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cysignals.memory.check_reallocarray", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cysignals/memory.pxd":109
 * 
 * 
 * cdef inline void* check_malloc(size_t n) except? NULL:             # <<<<<<<<<<<<<<
 *     """
 *     Allocate ``n`` bytes of memory.
 */

static CYTHON_INLINE void *__pyx_f_9cysignals_6memory_check_malloc(size_t __pyx_v_n) {
  void *__pyx_v_ret;
  void *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("check_malloc", 0);

  /* "cysignals/memory.pxd":113
 *     Allocate ``n`` bytes of memory.
 *     """
 *     if n == 0:             # <<<<<<<<<<<<<<
 *         return NULL
 *     cdef void* ret = sig_malloc(n)
 */
  __pyx_t_1 = ((__pyx_v_n == 0) != 0);
  if (__pyx_t_1) {

    /* "cysignals/memory.pxd":114
 *     """
 *     if n == 0:
 *         return NULL             # <<<<<<<<<<<<<<
 *     cdef void* ret = sig_malloc(n)
 *     if unlikely(ret == NULL):
 */
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "cysignals/memory.pxd":113
 *     Allocate ``n`` bytes of memory.
 *     """
 *     if n == 0:             # <<<<<<<<<<<<<<
 *         return NULL
 *     cdef void* ret = sig_malloc(n)
 */
  }

  /* "cysignals/memory.pxd":115
 *     if n == 0:
 *         return NULL
 *     cdef void* ret = sig_malloc(n)             # <<<<<<<<<<<<<<
 *     if unlikely(ret == NULL):
 *         raise MemoryError("failed to allocate %s bytes" % n)
 */
  __pyx_v_ret = sig_malloc(__pyx_v_n);

  /* "cysignals/memory.pxd":116
 *         return NULL
 *     cdef void* ret = sig_malloc(n)
 *     if unlikely(ret == NULL):             # <<<<<<<<<<<<<<
 *         raise MemoryError("failed to allocate %s bytes" % n)
 *     return ret
 */
  __pyx_t_1 = (unlikely((__pyx_v_ret == NULL)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "cysignals/memory.pxd":117
 *     cdef void* ret = sig_malloc(n)
 *     if unlikely(ret == NULL):
 *         raise MemoryError("failed to allocate %s bytes" % n)             # <<<<<<<<<<<<<<
 *     return ret
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_n); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_failed_to_allocate_s_bytes, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_MemoryError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(2, 117, __pyx_L1_error)

    /* "cysignals/memory.pxd":116
 *         return NULL
 *     cdef void* ret = sig_malloc(n)
 *     if unlikely(ret == NULL):             # <<<<<<<<<<<<<<
 *         raise MemoryError("failed to allocate %s bytes" % n)
 *     return ret
 */
  }

  /* "cysignals/memory.pxd":118
 *     if unlikely(ret == NULL):
 *         raise MemoryError("failed to allocate %s bytes" % n)
 *     return ret             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "cysignals/memory.pxd":109
 * 
 * 
 * cdef inline void* check_malloc(size_t n) except? NULL:             # <<<<<<<<<<<<<<
 *     """
 *     Allocate ``n`` bytes of memory.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cysignals.memory.check_malloc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cysignals/memory.pxd":121
 * 
 * 
 * cdef inline void* check_realloc(void* ptr, size_t n) except? NULL:             # <<<<<<<<<<<<<<
 *     """
 *     Re-allocate memory at ``ptr`` to hold ``n`` bytes.
 */

static CYTHON_INLINE void *__pyx_f_9cysignals_6memory_check_realloc(void *__pyx_v_ptr, size_t __pyx_v_n) {
  void *__pyx_v_ret;
  void *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("check_realloc", 0);

  /* "cysignals/memory.pxd":126
 *     If ``ptr`` equals ``NULL``, this behaves as ``check_malloc``.
 *     """
 *     if n == 0:             # <<<<<<<<<<<<<<
 *         sig_free(ptr)
 *         return NULL
 */
  __pyx_t_1 = ((__pyx_v_n == 0) != 0);
  if (__pyx_t_1) {

    /* "cysignals/memory.pxd":127
 *     """
 *     if n == 0:
 *         sig_free(ptr)             # <<<<<<<<<<<<<<
 *         return NULL
 *     cdef void* ret = sig_realloc(ptr, n)
 */
    sig_free(__pyx_v_ptr);

    /* "cysignals/memory.pxd":128
 *     if n == 0:
 *         sig_free(ptr)
 *         return NULL             # <<<<<<<<<<<<<<
 *     cdef void* ret = sig_realloc(ptr, n)
 *     if unlikely(ret == NULL):
 */
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "cysignals/memory.pxd":126
 *     If ``ptr`` equals ``NULL``, this behaves as ``check_malloc``.
 *     """
 *     if n == 0:             # <<<<<<<<<<<<<<
 *         sig_free(ptr)
 *         return NULL
 */
  }

  /* "cysignals/memory.pxd":129
 *         sig_free(ptr)
 *         return NULL
 *     cdef void* ret = sig_realloc(ptr, n)             # <<<<<<<<<<<<<<
 *     if unlikely(ret == NULL):
 *         raise MemoryError("failed to allocate %s bytes" % n)
 */
  __pyx_v_ret = sig_realloc(__pyx_v_ptr, __pyx_v_n);

  /* "cysignals/memory.pxd":130
 *         return NULL
 *     cdef void* ret = sig_realloc(ptr, n)
 *     if unlikely(ret == NULL):             # <<<<<<<<<<<<<<
 *         raise MemoryError("failed to allocate %s bytes" % n)
 *     return ret
 */
  __pyx_t_1 = (unlikely((__pyx_v_ret == NULL)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "cysignals/memory.pxd":131
 *     cdef void* ret = sig_realloc(ptr, n)
 *     if unlikely(ret == NULL):
 *         raise MemoryError("failed to allocate %s bytes" % n)             # <<<<<<<<<<<<<<
 *     return ret
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_n); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_failed_to_allocate_s_bytes, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_MemoryError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(2, 131, __pyx_L1_error)

    /* "cysignals/memory.pxd":130
 *         return NULL
 *     cdef void* ret = sig_realloc(ptr, n)
 *     if unlikely(ret == NULL):             # <<<<<<<<<<<<<<
 *         raise MemoryError("failed to allocate %s bytes" % n)
 *     return ret
 */
  }

  /* "cysignals/memory.pxd":132
 *     if unlikely(ret == NULL):
 *         raise MemoryError("failed to allocate %s bytes" % n)
 *     return ret             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "cysignals/memory.pxd":121
 * 
 * 
 * cdef inline void* check_realloc(void* ptr, size_t n) except? NULL:             # <<<<<<<<<<<<<<
 *     """
 *     Re-allocate memory at ``ptr`` to hold ``n`` bytes.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cysignals.memory.check_realloc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cysignals/memory.pxd":135
 * 
 * 
 * cdef inline void* check_calloc(size_t nmemb, size_t size) except? NULL:             # <<<<<<<<<<<<<<
 *     """
 *     Allocate memory for ``nmemb`` elements of size ``size``. The
 */

static CYTHON_INLINE void *__pyx_f_9cysignals_6memory_check_calloc(size_t __pyx_v_nmemb, size_t __pyx_v_size) {
  void *__pyx_v_ret;
  void *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("check_calloc", 0);

  /* "cysignals/memory.pxd":140
 *     resulting memory is zeroed.
 *     """
 *     if nmemb == 0:             # <<<<<<<<<<<<<<
 *         return NULL
 *     cdef void* ret = sig_calloc(nmemb, size)
 */
  __pyx_t_1 = ((__pyx_v_nmemb == 0) != 0);
  if (__pyx_t_1) {

    /* "cysignals/memory.pxd":141
 *     """
 *     if nmemb == 0:
 *         return NULL             # <<<<<<<<<<<<<<
 *     cdef void* ret = sig_calloc(nmemb, size)
 *     if unlikely(ret == NULL):
 */
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "cysignals/memory.pxd":140
 *     resulting memory is zeroed.
 *     """
 *     if nmemb == 0:             # <<<<<<<<<<<<<<
 *         return NULL
 *     cdef void* ret = sig_calloc(nmemb, size)
 */
  }

  /* "cysignals/memory.pxd":142
 *     if nmemb == 0:
 *         return NULL
 *     cdef void* ret = sig_calloc(nmemb, size)             # <<<<<<<<<<<<<<
 *     if unlikely(ret == NULL):
 *         raise MemoryError("failed to allocate %s * %s bytes" % (nmemb, size))
 */
  __pyx_v_ret = sig_calloc(__pyx_v_nmemb, __pyx_v_size);

  /* "cysignals/memory.pxd":143
 *         return NULL
 *     cdef void* ret = sig_calloc(nmemb, size)
 *     if unlikely(ret == NULL):             # <<<<<<<<<<<<<<
 *         raise MemoryError("failed to allocate %s * %s bytes" % (nmemb, size))
 *     return ret
 */
  __pyx_t_1 = (unlikely((__pyx_v_ret == NULL)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "cysignals/memory.pxd":144
 *     cdef void* ret = sig_calloc(nmemb, size)
 *     if unlikely(ret == NULL):
 *         raise MemoryError("failed to allocate %s * %s bytes" % (nmemb, size))             # <<<<<<<<<<<<<<
 *     return ret
 */
    __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_v_nmemb); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_failed_to_allocate_s_s_bytes, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_MemoryError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(2, 144, __pyx_L1_error)

    /* "cysignals/memory.pxd":143
 *         return NULL
 *     cdef void* ret = sig_calloc(nmemb, size)
 *     if unlikely(ret == NULL):             # <<<<<<<<<<<<<<
 *         raise MemoryError("failed to allocate %s * %s bytes" % (nmemb, size))
 *     return ret
 */
  }

  /* "cysignals/memory.pxd":145
 *     if unlikely(ret == NULL):
 *         raise MemoryError("failed to allocate %s * %s bytes" % (nmemb, size))
 *     return ret             # <<<<<<<<<<<<<<
 */
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "cysignals/memory.pxd":135
 * 
 * 
 * cdef inline void* check_calloc(size_t nmemb, size_t size) except? NULL:             # <<<<<<<<<<<<<<
 *     """
 *     Allocate memory for ``nmemb`` elements of size ``size``. The
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cysignals.memory.check_calloc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sage/ext/stdsage.pxd":16
 * 
 * 
 * cdef inline PY_NEW(type t):             # <<<<<<<<<<<<<<
 *     """
 *     Return ``t.__new__(t)``.  This works even for types like
 */

static CYTHON_INLINE PyObject *__pyx_f_4sage_3ext_7stdsage_PY_NEW(PyTypeObject *__pyx_v_t) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PY_NEW", 0);

  /* "sage/ext/stdsage.pxd":22
 *     optimizations assume that ``tp_new`` doesn't change).
 *     """
 *     return (<PyTypeObject*>t).tp_new(t, <object>NULL, <object>NULL)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyTypeObject *)__pyx_v_t)->tp_new(((PyTypeObject *)__pyx_v_t), ((PyObject *)NULL), ((PyObject *)NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "sage/ext/stdsage.pxd":16
 * 
 * 
 * cdef inline PY_NEW(type t):             # <<<<<<<<<<<<<<
 *     """
 *     Return ``t.__new__(t)``.  This works even for types like
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sage.ext.stdsage.PY_NEW", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sage/ext/stdsage.pxd":25
 * 
 * 
 * cdef inline void PY_SET_TP_NEW(type dst, type src):             # <<<<<<<<<<<<<<
 *     """
 *     Manually set ``dst.__new__`` to ``src.__new__``.  This is used to
 */

static CYTHON_INLINE void __pyx_f_4sage_3ext_7stdsage_PY_SET_TP_NEW(PyTypeObject *__pyx_v_dst, PyTypeObject *__pyx_v_src) {
  __Pyx_RefNannyDeclarations
  newfunc __pyx_t_1;
  __Pyx_RefNannySetupContext("PY_SET_TP_NEW", 0);

  /* "sage/ext/stdsage.pxd":31
 *     irrelevant base class ``tp_new`` methods.
 *     """
 *     (<PyTypeObject*>dst).tp_new = (<PyTypeObject*>src).tp_new             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = ((PyTypeObject *)__pyx_v_src)->tp_new;
  ((PyTypeObject *)__pyx_v_dst)->tp_new = __pyx_t_1;

  /* "sage/ext/stdsage.pxd":25
 * 
 * 
 * cdef inline void PY_SET_TP_NEW(type dst, type src):             # <<<<<<<<<<<<<<
 *     """
 *     Manually set ``dst.__new__`` to ``src.__new__``.  This is used to
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "sage/ext/stdsage.pxd":34
 * 
 * 
 * cdef inline bint HAS_DICTIONARY(obj):             # <<<<<<<<<<<<<<
 *     """
 *     Test whether the given object has a Python dictionary.
 */

static CYTHON_INLINE int __pyx_f_4sage_3ext_7stdsage_HAS_DICTIONARY(PyObject *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("HAS_DICTIONARY", 0);

  /* "sage/ext/stdsage.pxd":38
 *     Test whether the given object has a Python dictionary.
 *     """
 *     return Py_TYPE(obj).tp_dictoffset != 0             # <<<<<<<<<<<<<<
 */
  __pyx_r = (Py_TYPE(__pyx_v_obj)->tp_dictoffset != 0);
  goto __pyx_L0;

  /* "sage/ext/stdsage.pxd":34
 * 
 * 
 * cdef inline bint HAS_DICTIONARY(obj):             # <<<<<<<<<<<<<<
 *     """
 *     Test whether the given object has a Python dictionary.
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "parent.pxd":23
 *     # Flags, see below
 *     cdef int flags
 *     cdef inline bint get_flag(self, int flag):             # <<<<<<<<<<<<<<
 *         return self.flags & flag
 * 
 */

static CYTHON_INLINE int __pyx_f_4sage_9structure_6parent_6Parent_get_flag(struct __pyx_obj_4sage_9structure_6parent_Parent *__pyx_v_self, int __pyx_v_flag) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_flag", 0);

  /* "parent.pxd":24
 *     cdef int flags
 *     cdef inline bint get_flag(self, int flag):
 *         return self.flags & flag             # <<<<<<<<<<<<<<
 * 
 *     cpdef register_coercion(self, mor)
 */
  __pyx_r = (__pyx_v_self->flags & __pyx_v_flag);
  goto __pyx_L0;

  /* "parent.pxd":23
 *     # Flags, see below
 *     cdef int flags
 *     cdef inline bint get_flag(self, int flag):             # <<<<<<<<<<<<<<
 *         return self.flags & flag
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sage/structure/element.pxd":6
 * 
 * 
 * cpdef inline parent(x):             # <<<<<<<<<<<<<<
 *     """
 *     Return the parent of the element ``x``.
 */

static PyObject *__pyx_pw_4sage_9structure_7element_1parent(PyObject *__pyx_self, PyObject *__pyx_v_x); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4sage_9structure_7element_parent(PyObject *__pyx_v_x, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("parent", 0);

  /* "sage/structure/element.pxd":59
 *         <... 'list'>
 *     """
 *     if isinstance(x, Element):             # <<<<<<<<<<<<<<
 *         return (<Element>x)._parent
 *     return type(x)
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_x, __pyx_ptype_4sage_9structure_7element_Element); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "sage/structure/element.pxd":60
 *     """
 *     if isinstance(x, Element):
 *         return (<Element>x)._parent             # <<<<<<<<<<<<<<
 *     return type(x)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)((struct __pyx_obj_4sage_9structure_7element_Element *)__pyx_v_x)->_parent));
    __pyx_r = ((PyObject *)((struct __pyx_obj_4sage_9structure_7element_Element *)__pyx_v_x)->_parent);
    goto __pyx_L0;

    /* "sage/structure/element.pxd":59
 *         <... 'list'>
 *     """
 *     if isinstance(x, Element):             # <<<<<<<<<<<<<<
 *         return (<Element>x)._parent
 *     return type(x)
 */
  }

  /* "sage/structure/element.pxd":61
 *     if isinstance(x, Element):
 *         return (<Element>x)._parent
 *     return type(x)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_x)));
  __pyx_r = ((PyObject *)Py_TYPE(__pyx_v_x));
  goto __pyx_L0;

  /* "sage/structure/element.pxd":6
 * 
 * 
 * cpdef inline parent(x):             # <<<<<<<<<<<<<<
 *     """
 *     Return the parent of the element ``x``.
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4sage_9structure_7element_1parent(PyObject *__pyx_self, PyObject *__pyx_v_x); /*proto*/
static char __pyx_doc_4sage_9structure_7element_parent[] = "\n    Return the parent of the element ``x``.\n\n    Usually, this means the mathematical object of which ``x`` is an\n    element.\n\n    INPUT:\n\n    - ``x`` -- an element\n\n    OUTPUT:\n\n    - If ``x`` is a Sage :class:`Element`, return ``x.parent()``.\n\n    - Otherwise, return ``type(x)``.\n\n    .. SEEALSO::\n\n        `Parents, Conversion and Coercion <http://doc.sagemath.org/html/en/tutorial/tour_coercion.html>`_\n        Section in the Sage Tutorial\n\n    EXAMPLES::\n\n        sage: a = 42\n        sage: parent(a)\n        Integer Ring\n        sage: b = 42/1\n        sage: parent(b)\n        Rational Field\n        sage: c = 42.0\n        sage: parent(c)\n        Real Field with 53 bits of precision\n\n    Some more complicated examples::\n\n        sage: x = Partition([3,2,1,1,1])\n        sage: parent(x)\n        Partitions\n        sage: v = vector(RDF, [1,2,3])\n        sage: parent(v)\n        Vector space of dimension 3 over Real Double Field\n\n    The following are not considered to be elements, so the type is\n    returned::\n\n        sage: d = int(42)  # Python int\n        sage: parent(d)\n        <... 'int'>\n        sage: L = list(range(10))\n        sage: parent(L)\n        <... 'list'>\n    ";
static PyObject *__pyx_pw_4sage_9structure_7element_1parent(PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("parent (wrapper)", 0);
  __pyx_r = __pyx_pf_4sage_9structure_7element_parent(__pyx_self, ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4sage_9structure_7element_parent(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("parent", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4sage_9structure_7element_parent(__pyx_v_x, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sage.structure.element.parent", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sage/structure/element.pxd":64
 * 
 * 
 * cdef inline int classify_elements(left, right):             # <<<<<<<<<<<<<<
 *     """
 *     Given two objects, at least one which is an :class:`Element`,
 */

static CYTHON_INLINE int __pyx_f_4sage_9structure_7element_classify_elements(PyObject *__pyx_v_left, PyObject *__pyx_v_right) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("classify_elements", 0);

  /* "sage/structure/element.pxd":87
 *     - 0o37: both are Element, same type, same parent
 *     """
 *     if type(left) is type(right):             # <<<<<<<<<<<<<<
 *         # We know at least one of the arguments is an Element. So if
 *         # their types are *equal* (fast to check) then they are both
 */
  __pyx_t_1 = (((PyObject *)Py_TYPE(__pyx_v_left)) == ((PyObject *)Py_TYPE(__pyx_v_right)));
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "sage/structure/element.pxd":91
 *         # their types are *equal* (fast to check) then they are both
 *         # Elements.
 *         if (<Element>left)._parent is (<Element>right)._parent:             # <<<<<<<<<<<<<<
 *             return 0o37
 *         else:
 */
    __pyx_t_2 = (((struct __pyx_obj_4sage_9structure_7element_Element *)__pyx_v_left)->_parent == ((struct __pyx_obj_4sage_9structure_7element_Element *)__pyx_v_right)->_parent);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "sage/structure/element.pxd":92
 *         # Elements.
 *         if (<Element>left)._parent is (<Element>right)._parent:
 *             return 0o37             # <<<<<<<<<<<<<<
 *         else:
 *             return 0o17
 */
      __pyx_r = 037;
      goto __pyx_L0;

      /* "sage/structure/element.pxd":91
 *         # their types are *equal* (fast to check) then they are both
 *         # Elements.
 *         if (<Element>left)._parent is (<Element>right)._parent:             # <<<<<<<<<<<<<<
 *             return 0o37
 *         else:
 */
    }

    /* "sage/structure/element.pxd":94
 *             return 0o37
 *         else:
 *             return 0o17             # <<<<<<<<<<<<<<
 *     if not isinstance(right, Element):
 *         return 0o01
 */
    /*else*/ {
      __pyx_r = 017;
      goto __pyx_L0;
    }

    /* "sage/structure/element.pxd":87
 *     - 0o37: both are Element, same type, same parent
 *     """
 *     if type(left) is type(right):             # <<<<<<<<<<<<<<
 *         # We know at least one of the arguments is an Element. So if
 *         # their types are *equal* (fast to check) then they are both
 */
  }

  /* "sage/structure/element.pxd":95
 *         else:
 *             return 0o17
 *     if not isinstance(right, Element):             # <<<<<<<<<<<<<<
 *         return 0o01
 *     if not isinstance(left, Element):
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_right, __pyx_ptype_4sage_9structure_7element_Element); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "sage/structure/element.pxd":96
 *             return 0o17
 *     if not isinstance(right, Element):
 *         return 0o01             # <<<<<<<<<<<<<<
 *     if not isinstance(left, Element):
 *         return 0o02
 */
    __pyx_r = 001;
    goto __pyx_L0;

    /* "sage/structure/element.pxd":95
 *         else:
 *             return 0o17
 *     if not isinstance(right, Element):             # <<<<<<<<<<<<<<
 *         return 0o01
 *     if not isinstance(left, Element):
 */
  }

  /* "sage/structure/element.pxd":97
 *     if not isinstance(right, Element):
 *         return 0o01
 *     if not isinstance(left, Element):             # <<<<<<<<<<<<<<
 *         return 0o02
 *     if (<Element>left)._parent is (<Element>right)._parent:
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_left, __pyx_ptype_4sage_9structure_7element_Element); 
  __pyx_t_1 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_1) {

    /* "sage/structure/element.pxd":98
 *         return 0o01
 *     if not isinstance(left, Element):
 *         return 0o02             # <<<<<<<<<<<<<<
 *     if (<Element>left)._parent is (<Element>right)._parent:
 *         return 0o27
 */
    __pyx_r = 002;
    goto __pyx_L0;

    /* "sage/structure/element.pxd":97
 *     if not isinstance(right, Element):
 *         return 0o01
 *     if not isinstance(left, Element):             # <<<<<<<<<<<<<<
 *         return 0o02
 *     if (<Element>left)._parent is (<Element>right)._parent:
 */
  }

  /* "sage/structure/element.pxd":99
 *     if not isinstance(left, Element):
 *         return 0o02
 *     if (<Element>left)._parent is (<Element>right)._parent:             # <<<<<<<<<<<<<<
 *         return 0o27
 *     else:
 */
  __pyx_t_1 = (((struct __pyx_obj_4sage_9structure_7element_Element *)__pyx_v_left)->_parent == ((struct __pyx_obj_4sage_9structure_7element_Element *)__pyx_v_right)->_parent);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "sage/structure/element.pxd":100
 *         return 0o02
 *     if (<Element>left)._parent is (<Element>right)._parent:
 *         return 0o27             # <<<<<<<<<<<<<<
 *     else:
 *         return 0o07
 */
    __pyx_r = 027;
    goto __pyx_L0;

    /* "sage/structure/element.pxd":99
 *     if not isinstance(left, Element):
 *         return 0o02
 *     if (<Element>left)._parent is (<Element>right)._parent:             # <<<<<<<<<<<<<<
 *         return 0o27
 *     else:
 */
  }

  /* "sage/structure/element.pxd":102
 *         return 0o27
 *     else:
 *         return 0o07             # <<<<<<<<<<<<<<
 * 
 * # Functions to help understand the result of classify_elements()
 */
  /*else*/ {
    __pyx_r = 007;
    goto __pyx_L0;
  }

  /* "sage/structure/element.pxd":64
 * 
 * 
 * cdef inline int classify_elements(left, right):             # <<<<<<<<<<<<<<
 *     """
 *     Given two objects, at least one which is an :class:`Element`,
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sage/structure/element.pxd":105
 * 
 * # Functions to help understand the result of classify_elements()
 * cdef inline bint BOTH_ARE_ELEMENT(int cl):             # <<<<<<<<<<<<<<
 *     return cl & 0o04
 * cdef inline bint HAVE_SAME_PARENT(int cl):
 */

static CYTHON_INLINE int __pyx_f_4sage_9structure_7element_BOTH_ARE_ELEMENT(int __pyx_v_cl) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("BOTH_ARE_ELEMENT", 0);

  /* "sage/structure/element.pxd":106
 * # Functions to help understand the result of classify_elements()
 * cdef inline bint BOTH_ARE_ELEMENT(int cl):
 *     return cl & 0o04             # <<<<<<<<<<<<<<
 * cdef inline bint HAVE_SAME_PARENT(int cl):
 *     return cl & 0o20
 */
  __pyx_r = (__pyx_v_cl & 004);
  goto __pyx_L0;

  /* "sage/structure/element.pxd":105
 * 
 * # Functions to help understand the result of classify_elements()
 * cdef inline bint BOTH_ARE_ELEMENT(int cl):             # <<<<<<<<<<<<<<
 *     return cl & 0o04
 * cdef inline bint HAVE_SAME_PARENT(int cl):
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sage/structure/element.pxd":107
 * cdef inline bint BOTH_ARE_ELEMENT(int cl):
 *     return cl & 0o04
 * cdef inline bint HAVE_SAME_PARENT(int cl):             # <<<<<<<<<<<<<<
 *     return cl & 0o20
 * 
 */

static CYTHON_INLINE int __pyx_f_4sage_9structure_7element_HAVE_SAME_PARENT(int __pyx_v_cl) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("HAVE_SAME_PARENT", 0);

  /* "sage/structure/element.pxd":108
 *     return cl & 0o04
 * cdef inline bint HAVE_SAME_PARENT(int cl):
 *     return cl & 0o20             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_cl & 020);
  goto __pyx_L0;

  /* "sage/structure/element.pxd":107
 * cdef inline bint BOTH_ARE_ELEMENT(int cl):
 *     return cl & 0o04
 * cdef inline bint HAVE_SAME_PARENT(int cl):             # <<<<<<<<<<<<<<
 *     return cl & 0o20
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "sage/structure/element.pxd":111
 * 
 * 
 * cpdef inline bint have_same_parent(left, right):             # <<<<<<<<<<<<<<
 *     """
 *     Return ``True`` if and only if ``left`` and ``right`` have the
 */

static PyObject *__pyx_pw_4sage_9structure_7element_3have_same_parent(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static CYTHON_INLINE int __pyx_f_4sage_9structure_7element_have_same_parent(PyObject *__pyx_v_left, PyObject *__pyx_v_right, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("have_same_parent", 0);

  /* "sage/structure/element.pxd":143
 *         True
 *     """
 *     return HAVE_SAME_PARENT(classify_elements(left, right))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_f_4sage_9structure_7element_HAVE_SAME_PARENT(__pyx_f_4sage_9structure_7element_classify_elements(__pyx_v_left, __pyx_v_right));
  goto __pyx_L0;

  /* "sage/structure/element.pxd":111
 * 
 * 
 * cpdef inline bint have_same_parent(left, right):             # <<<<<<<<<<<<<<
 *     """
 *     Return ``True`` if and only if ``left`` and ``right`` have the
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4sage_9structure_7element_3have_same_parent(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4sage_9structure_7element_2have_same_parent[] = "\n    Return ``True`` if and only if ``left`` and ``right`` have the\n    same parent.\n\n    .. WARNING::\n\n        This function assumes that at least one of the arguments is a\n        Sage :class:`Element`. When in doubt, use the slower\n        ``parent(left) is parent(right)`` instead.\n\n    EXAMPLES::\n\n        sage: from sage.structure.element import have_same_parent\n        sage: have_same_parent(1, 3)\n        True\n        sage: have_same_parent(1, 1/2)\n        False\n        sage: have_same_parent(gap(1), gap(1/2))\n        True\n\n    These have different types but the same parent::\n\n        sage: a = RLF(2)\n        sage: b = exp(a)\n        sage: type(a)\n        <... 'sage.rings.real_lazy.LazyWrapper'>\n        sage: type(b)\n        <... 'sage.rings.real_lazy.LazyNamedUnop'>\n        sage: have_same_parent(a, b)\n        True\n    ";
static PyObject *__pyx_pw_4sage_9structure_7element_3have_same_parent(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_left = 0;
  PyObject *__pyx_v_right = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("have_same_parent (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_left,&__pyx_n_s_right,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_left)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_right)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("have_same_parent", 1, 2, 2, 1); __PYX_ERR(4, 111, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "have_same_parent") < 0)) __PYX_ERR(4, 111, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_left = values[0];
    __pyx_v_right = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("have_same_parent", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(4, 111, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("sage.structure.element.have_same_parent", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4sage_9structure_7element_2have_same_parent(__pyx_self, __pyx_v_left, __pyx_v_right);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4sage_9structure_7element_2have_same_parent(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_left, PyObject *__pyx_v_right) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("have_same_parent", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_4sage_9structure_7element_have_same_parent(__pyx_v_left, __pyx_v_right, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("sage.structure.element.have_same_parent", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily __pyx_vtable_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily;

static PyObject *__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)o);
  p->__pyx_vtab = __pyx_vtabptr_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily;
  p->poly = ((struct __pyx_obj_3ppl_10polyhedron_C_Polyhedron *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily(PyObject *o) {
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *p = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->poly);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *p = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)o;
  if (p->poly) {
    e = (*v)(((PyObject *)p->poly), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *p = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)o;
  tmp = ((PyObject*)p->poly);
  p->poly = ((struct __pyx_obj_3ppl_10polyhedron_C_Polyhedron *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily[] = {
  {"ray_coefficient", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_7ray_coefficient, METH_VARARGS|METH_KEYWORDS, __pyx_doc_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_6ray_coefficient},
  {"rays", (PyCFunction)__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_9rays, METH_NOARGS, __pyx_doc_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_8rays},
  {"_debug_info", (PyCFunction)__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_13_debug_info, METH_NOARGS, 0},
  {"permutation", (PyCFunction)__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_15permutation, METH_NOARGS, __pyx_doc_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_14permutation},
  {"has_connection", (PyCFunction)__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_21has_connection, METH_NOARGS, __pyx_doc_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_20has_connection},
  {"children", (PyCFunction)__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_23children, METH_NOARGS, __pyx_doc_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_22children},
  {"tree", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_25tree, METH_VARARGS|METH_KEYWORDS, __pyx_doc_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_24tree},
  {"random_element", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_28random_element, METH_VARARGS|METH_KEYWORDS, __pyx_doc_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_27random_element},
  {"random_element_statistics", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_30random_element_statistics, METH_VARARGS|METH_KEYWORDS, __pyx_doc_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_29random_element_statistics},
  {"random_integer_point_statistics", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_32random_integer_point_statistics, METH_VARARGS|METH_KEYWORDS, __pyx_doc_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_31random_integer_point_statistics},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_34__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_36__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily = {
  PyVarObject_HEAD_INIT(0, 0)
  "surface_dynamics.interval_exchanges.iet_family.IETFamily", /*tp_name*/
  sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_11__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  __pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_17__hash__, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "\n    A linear family of interval exchange transformations\n\n    This class consists of two attributes:\n\n    - a permutation (must be labeled)\n\n    - a cone of lengths (a subcone of the positive cone)\n\n    EXAMPLES::\n\n        sage: from surface_dynamics import *\n\n        sage: p = iet.Permutation('a b c d', 'd c b a')\n        sage: F = iet.IETFamily(p, (ZZ**4).basis())    # optional - pplpy\n        sage: F  # optional - pplpy\n        top 0 1 2 3\n        bot 3 2 1 0\n        0 0 0 1\n        0 0 1 0\n        0 1 0 0\n        1 0 0 0\n\n        sage: iet.IETFamily(p, [(2,3,0,0), (0,1,1,1)]) # optional - pplpy\n        top 0 1 2 3\n        bot 3 2 1 0\n        0 1 1 1\n        2 3 0 0\n\n        sage: iet.IETFamily(p, [(1,0,0,0), (1,-1,1,1)]) # optional - pplpy\n        Traceback (most recent call last):\n        ...\n        ValueError: C must be a subcone of the non-negative cone\n        sage: iet.IETFamily(p, Polyhedron(vertices=[(0,0,0,0), (1,0,0,0),(0,1,0,0),(1,1,1,1)])) # optional - pplpy\n        Traceback (most recent call last):\n        ...\n        ValueError: should have only zero as vertices\n    ", /*tp_doc*/
  __pyx_tp_traverse_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily, /*tp_traverse*/
  __pyx_tp_clear_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily, /*tp_clear*/
  __pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_19__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_5__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__ *__pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__[8];
static int __pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__ = 0;

static PyObject *__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__)))) {
    o = (PyObject*)__pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__[--__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__];
    memset(o, 0, sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__(PyObject *o) {
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__ *p = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__)))) {
    __pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__[__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__++] = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__ *p = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__ *p = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__ *)o;
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "surface_dynamics.interval_exchanges.iet_family.__pyx_scope_struct____repr__", /*tp_name*/
  sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__, /*tp_traverse*/
  __pyx_tp_clear_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr *__pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr[8];
static int __pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr = 0;

static PyObject *__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr)))) {
    o = (PyObject*)__pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr[--__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr(PyObject *o) {
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr)))) {
    __pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr[__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "surface_dynamics.interval_exchanges.iet_family.__pyx_scope_struct_1_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr *__pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr[8];
static int __pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr = 0;

static PyObject *__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr)))) {
    o = (PyObject*)__pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr[--__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr(PyObject *o) {
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_i);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr)))) {
    __pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr[__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr++] = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_i) {
    e = (*v)(p->__pyx_v_i, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "surface_dynamics.interval_exchanges.iet_family.__pyx_scope_struct_2_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr *__pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr[8];
static int __pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr = 0;

static PyObject *__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr)))) {
    o = (PyObject*)__pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr[--__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr(PyObject *o) {
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr *p = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr)))) {
    __pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr[__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr++] = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr *p = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "surface_dynamics.interval_exchanges.iet_family.__pyx_scope_struct_3_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree *__pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree[8];
static int __pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree = 0;

static PyObject *__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree)))) {
    o = (PyObject*)__pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree[--__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree];
    memset(o, 0, sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree(PyObject *o) {
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree *p = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_branch);
  Py_CLEAR(p->__pyx_v_f);
  Py_CLEAR(p->__pyx_v_ff);
  Py_CLEAR(p->__pyx_v_max_depth);
  Py_CLEAR(p->__pyx_v_s);
  Py_CLEAR(p->__pyx_v_seen);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_ss);
  Py_CLEAR(p->__pyx_v_verbose);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree)))) {
    __pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree[__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree++] = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree *p = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree *)o;
  if (p->__pyx_v_branch) {
    e = (*v)(p->__pyx_v_branch, a); if (e) return e;
  }
  if (p->__pyx_v_f) {
    e = (*v)(p->__pyx_v_f, a); if (e) return e;
  }
  if (p->__pyx_v_ff) {
    e = (*v)(p->__pyx_v_ff, a); if (e) return e;
  }
  if (p->__pyx_v_max_depth) {
    e = (*v)(p->__pyx_v_max_depth, a); if (e) return e;
  }
  if (p->__pyx_v_s) {
    e = (*v)(p->__pyx_v_s, a); if (e) return e;
  }
  if (p->__pyx_v_seen) {
    e = (*v)(p->__pyx_v_seen, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_ss) {
    e = (*v)(p->__pyx_v_ss, a); if (e) return e;
  }
  if (p->__pyx_v_verbose) {
    e = (*v)(p->__pyx_v_verbose, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree = {
  PyVarObject_HEAD_INIT(0, 0)
  "surface_dynamics.interval_exchanges.iet_family.__pyx_scope_struct_4_tree", /*tp_name*/
  sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element *__pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element[8];
static int __pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element = 0;

static PyObject *__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element)))) {
    o = (PyObject*)__pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element[--__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element];
    memset(o, 0, sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element(PyObject *o) {
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element *p = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_coeffs);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element)))) {
    __pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element[__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element++] = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element *p = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element *)o;
  if (p->__pyx_v_coeffs) {
    e = (*v)(p->__pyx_v_coeffs, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element *p = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element *)o;
  tmp = ((PyObject*)p->__pyx_v_coeffs);
  p->__pyx_v_coeffs = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element = {
  PyVarObject_HEAD_INIT(0, 0)
  "surface_dynamics.interval_exchanges.iet_family.__pyx_scope_struct_5_random_element", /*tp_name*/
  sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element, /*tp_traverse*/
  __pyx_tp_clear_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr *__pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr[8];
static int __pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr = 0;

static PyObject *__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr)))) {
    o = (PyObject*)__pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr[--__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr(PyObject *o) {
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr *p = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_i);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr)))) {
    __pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr[__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr++] = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr *p = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_i) {
    e = (*v)(p->__pyx_v_i, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "surface_dynamics.interval_exchanges.iet_family.__pyx_scope_struct_6_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics *__pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics[8];
static int __pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics = 0;

static PyObject *__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics)))) {
    o = (PyObject*)__pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics[--__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics];
    memset(o, 0, sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics(PyObject *o) {
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics *p = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_ZZ);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics)))) {
    __pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics[__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics++] = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics *p = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics *)o;
  if (p->__pyx_v_ZZ) {
    e = (*v)(p->__pyx_v_ZZ, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics *p = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics *)o;
  tmp = ((PyObject*)p->__pyx_v_ZZ);
  p->__pyx_v_ZZ = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics = {
  PyVarObject_HEAD_INIT(0, 0)
  "surface_dynamics.interval_exchanges.iet_family.__pyx_scope_struct_7_random_integer_point_statistics", /*tp_name*/
  sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics, /*tp_traverse*/
  __pyx_tp_clear_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr *__pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr[8];
static int __pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr = 0;

static PyObject *__pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr)))) {
    o = (PyObject*)__pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr[--__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr(PyObject *o) {
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr *p = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_r);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr)))) {
    __pyx_freelist_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr[__pyx_freecount_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr++] = ((struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr *p = (struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_r) {
    e = (*v)(p->__pyx_v_r, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "surface_dynamics.interval_exchanges.iet_family.__pyx_scope_struct_8_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_iet_family(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_iet_family},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "iet_family",
    __pyx_k_Linear_families_of_interval_exc, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_C, __pyx_k_C, sizeof(__pyx_k_C), 0, 0, 1, 1},
  {&__pyx_n_s_FreeModule, __pyx_k_FreeModule, sizeof(__pyx_k_FreeModule), 0, 0, 1, 1},
  {&__pyx_n_s_IET, __pyx_k_IET, sizeof(__pyx_k_IET), 0, 0, 1, 1},
  {&__pyx_n_s_IETFamily, __pyx_k_IETFamily, sizeof(__pyx_k_IETFamily), 0, 0, 1, 1},
  {&__pyx_n_s_IETFamily_tree, __pyx_k_IETFamily_tree, sizeof(__pyx_k_IETFamily_tree), 0, 0, 1, 1},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_n_s_LabelledPermutationIET, __pyx_k_LabelledPermutationIET, sizeof(__pyx_k_LabelledPermutationIET), 0, 0, 1, 1},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_n_s_Permutation, __pyx_k_Permutation, sizeof(__pyx_k_Permutation), 0, 0, 1, 1},
  {&__pyx_n_s_QQ, __pyx_k_QQ, sizeof(__pyx_k_QQ), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_ZZ, __pyx_k_ZZ, sizeof(__pyx_k_ZZ), 0, 0, 1, 1},
  {&__pyx_kp_s__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 0, 1, 0},
  {&__pyx_kp_s__6, __pyx_k__6, sizeof(__pyx_k__6), 0, 0, 1, 0},
  {&__pyx_kp_s__8, __pyx_k__8, sizeof(__pyx_k__8), 0, 0, 1, 0},
  {&__pyx_n_s_abs, __pyx_k_abs, sizeof(__pyx_k_abs), 0, 0, 1, 1},
  {&__pyx_n_s_add_space_dimensions_and_embed, __pyx_k_add_space_dimensions_and_embed, sizeof(__pyx_k_add_space_dimensions_and_embed), 0, 0, 1, 1},
  {&__pyx_n_s_affine_dimension, __pyx_k_affine_dimension, sizeof(__pyx_k_affine_dimension), 0, 0, 1, 1},
  {&__pyx_n_s_affine_image, __pyx_k_affine_image, sizeof(__pyx_k_affine_image), 0, 0, 1, 1},
  {&__pyx_kp_s_alloc_lu, __pyx_k_alloc_lu, sizeof(__pyx_k_alloc_lu), 0, 0, 1, 0},
  {&__pyx_n_s_alphabet, __pyx_k_alphabet, sizeof(__pyx_k_alphabet), 0, 0, 1, 1},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_autosim, __pyx_k_autosim, sizeof(__pyx_k_autosim), 0, 0, 1, 1},
  {&__pyx_n_s_b, __pyx_k_b, sizeof(__pyx_k_b), 0, 0, 1, 1},
  {&__pyx_kp_s_bot, __pyx_k_bot, sizeof(__pyx_k_bot), 0, 0, 1, 0},
  {&__pyx_kp_s_branch, __pyx_k_branch, sizeof(__pyx_k_branch), 0, 0, 1, 0},
  {&__pyx_n_s_children, __pyx_k_children, sizeof(__pyx_k_children), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_kp_s_column_index_out_of_range, __pyx_k_column_index_out_of_range, sizeof(__pyx_k_column_index_out_of_range), 0, 0, 1, 0},
  {&__pyx_n_s_cone, __pyx_k_cone, sizeof(__pyx_k_cone), 0, 0, 1, 1},
  {&__pyx_n_s_constructors, __pyx_k_constructors, sizeof(__pyx_k_constructors), 0, 0, 1, 1},
  {&__pyx_kp_s_d_saddles_found_on_d_random_poi, __pyx_k_d_saddles_found_on_d_random_poi, sizeof(__pyx_k_d_saddles_found_on_d_random_poi), 0, 0, 1, 0},
  {&__pyx_kp_s_dim_lu, __pyx_k_dim_lu, sizeof(__pyx_k_dim_lu), 0, 0, 1, 0},
  {&__pyx_kp_s_dimension_mismatch, __pyx_k_dimension_mismatch, sizeof(__pyx_k_dimension_mismatch), 0, 0, 1, 0},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_kp_s_entries_lu, __pyx_k_entries_lu, sizeof(__pyx_k_entries_lu), 0, 0, 1, 0},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_kp_s_failed_to_allocate_s_bytes, __pyx_k_failed_to_allocate_s_bytes, sizeof(__pyx_k_failed_to_allocate_s_bytes), 0, 0, 1, 0},
  {&__pyx_kp_s_failed_to_allocate_s_s_bytes, __pyx_k_failed_to_allocate_s_s_bytes, sizeof(__pyx_k_failed_to_allocate_s_s_bytes), 0, 0, 1, 0},
  {&__pyx_n_s_file, __pyx_k_file, sizeof(__pyx_k_file), 0, 0, 1, 1},
  {&__pyx_n_s_flush, __pyx_k_flush, sizeof(__pyx_k_flush), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_has_connection, __pyx_k_has_connection, sizeof(__pyx_k_has_connection), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_iet_interface, __pyx_k_iet_interface, sizeof(__pyx_k_iet_interface), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_intersection_assign, __pyx_k_intersection_assign, sizeof(__pyx_k_intersection_assign), 0, 0, 1, 1},
  {&__pyx_n_s_is_zero, __pyx_k_is_zero, sizeof(__pyx_k_is_zero), 0, 0, 1, 1},
  {&__pyx_n_s_iterations, __pyx_k_iterations, sizeof(__pyx_k_iterations), 0, 0, 1, 1},
  {&__pyx_n_s_join, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
  {&__pyx_n_s_labels, __pyx_k_labels, sizeof(__pyx_k_labels), 0, 0, 1, 1},
  {&__pyx_n_s_left, __pyx_k_left, sizeof(__pyx_k_left), 0, 0, 1, 1},
  {&__pyx_kp_s_length_lu, __pyx_k_length_lu, sizeof(__pyx_k_length_lu), 0, 0, 1, 0},
  {&__pyx_kp_s_looking_children, __pyx_k_looking_children, sizeof(__pyx_k_looking_children), 0, 0, 1, 0},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_max_depth, __pyx_k_max_depth, sizeof(__pyx_k_max_depth), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_nray, __pyx_k_nray, sizeof(__pyx_k_nray), 0, 0, 1, 1},
  {&__pyx_n_s_num_exp, __pyx_k_num_exp, sizeof(__pyx_k_num_exp), 0, 0, 1, 1},
  {&__pyx_n_s_num_iterations, __pyx_k_num_iterations, sizeof(__pyx_k_num_iterations), 0, 0, 1, 1},
  {&__pyx_n_s_number_of_cylinders, __pyx_k_number_of_cylinders, sizeof(__pyx_k_number_of_cylinders), 0, 0, 1, 1},
  {&__pyx_kp_s_only_equality_and_difference_tes, __pyx_k_only_equality_and_difference_tes, sizeof(__pyx_k_only_equality_and_difference_tes), 0, 0, 1, 0},
  {&__pyx_n_s_p, __pyx_k_p, sizeof(__pyx_k_p), 0, 0, 1, 1},
  {&__pyx_kp_s_p_must_be_a_labelled_permutation, __pyx_k_p_must_be_a_labelled_permutation, sizeof(__pyx_k_p_must_be_a_labelled_permutation), 0, 0, 1, 0},
  {&__pyx_n_s_permutation, __pyx_k_permutation, sizeof(__pyx_k_permutation), 0, 0, 1, 1},
  {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
  {&__pyx_n_s_ppl_check_non_negative_cone, __pyx_k_ppl_check_non_negative_cone, sizeof(__pyx_k_ppl_check_non_negative_cone), 0, 0, 1, 1},
  {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_random_element, __pyx_k_random_element, sizeof(__pyx_k_random_element), 0, 0, 1, 1},
  {&__pyx_n_s_random_element_locals_genexpr, __pyx_k_random_element_locals_genexpr, sizeof(__pyx_k_random_element_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_random_integer_point_statistics, __pyx_k_random_integer_point_statistics, sizeof(__pyx_k_random_integer_point_statistics), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_ray_coefficient, __pyx_k_ray_coefficient, sizeof(__pyx_k_ray_coefficient), 0, 0, 1, 1},
  {&__pyx_n_s_rays, __pyx_k_rays, sizeof(__pyx_k_rays), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_repr___locals_genexpr, __pyx_k_repr___locals_genexpr, sizeof(__pyx_k_repr___locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_right, __pyx_k_right, sizeof(__pyx_k_right), 0, 0, 1, 1},
  {&__pyx_n_s_ring, __pyx_k_ring, sizeof(__pyx_k_ring), 0, 0, 1, 1},
  {&__pyx_kp_s_row_index_out_of_range, __pyx_k_row_index_out_of_range, sizeof(__pyx_k_row_index_out_of_range), 0, 0, 1, 0},
  {&__pyx_kp_s_rows_d_lu, __pyx_k_rows_d_lu, sizeof(__pyx_k_rows_d_lu), 0, 0, 1, 0},
  {&__pyx_kp_s_rows_lu, __pyx_k_rows_lu, sizeof(__pyx_k_rows_lu), 0, 0, 1, 0},
  {&__pyx_kp_s_s_f, __pyx_k_s_f, sizeof(__pyx_k_s_f), 0, 0, 1, 0},
  {&__pyx_n_s_saddle, __pyx_k_saddle, sizeof(__pyx_k_saddle), 0, 0, 1, 1},
  {&__pyx_n_s_sage_modules_free_module, __pyx_k_sage_modules_free_module, sizeof(__pyx_k_sage_modules_free_module), 0, 0, 1, 1},
  {&__pyx_n_s_sage_rings_integer_ring, __pyx_k_sage_rings_integer_ring, sizeof(__pyx_k_sage_rings_integer_ring), 0, 0, 1, 1},
  {&__pyx_n_s_sage_rings_rational_field, __pyx_k_sage_rings_rational_field, sizeof(__pyx_k_sage_rings_rational_field), 0, 0, 1, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_space_dimension, __pyx_k_space_dimension, sizeof(__pyx_k_space_dimension), 0, 0, 1, 1},
  {&__pyx_n_s_stdout, __pyx_k_stdout, sizeof(__pyx_k_stdout), 0, 0, 1, 1},
  {&__pyx_n_s_sum, __pyx_k_sum, sizeof(__pyx_k_sum), 0, 0, 1, 1},
  {&__pyx_n_s_surface_dynamics_interval_exchan, __pyx_k_surface_dynamics_interval_exchan, sizeof(__pyx_k_surface_dynamics_interval_exchan), 0, 0, 1, 1},
  {&__pyx_n_s_surface_dynamics_interval_exchan_2, __pyx_k_surface_dynamics_interval_exchan_2, sizeof(__pyx_k_surface_dynamics_interval_exchan_2), 0, 0, 1, 1},
  {&__pyx_n_s_surface_dynamics_misc_ppl_utils, __pyx_k_surface_dynamics_misc_ppl_utils, sizeof(__pyx_k_surface_dynamics_misc_ppl_utils), 0, 0, 1, 1},
  {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
  {&__pyx_n_s_t, __pyx_k_t, sizeof(__pyx_k_t), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_kp_s_top, __pyx_k_top, sizeof(__pyx_k_top), 0, 0, 1, 0},
  {&__pyx_n_s_tree, __pyx_k_tree, sizeof(__pyx_k_tree), 0, 0, 1, 1},
  {&__pyx_n_s_verbose, __pyx_k_verbose, sizeof(__pyx_k_verbose), 0, 0, 1, 1},
  {&__pyx_n_s_write, __pyx_k_write, sizeof(__pyx_k_write), 0, 0, 1, 1},
  {&__pyx_n_s_zorich_move, __pyx_k_zorich_move, sizeof(__pyx_k_zorich_move), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 85, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 207, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 219, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(0, 250, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 427, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(0, 656, __pyx_L1_error)
  __pyx_builtin_sum = __Pyx_GetBuiltinName(__pyx_n_s_sum); if (!__pyx_builtin_sum) __PYX_ERR(0, 703, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(2, 87, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":207
 *         from surface_dynamics.interval_exchanges.labelled import LabelledPermutationIET
 *         if not isinstance(p, LabelledPermutationIET):
 *             raise ValueError('p must be a labelled permutation')             # <<<<<<<<<<<<<<
 * 
 *         from surface_dynamics.misc.ppl_utils import ppl_check_non_negative_cone
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_p_must_be_a_labelled_permutation); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":212
 *         ppl_check_non_negative_cone(C)
 *         if C.space_dimension() != len(p):
 *             raise ValueError('dimension mismatch')             # <<<<<<<<<<<<<<
 * 
 *         # fill C data from input
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_dimension_mismatch); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":250
 *         """
 *         if nray >= self.length:
 *             raise IndexError('row index out of range')             # <<<<<<<<<<<<<<
 *         if i >= self.dim:
 *             raise IndexError('column index out of range')
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_row_index_out_of_range); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":252
 *             raise IndexError('row index out of range')
 *         if i >= self.dim:
 *             raise IndexError('column index out of range')             # <<<<<<<<<<<<<<
 *         cdef Integer z= PY_NEW(Integer)
 *         mpz_set(z.value, self.rows[nray][i])
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_column_index_out_of_range); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":427
 *         """
 *         if op != Py_EQ and op != Py_NE:
 *             raise TypeError('only equality and difference tests available')             # <<<<<<<<<<<<<<
 *         if self.dim != other.dim or self.length != other.length:
 *             return op == Py_NE
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_only_equality_and_difference_tes); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":777
 *         cdef dict cyls = {}
 *         for _ in range(num_exp):
 *             lengths = sum(ZZ.random_element(0,100000000) * r for r in self.rays())             # <<<<<<<<<<<<<<
 *             ncyls = iet_interface.number_of_cylinders(top, bot, lengths)
 *             if ncyls not in cyls:
 */
  __pyx_tuple__9 = PyTuple_Pack(2, __pyx_int_0, __pyx_int_100000000); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyList_Type_pop.type = (PyObject*)&PyList_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_100 = PyInt_FromLong(100); if (unlikely(!__pyx_int_100)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_200 = PyInt_FromLong(200); if (unlikely(!__pyx_int_200)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_100000000 = PyInt_FromLong(100000000L); if (unlikely(!__pyx_int_100000000)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily = &__pyx_vtable_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily;
  __pyx_vtable_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily.fit_length = (void (*)(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *, size_t))__pyx_f_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_fit_length;
  __pyx_vtable_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily.fill_rays = (void (*)(struct __pyx_obj_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily *, Parma_Polyhedra_Library::Generator_System const &))__pyx_f_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_fill_rays;
  if (PyType_Ready(&__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily) < 0) __PYX_ERR(0, 140, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily.tp_dictoffset && __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #if CYTHON_COMPILING_IN_CPYTHON
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)&__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily, "__repr__"); if (unlikely(!wrapper)) __PYX_ERR(0, 140, __pyx_L1_error)
    if (Py_TYPE(wrapper) == &PyWrapperDescr_Type) {
      __pyx_wrapperbase_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_10__repr__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_10__repr__.doc = __pyx_doc_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_10__repr__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_10__repr__;
    }
  }
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)&__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily, "__hash__"); if (unlikely(!wrapper)) __PYX_ERR(0, 140, __pyx_L1_error)
    if (Py_TYPE(wrapper) == &PyWrapperDescr_Type) {
      __pyx_wrapperbase_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_16__hash__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_16__hash__.doc = __pyx_doc_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_16__hash__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_16surface_dynamics_18interval_exchanges_10iet_family_9IETFamily_16__hash__;
    }
  }
  #endif
  if (__Pyx_SetVtable(__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily.tp_dict, __pyx_vtabptr_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily) < 0) __PYX_ERR(0, 140, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_IETFamily, (PyObject *)&__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily) < 0) __PYX_ERR(0, 140, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily) < 0) __PYX_ERR(0, 140, __pyx_L1_error)
  __pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily = &__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family_IETFamily;
  if (PyType_Ready(&__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__) < 0) __PYX_ERR(0, 274, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__.tp_dictoffset && __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__ = &__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct____repr__;
  if (PyType_Ready(&__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(0, 286, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr.tp_dictoffset && __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr = &__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_1_genexpr;
  if (PyType_Ready(&__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr) < 0) __PYX_ERR(0, 287, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr.tp_dictoffset && __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr = &__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_2_genexpr;
  if (PyType_Ready(&__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr) < 0) __PYX_ERR(0, 289, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr.tp_dictoffset && __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr = &__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_3_genexpr;
  if (PyType_Ready(&__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree) < 0) __PYX_ERR(0, 598, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree.tp_dictoffset && __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree = &__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_4_tree;
  if (PyType_Ready(&__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element) < 0) __PYX_ERR(0, 667, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element.tp_dictoffset && __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element = &__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_5_random_element;
  if (PyType_Ready(&__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr) < 0) __PYX_ERR(0, 703, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr.tp_dictoffset && __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr = &__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_6_genexpr;
  if (PyType_Ready(&__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics) < 0) __PYX_ERR(0, 764, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics.tp_dictoffset && __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics = &__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_7_random_integer_point_statistics;
  if (PyType_Ready(&__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr) < 0) __PYX_ERR(0, 777, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr.tp_dictoffset && __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr = &__pyx_type_16surface_dynamics_18interval_exchanges_10iet_family___pyx_scope_struct_8_genexpr;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(5, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ppl.linear_algebra"); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_3ppl_14linear_algebra_Variable = __Pyx_ImportType(__pyx_t_1, "ppl.linear_algebra", "Variable", sizeof(struct __pyx_obj_3ppl_14linear_algebra_Variable), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3ppl_14linear_algebra_Variable) __PYX_ERR(6, 3, __pyx_L1_error)
  __pyx_ptype_3ppl_14linear_algebra_Variables_Set = __Pyx_ImportType(__pyx_t_1, "ppl.linear_algebra", "Variables_Set", sizeof(struct __pyx_obj_3ppl_14linear_algebra_Variables_Set), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3ppl_14linear_algebra_Variables_Set) __PYX_ERR(6, 6, __pyx_L1_error)
  __pyx_ptype_3ppl_14linear_algebra_Linear_Expression = __Pyx_ImportType(__pyx_t_1, "ppl.linear_algebra", "Linear_Expression", sizeof(struct __pyx_obj_3ppl_14linear_algebra_Linear_Expression), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3ppl_14linear_algebra_Linear_Expression) __PYX_ERR(6, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ppl.generator"); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_3ppl_9generator_Generator = __Pyx_ImportType(__pyx_t_1, "ppl.generator", "Generator", sizeof(struct __pyx_obj_3ppl_9generator_Generator), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3ppl_9generator_Generator) __PYX_ERR(7, 10, __pyx_L1_error)
  __pyx_ptype_3ppl_9generator_Generator_System = __Pyx_ImportType(__pyx_t_1, "ppl.generator", "Generator_System", sizeof(struct __pyx_obj_3ppl_9generator_Generator_System), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3ppl_9generator_Generator_System) __PYX_ERR(7, 13, __pyx_L1_error)
  __pyx_ptype_3ppl_9generator_Poly_Gen_Relation = __Pyx_ImportType(__pyx_t_1, "ppl.generator", "Poly_Gen_Relation", sizeof(struct __pyx_obj_3ppl_9generator_Poly_Gen_Relation), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3ppl_9generator_Poly_Gen_Relation) __PYX_ERR(7, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ppl.constraint"); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_3ppl_10constraint_Constraint = __Pyx_ImportType(__pyx_t_1, "ppl.constraint", "Constraint", sizeof(struct __pyx_obj_3ppl_10constraint_Constraint), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3ppl_10constraint_Constraint) __PYX_ERR(8, 9, __pyx_L1_error)
  __pyx_ptype_3ppl_10constraint_Constraint_System = __Pyx_ImportType(__pyx_t_1, "ppl.constraint", "Constraint_System", sizeof(struct __pyx_obj_3ppl_10constraint_Constraint_System), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3ppl_10constraint_Constraint_System) __PYX_ERR(8, 12, __pyx_L1_error)
  __pyx_ptype_3ppl_10constraint_Poly_Con_Relation = __Pyx_ImportType(__pyx_t_1, "ppl.constraint", "Poly_Con_Relation", sizeof(struct __pyx_obj_3ppl_10constraint_Poly_Con_Relation), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3ppl_10constraint_Poly_Con_Relation) __PYX_ERR(8, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("ppl.polyhedron"); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_3ppl_10polyhedron_Polyhedron = __Pyx_ImportType(__pyx_t_1, "ppl.polyhedron", "Polyhedron", sizeof(struct __pyx_obj_3ppl_10polyhedron_Polyhedron), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3ppl_10polyhedron_Polyhedron) __PYX_ERR(9, 5, __pyx_L1_error)
  __pyx_vtabptr_3ppl_10polyhedron_Polyhedron = (struct __pyx_vtabstruct_3ppl_10polyhedron_Polyhedron*)__Pyx_GetVtable(__pyx_ptype_3ppl_10polyhedron_Polyhedron->tp_dict); if (unlikely(!__pyx_vtabptr_3ppl_10polyhedron_Polyhedron)) __PYX_ERR(9, 5, __pyx_L1_error)
  __pyx_ptype_3ppl_10polyhedron_C_Polyhedron = __Pyx_ImportType(__pyx_t_1, "ppl.polyhedron", "C_Polyhedron", sizeof(struct __pyx_obj_3ppl_10polyhedron_C_Polyhedron), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3ppl_10polyhedron_C_Polyhedron) __PYX_ERR(9, 10, __pyx_L1_error)
  __pyx_vtabptr_3ppl_10polyhedron_C_Polyhedron = (struct __pyx_vtabstruct_3ppl_10polyhedron_C_Polyhedron*)__Pyx_GetVtable(__pyx_ptype_3ppl_10polyhedron_C_Polyhedron->tp_dict); if (unlikely(!__pyx_vtabptr_3ppl_10polyhedron_C_Polyhedron)) __PYX_ERR(9, 10, __pyx_L1_error)
  __pyx_ptype_3ppl_10polyhedron_NNC_Polyhedron = __Pyx_ImportType(__pyx_t_1, "ppl.polyhedron", "NNC_Polyhedron", sizeof(struct __pyx_obj_3ppl_10polyhedron_NNC_Polyhedron), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3ppl_10polyhedron_NNC_Polyhedron) __PYX_ERR(9, 13, __pyx_L1_error)
  __pyx_vtabptr_3ppl_10polyhedron_NNC_Polyhedron = (struct __pyx_vtabstruct_3ppl_10polyhedron_NNC_Polyhedron*)__Pyx_GetVtable(__pyx_ptype_3ppl_10polyhedron_NNC_Polyhedron->tp_dict); if (unlikely(!__pyx_vtabptr_3ppl_10polyhedron_NNC_Polyhedron)) __PYX_ERR(9, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("sage.structure.sage_object"); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_4sage_9structure_11sage_object_SageObject = __Pyx_ImportType(__pyx_t_1, "sage.structure.sage_object", "SageObject", sizeof(struct __pyx_obj_4sage_9structure_11sage_object_SageObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_9structure_11sage_object_SageObject) __PYX_ERR(10, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("sage.structure.category_object"); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_4sage_9structure_15category_object_CategoryObject = __Pyx_ImportType(__pyx_t_1, "sage.structure.category_object", "CategoryObject", sizeof(struct __pyx_obj_4sage_9structure_15category_object_CategoryObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_9structure_15category_object_CategoryObject) __PYX_ERR(11, 15, __pyx_L1_error)
  __pyx_vtabptr_4sage_9structure_15category_object_CategoryObject = (struct __pyx_vtabstruct_4sage_9structure_15category_object_CategoryObject*)__Pyx_GetVtable(__pyx_ptype_4sage_9structure_15category_object_CategoryObject->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_9structure_15category_object_CategoryObject)) __PYX_ERR(11, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("sage.structure.coerce_dict"); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_4sage_9structure_11coerce_dict_MonoDict = __Pyx_ImportType(__pyx_t_1, "sage.structure.coerce_dict", "MonoDict", sizeof(struct __pyx_obj_4sage_9structure_11coerce_dict_MonoDict), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_9structure_11coerce_dict_MonoDict) __PYX_ERR(12, 12, __pyx_L1_error)
  __pyx_vtabptr_4sage_9structure_11coerce_dict_MonoDict = (struct __pyx_vtabstruct_4sage_9structure_11coerce_dict_MonoDict*)__Pyx_GetVtable(__pyx_ptype_4sage_9structure_11coerce_dict_MonoDict->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_9structure_11coerce_dict_MonoDict)) __PYX_ERR(12, 12, __pyx_L1_error)
  __pyx_ptype_4sage_9structure_11coerce_dict_TripleDict = __Pyx_ImportType(__pyx_t_1, "sage.structure.coerce_dict", "TripleDict", sizeof(struct __pyx_obj_4sage_9structure_11coerce_dict_TripleDict), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_9structure_11coerce_dict_TripleDict) __PYX_ERR(12, 38, __pyx_L1_error)
  __pyx_vtabptr_4sage_9structure_11coerce_dict_TripleDict = (struct __pyx_vtabstruct_4sage_9structure_11coerce_dict_TripleDict*)__Pyx_GetVtable(__pyx_ptype_4sage_9structure_11coerce_dict_TripleDict->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_9structure_11coerce_dict_TripleDict)) __PYX_ERR(12, 38, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("sage.structure.parent"); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_4sage_9structure_6parent_Parent = __Pyx_ImportType(__pyx_t_1, "sage.structure.parent", "Parent", sizeof(struct __pyx_obj_4sage_9structure_6parent_Parent), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_9structure_6parent_Parent) __PYX_ERR(13, 12, __pyx_L1_error)
  __pyx_vtabptr_4sage_9structure_6parent_Parent = (struct __pyx_vtabstruct_4sage_9structure_6parent_Parent*)__Pyx_GetVtable(__pyx_ptype_4sage_9structure_6parent_Parent->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_9structure_6parent_Parent)) __PYX_ERR(13, 12, __pyx_L1_error)
  __pyx_ptype_4sage_9structure_6parent_Set_generic = __Pyx_ImportType(__pyx_t_1, "sage.structure.parent", "Set_generic", sizeof(struct __pyx_obj_4sage_9structure_6parent_Set_generic), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_9structure_6parent_Set_generic) __PYX_ERR(13, 103, __pyx_L1_error)
  __pyx_vtabptr_4sage_9structure_6parent_Set_generic = (struct __pyx_vtabstruct_4sage_9structure_6parent_Set_generic*)__Pyx_GetVtable(__pyx_ptype_4sage_9structure_6parent_Set_generic->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_9structure_6parent_Set_generic)) __PYX_ERR(13, 103, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("sage.misc.inherit_comparison"); if (unlikely(!__pyx_t_1)) __PYX_ERR(14, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_4sage_4misc_18inherit_comparison_InheritComparisonMetaclass = __Pyx_ImportType(__pyx_t_1, "sage.misc.inherit_comparison", "InheritComparisonMetaclass", sizeof(struct __pyx_obj_4sage_4misc_18inherit_comparison_InheritComparisonMetaclass), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_4misc_18inherit_comparison_InheritComparisonMetaclass) __PYX_ERR(14, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("sage.structure.element"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_4sage_9structure_7element_Element = __Pyx_ImportType(__pyx_t_1, "sage.structure.element", "Element", sizeof(struct __pyx_obj_4sage_9structure_7element_Element), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_9structure_7element_Element) __PYX_ERR(4, 150, __pyx_L1_error)
  __pyx_vtabptr_4sage_9structure_7element_Element = (struct __pyx_vtabstruct_4sage_9structure_7element_Element*)__Pyx_GetVtable(__pyx_ptype_4sage_9structure_7element_Element->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_9structure_7element_Element)) __PYX_ERR(4, 150, __pyx_L1_error)
  __pyx_ptype_4sage_9structure_7element_ElementWithCachedMethod = __Pyx_ImportType(__pyx_t_1, "sage.structure.element", "ElementWithCachedMethod", sizeof(struct __pyx_obj_4sage_9structure_7element_ElementWithCachedMethod), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_9structure_7element_ElementWithCachedMethod) __PYX_ERR(4, 177, __pyx_L1_error)
  __pyx_vtabptr_4sage_9structure_7element_ElementWithCachedMethod = (struct __pyx_vtabstruct_4sage_9structure_7element_ElementWithCachedMethod*)__Pyx_GetVtable(__pyx_ptype_4sage_9structure_7element_ElementWithCachedMethod->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_9structure_7element_ElementWithCachedMethod)) __PYX_ERR(4, 177, __pyx_L1_error)
  __pyx_ptype_4sage_9structure_7element_ModuleElement = __Pyx_ImportType(__pyx_t_1, "sage.structure.element", "ModuleElement", sizeof(struct __pyx_obj_4sage_9structure_7element_ModuleElement), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_9structure_7element_ModuleElement) __PYX_ERR(4, 180, __pyx_L1_error)
  __pyx_vtabptr_4sage_9structure_7element_ModuleElement = (struct __pyx_vtabstruct_4sage_9structure_7element_ModuleElement*)__Pyx_GetVtable(__pyx_ptype_4sage_9structure_7element_ModuleElement->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_9structure_7element_ModuleElement)) __PYX_ERR(4, 180, __pyx_L1_error)
  __pyx_ptype_4sage_9structure_7element_RingElement = __Pyx_ImportType(__pyx_t_1, "sage.structure.element", "RingElement", sizeof(struct __pyx_obj_4sage_9structure_7element_RingElement), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_9structure_7element_RingElement) __PYX_ERR(4, 182, __pyx_L1_error)
  __pyx_vtabptr_4sage_9structure_7element_RingElement = (struct __pyx_vtabstruct_4sage_9structure_7element_RingElement*)__Pyx_GetVtable(__pyx_ptype_4sage_9structure_7element_RingElement->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_9structure_7element_RingElement)) __PYX_ERR(4, 182, __pyx_L1_error)
  __pyx_ptype_4sage_9structure_7element_MonoidElement = __Pyx_ImportType(__pyx_t_1, "sage.structure.element", "MonoidElement", sizeof(struct __pyx_obj_4sage_9structure_7element_MonoidElement), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_9structure_7element_MonoidElement) __PYX_ERR(4, 194, __pyx_L1_error)
  __pyx_vtabptr_4sage_9structure_7element_MonoidElement = (struct __pyx_vtabstruct_4sage_9structure_7element_MonoidElement*)__Pyx_GetVtable(__pyx_ptype_4sage_9structure_7element_MonoidElement->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_9structure_7element_MonoidElement)) __PYX_ERR(4, 194, __pyx_L1_error)
  __pyx_ptype_4sage_9structure_7element_MultiplicativeGroupElement = __Pyx_ImportType(__pyx_t_1, "sage.structure.element", "MultiplicativeGroupElement", sizeof(struct __pyx_obj_4sage_9structure_7element_MultiplicativeGroupElement), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_9structure_7element_MultiplicativeGroupElement) __PYX_ERR(4, 197, __pyx_L1_error)
  __pyx_vtabptr_4sage_9structure_7element_MultiplicativeGroupElement = (struct __pyx_vtabstruct_4sage_9structure_7element_MultiplicativeGroupElement*)__Pyx_GetVtable(__pyx_ptype_4sage_9structure_7element_MultiplicativeGroupElement->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_9structure_7element_MultiplicativeGroupElement)) __PYX_ERR(4, 197, __pyx_L1_error)
  __pyx_ptype_4sage_9structure_7element_AdditiveGroupElement = __Pyx_ImportType(__pyx_t_1, "sage.structure.element", "AdditiveGroupElement", sizeof(struct __pyx_obj_4sage_9structure_7element_AdditiveGroupElement), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_9structure_7element_AdditiveGroupElement) __PYX_ERR(4, 200, __pyx_L1_error)
  __pyx_vtabptr_4sage_9structure_7element_AdditiveGroupElement = (struct __pyx_vtabstruct_4sage_9structure_7element_AdditiveGroupElement*)__Pyx_GetVtable(__pyx_ptype_4sage_9structure_7element_AdditiveGroupElement->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_9structure_7element_AdditiveGroupElement)) __PYX_ERR(4, 200, __pyx_L1_error)
  __pyx_ptype_4sage_9structure_7element_CommutativeRingElement = __Pyx_ImportType(__pyx_t_1, "sage.structure.element", "CommutativeRingElement", sizeof(struct __pyx_obj_4sage_9structure_7element_CommutativeRingElement), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_9structure_7element_CommutativeRingElement) __PYX_ERR(4, 208, __pyx_L1_error)
  __pyx_vtabptr_4sage_9structure_7element_CommutativeRingElement = (struct __pyx_vtabstruct_4sage_9structure_7element_CommutativeRingElement*)__Pyx_GetVtable(__pyx_ptype_4sage_9structure_7element_CommutativeRingElement->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_9structure_7element_CommutativeRingElement)) __PYX_ERR(4, 208, __pyx_L1_error)
  __pyx_ptype_4sage_9structure_7element_IntegralDomainElement = __Pyx_ImportType(__pyx_t_1, "sage.structure.element", "IntegralDomainElement", sizeof(struct __pyx_obj_4sage_9structure_7element_IntegralDomainElement), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_9structure_7element_IntegralDomainElement) __PYX_ERR(4, 211, __pyx_L1_error)
  __pyx_vtabptr_4sage_9structure_7element_IntegralDomainElement = (struct __pyx_vtabstruct_4sage_9structure_7element_IntegralDomainElement*)__Pyx_GetVtable(__pyx_ptype_4sage_9structure_7element_IntegralDomainElement->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_9structure_7element_IntegralDomainElement)) __PYX_ERR(4, 211, __pyx_L1_error)
  __pyx_ptype_4sage_9structure_7element_DedekindDomainElement = __Pyx_ImportType(__pyx_t_1, "sage.structure.element", "DedekindDomainElement", sizeof(struct __pyx_obj_4sage_9structure_7element_DedekindDomainElement), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_9structure_7element_DedekindDomainElement) __PYX_ERR(4, 214, __pyx_L1_error)
  __pyx_vtabptr_4sage_9structure_7element_DedekindDomainElement = (struct __pyx_vtabstruct_4sage_9structure_7element_DedekindDomainElement*)__Pyx_GetVtable(__pyx_ptype_4sage_9structure_7element_DedekindDomainElement->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_9structure_7element_DedekindDomainElement)) __PYX_ERR(4, 214, __pyx_L1_error)
  __pyx_ptype_4sage_9structure_7element_PrincipalIdealDomainElement = __Pyx_ImportType(__pyx_t_1, "sage.structure.element", "PrincipalIdealDomainElement", sizeof(struct __pyx_obj_4sage_9structure_7element_PrincipalIdealDomainElement), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_9structure_7element_PrincipalIdealDomainElement) __PYX_ERR(4, 217, __pyx_L1_error)
  __pyx_vtabptr_4sage_9structure_7element_PrincipalIdealDomainElement = (struct __pyx_vtabstruct_4sage_9structure_7element_PrincipalIdealDomainElement*)__Pyx_GetVtable(__pyx_ptype_4sage_9structure_7element_PrincipalIdealDomainElement->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_9structure_7element_PrincipalIdealDomainElement)) __PYX_ERR(4, 217, __pyx_L1_error)
  __pyx_ptype_4sage_9structure_7element_EuclideanDomainElement = __Pyx_ImportType(__pyx_t_1, "sage.structure.element", "EuclideanDomainElement", sizeof(struct __pyx_obj_4sage_9structure_7element_EuclideanDomainElement), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_9structure_7element_EuclideanDomainElement) __PYX_ERR(4, 220, __pyx_L1_error)
  __pyx_vtabptr_4sage_9structure_7element_EuclideanDomainElement = (struct __pyx_vtabstruct_4sage_9structure_7element_EuclideanDomainElement*)__Pyx_GetVtable(__pyx_ptype_4sage_9structure_7element_EuclideanDomainElement->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_9structure_7element_EuclideanDomainElement)) __PYX_ERR(4, 220, __pyx_L1_error)
  __pyx_ptype_4sage_9structure_7element_FieldElement = __Pyx_ImportType(__pyx_t_1, "sage.structure.element", "FieldElement", sizeof(struct __pyx_obj_4sage_9structure_7element_FieldElement), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_9structure_7element_FieldElement) __PYX_ERR(4, 224, __pyx_L1_error)
  __pyx_vtabptr_4sage_9structure_7element_FieldElement = (struct __pyx_vtabstruct_4sage_9structure_7element_FieldElement*)__Pyx_GetVtable(__pyx_ptype_4sage_9structure_7element_FieldElement->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_9structure_7element_FieldElement)) __PYX_ERR(4, 224, __pyx_L1_error)
  __pyx_ptype_4sage_9structure_7element_AlgebraElement = __Pyx_ImportType(__pyx_t_1, "sage.structure.element", "AlgebraElement", sizeof(struct __pyx_obj_4sage_9structure_7element_AlgebraElement), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_9structure_7element_AlgebraElement) __PYX_ERR(4, 227, __pyx_L1_error)
  __pyx_vtabptr_4sage_9structure_7element_AlgebraElement = (struct __pyx_vtabstruct_4sage_9structure_7element_AlgebraElement*)__Pyx_GetVtable(__pyx_ptype_4sage_9structure_7element_AlgebraElement->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_9structure_7element_AlgebraElement)) __PYX_ERR(4, 227, __pyx_L1_error)
  __pyx_ptype_4sage_9structure_7element_CommutativeAlgebraElement = __Pyx_ImportType(__pyx_t_1, "sage.structure.element", "CommutativeAlgebraElement", sizeof(struct __pyx_obj_4sage_9structure_7element_CommutativeAlgebraElement), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_9structure_7element_CommutativeAlgebraElement) __PYX_ERR(4, 230, __pyx_L1_error)
  __pyx_vtabptr_4sage_9structure_7element_CommutativeAlgebraElement = (struct __pyx_vtabstruct_4sage_9structure_7element_CommutativeAlgebraElement*)__Pyx_GetVtable(__pyx_ptype_4sage_9structure_7element_CommutativeAlgebraElement->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_9structure_7element_CommutativeAlgebraElement)) __PYX_ERR(4, 230, __pyx_L1_error)
  __pyx_ptype_4sage_9structure_7element_InfinityElement = __Pyx_ImportType(__pyx_t_1, "sage.structure.element", "InfinityElement", sizeof(struct __pyx_obj_4sage_9structure_7element_InfinityElement), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_9structure_7element_InfinityElement) __PYX_ERR(4, 233, __pyx_L1_error)
  __pyx_vtabptr_4sage_9structure_7element_InfinityElement = (struct __pyx_vtabstruct_4sage_9structure_7element_InfinityElement*)__Pyx_GetVtable(__pyx_ptype_4sage_9structure_7element_InfinityElement->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_9structure_7element_InfinityElement)) __PYX_ERR(4, 233, __pyx_L1_error)
  __pyx_ptype_4sage_9structure_7element_Vector = __Pyx_ImportType(__pyx_t_1, "sage.structure.element", "Vector", sizeof(struct __pyx_obj_4sage_9structure_7element_Vector), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_9structure_7element_Vector) __PYX_ERR(4, 237, __pyx_L1_error)
  __pyx_vtabptr_4sage_9structure_7element_Vector = (struct __pyx_vtabstruct_4sage_9structure_7element_Vector*)__Pyx_GetVtable(__pyx_ptype_4sage_9structure_7element_Vector->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_9structure_7element_Vector)) __PYX_ERR(4, 237, __pyx_L1_error)
  __pyx_ptype_4sage_9structure_7element_Matrix = __Pyx_ImportType(__pyx_t_1, "sage.structure.element", "Matrix", sizeof(struct __pyx_obj_4sage_9structure_7element_Matrix), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_9structure_7element_Matrix) __PYX_ERR(4, 252, __pyx_L1_error)
  __pyx_vtabptr_4sage_9structure_7element_Matrix = (struct __pyx_vtabstruct_4sage_9structure_7element_Matrix*)__Pyx_GetVtable(__pyx_ptype_4sage_9structure_7element_Matrix->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_9structure_7element_Matrix)) __PYX_ERR(4, 252, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("sage.categories.map"); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_4sage_10categories_3map_Map = __Pyx_ImportType(__pyx_t_1, "sage.categories.map", "Map", sizeof(struct __pyx_obj_4sage_10categories_3map_Map), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_10categories_3map_Map) __PYX_ERR(15, 4, __pyx_L1_error)
  __pyx_vtabptr_4sage_10categories_3map_Map = (struct __pyx_vtabstruct_4sage_10categories_3map_Map*)__Pyx_GetVtable(__pyx_ptype_4sage_10categories_3map_Map->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_10categories_3map_Map)) __PYX_ERR(15, 4, __pyx_L1_error)
  __pyx_ptype_4sage_10categories_3map_Section = __Pyx_ImportType(__pyx_t_1, "sage.categories.map", "Section", sizeof(struct __pyx_obj_4sage_10categories_3map_Section), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_10categories_3map_Section) __PYX_ERR(15, 28, __pyx_L1_error)
  __pyx_vtabptr_4sage_10categories_3map_Section = (struct __pyx_vtabstruct_4sage_10categories_3map_Section*)__Pyx_GetVtable(__pyx_ptype_4sage_10categories_3map_Section->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_10categories_3map_Section)) __PYX_ERR(15, 28, __pyx_L1_error)
  __pyx_ptype_4sage_10categories_3map_FormalCompositeMap = __Pyx_ImportType(__pyx_t_1, "sage.categories.map", "FormalCompositeMap", sizeof(struct __pyx_obj_4sage_10categories_3map_FormalCompositeMap), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_10categories_3map_FormalCompositeMap) __PYX_ERR(15, 31, __pyx_L1_error)
  __pyx_vtabptr_4sage_10categories_3map_FormalCompositeMap = (struct __pyx_vtabstruct_4sage_10categories_3map_FormalCompositeMap*)__Pyx_GetVtable(__pyx_ptype_4sage_10categories_3map_FormalCompositeMap->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_10categories_3map_FormalCompositeMap)) __PYX_ERR(15, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("sage.categories.morphism"); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_4sage_10categories_8morphism_Morphism = __Pyx_ImportType(__pyx_t_1, "sage.categories.morphism", "Morphism", sizeof(struct __pyx_obj_4sage_10categories_8morphism_Morphism), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_10categories_8morphism_Morphism) __PYX_ERR(16, 5, __pyx_L1_error)
  __pyx_vtabptr_4sage_10categories_8morphism_Morphism = (struct __pyx_vtabstruct_4sage_10categories_8morphism_Morphism*)__Pyx_GetVtable(__pyx_ptype_4sage_10categories_8morphism_Morphism->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_10categories_8morphism_Morphism)) __PYX_ERR(16, 5, __pyx_L1_error)
  __pyx_ptype_4sage_10categories_8morphism_SetMorphism = __Pyx_ImportType(__pyx_t_1, "sage.categories.morphism", "SetMorphism", sizeof(struct __pyx_obj_4sage_10categories_8morphism_SetMorphism), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_10categories_8morphism_SetMorphism) __PYX_ERR(16, 8, __pyx_L1_error)
  __pyx_vtabptr_4sage_10categories_8morphism_SetMorphism = (struct __pyx_vtabstruct_4sage_10categories_8morphism_SetMorphism*)__Pyx_GetVtable(__pyx_ptype_4sage_10categories_8morphism_SetMorphism->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_10categories_8morphism_SetMorphism)) __PYX_ERR(16, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("sage.rings.integer"); if (unlikely(!__pyx_t_1)) __PYX_ERR(17, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_4sage_5rings_7integer_Integer = __Pyx_ImportType(__pyx_t_1, "sage.rings.integer", "Integer", sizeof(struct __pyx_obj_4sage_5rings_7integer_Integer), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_5rings_7integer_Integer) __PYX_ERR(17, 7, __pyx_L1_error)
  __pyx_vtabptr_4sage_5rings_7integer_Integer = (struct __pyx_vtabstruct_4sage_5rings_7integer_Integer*)__Pyx_GetVtable(__pyx_ptype_4sage_5rings_7integer_Integer->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_5rings_7integer_Integer)) __PYX_ERR(17, 7, __pyx_L1_error)
  __pyx_ptype_4sage_5rings_7integer_int_to_Z = __Pyx_ImportType(__pyx_t_1, "sage.rings.integer", "int_to_Z", sizeof(struct __pyx_obj_4sage_5rings_7integer_int_to_Z), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_4sage_5rings_7integer_int_to_Z) __PYX_ERR(17, 41, __pyx_L1_error)
  __pyx_vtabptr_4sage_5rings_7integer_int_to_Z = (struct __pyx_vtabstruct_4sage_5rings_7integer_int_to_Z*)__Pyx_GetVtable(__pyx_ptype_4sage_5rings_7integer_int_to_Z->tp_dict); if (unlikely(!__pyx_vtabptr_4sage_5rings_7integer_int_to_Z)) __PYX_ERR(17, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __pyx_t_1 = PyImport_ImportModule("cysignals.signals"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "cysigs", (void **)&__pyx_vp_9cysignals_7signals_cysigs, "cysigs_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("cysignals.signals"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "verify_exc_value", (void (**)(void))&__pyx_f_9cysignals_7signals_verify_exc_value, "void (void)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "_sig_on_interrupt_received", (void (**)(void))&_sig_on_interrupt_received, "void (void)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "_sig_on_recover", (void (**)(void))&_sig_on_recover, "void (void)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "_sig_off_warning", (void (**)(void))&_sig_off_warning, "void (char const *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "print_backtrace", (void (**)(void))&print_backtrace, "void (void)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initiet_family(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initiet_family(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_iet_family(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_iet_family(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_iet_family(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'iet_family' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_iet_family(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("iet_family", __pyx_methods, __pyx_k_Linear_families_of_interval_exc, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_surface_dynamics__interval_exchanges__iet_family) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "surface_dynamics.interval_exchanges.iet_family")) {
      if (unlikely(PyDict_SetItemString(modules, "surface_dynamics.interval_exchanges.iet_family", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_variable_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_function_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "surface_dynamics/interval_exchanges/iet_family.pyx":1
 * # distutils: language=c++             # <<<<<<<<<<<<<<
 * # distutils: extra_compile_args=-std=c++11
 * # distutils: libraries=ppl gmp
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "sage/structure/element.pxd":111
 * 
 * 
 * cpdef inline bint have_same_parent(left, right):             # <<<<<<<<<<<<<<
 *     """
 *     Return ``True`` if and only if ``left`` and ``right`` have the
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init surface_dynamics.interval_exchanges.iet_family", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init surface_dynamics.interval_exchanges.iet_family");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* StringJoin */
#if !CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values) {
    return PyObject_CallMethodObjArgs(sep, __pyx_n_s_join, values, NULL);
}
#endif

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* pyfrozenset_new */
static CYTHON_INLINE PyObject* __Pyx_PyFrozenSet_New(PyObject* it) {
    if (it) {
        PyObject* result;
#if CYTHON_COMPILING_IN_PYPY
        PyObject* args;
        args = PyTuple_Pack(1, it);
        if (unlikely(!args))
            return NULL;
        result = PyObject_Call((PyObject*)&PyFrozenSet_Type, args, NULL);
        Py_DECREF(args);
        return result;
#else
        if (PyFrozenSet_CheckExact(it)) {
            Py_INCREF(it);
            return it;
        }
        result = PyFrozenSet_New(it);
        if (unlikely(!result))
            return NULL;
        if (likely(PySet_GET_SIZE(result)))
            return result;
        Py_DECREF(result);
#endif
    }
#if CYTHON_USE_TYPE_SLOTS
    return PyFrozenSet_Type.tp_new(&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#else
    return PyObject_Call((PyObject*)&PyFrozenSet_Type, __pyx_empty_tuple, NULL);
#endif
}

/* PySetContains */
static int __Pyx_PySet_ContainsUnhashable(PyObject *set, PyObject *key) {
    int result = -1;
    if (PySet_Check(key) && PyErr_ExceptionMatches(PyExc_TypeError)) {
        PyObject *tmpkey;
        PyErr_Clear();
        tmpkey = __Pyx_PyFrozenSet_New(key);
        if (tmpkey != NULL) {
            result = PySet_Contains(set, tmpkey);
            Py_DECREF(tmpkey);
        }
    }
    return result;
}
static CYTHON_INLINE int __Pyx_PySet_ContainsTF(PyObject* key, PyObject* set, int eq) {
    int result = PySet_Contains(set, key);
    if (unlikely(result < 0)) {
        result = __Pyx_PySet_ContainsUnhashable(set, key);
    }
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* append */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* pop_index */
static PyObject* __Pyx__PyObject_PopNewIndex(PyObject* L, PyObject* py_ix) {
    PyObject *r;
    if (unlikely(!py_ix)) return NULL;
    r = __Pyx__PyObject_PopIndex(L, py_ix);
    Py_DECREF(py_ix);
    return r;
}
static PyObject* __Pyx__PyObject_PopIndex(PyObject* L, PyObject* py_ix) {
    return __Pyx_PyObject_CallMethod1(L, __pyx_n_s_pop, py_ix);
}
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static PyObject* __Pyx__PyList_PopIndex(PyObject* L, PyObject* py_ix, Py_ssize_t ix) {
    Py_ssize_t size = PyList_GET_SIZE(L);
    if (likely(size > (((PyListObject*)L)->allocated >> 1))) {
        Py_ssize_t cix = ix;
        if (cix < 0) {
            cix += size;
        }
        if (likely(__Pyx_is_valid_index(cix, size))) {
            PyObject* v = PyList_GET_ITEM(L, cix);
            __Pyx_SET_SIZE(L, Py_SIZE(L) - 1);
            size -= 1;
            memmove(&PyList_GET_ITEM(L, cix), &PyList_GET_ITEM(L, cix+1), (size_t)(size-cix)*sizeof(PyObject*));
            return v;
        }
    }
    if (py_ix == Py_None) {
        return __Pyx__PyObject_PopNewIndex(L, PyInt_FromSsize_t(ix));
    } else {
        return __Pyx__PyObject_PopIndex(L, py_ix);
    }
}
#endif

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* pop */
static CYTHON_INLINE PyObject* __Pyx__PyObject_Pop(PyObject* L) {
    if (Py_TYPE(L) == &PySet_Type) {
        return PySet_Pop(L);
    }
    return __Pyx_PyObject_CallMethod0(L, __pyx_n_s_pop);
}
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE PyObject* __Pyx_PyList_Pop(PyObject* L) {
    if (likely(PyList_GET_SIZE(L) > (((PyListObject*)L)->allocated >> 1))) {
        __Pyx_SET_SIZE(L, Py_SIZE(L) - 1);
        return PyList_GET_ITEM(L, PyList_GET_SIZE(L));
    }
    return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyList_Type_pop, L);
}
#endif

/* py_set_discard_unhashable */
static int __Pyx_PySet_DiscardUnhashable(PyObject *set, PyObject *key) {
    PyObject *tmpkey;
    int rv;
    if (likely(!PySet_Check(key) || !PyErr_ExceptionMatches(PyExc_TypeError)))
        return -1;
    PyErr_Clear();
    tmpkey = __Pyx_PyFrozenSet_New(key);
    if (tmpkey == NULL)
        return -1;
    rv = PySet_Discard(set, tmpkey);
    Py_DECREF(tmpkey);
    return rv;
}

/* py_set_remove */
static int __Pyx_PySet_RemoveNotFound(PyObject *set, PyObject *key, int found) {
    if (unlikely(found < 0)) {
        found = __Pyx_PySet_DiscardUnhashable(set, key);
    }
    if (likely(found == 0)) {
        PyObject *tup;
        tup = PyTuple_Pack(1, key);
        if (!tup)
            return -1;
        PyErr_SetObject(PyExc_KeyError, tup);
        Py_DECREF(tup);
        return -1;
    }
    return found;
}
static CYTHON_INLINE int __Pyx_PySet_Remove(PyObject *set, PyObject *key) {
    int found = PySet_Discard(set, key);
    if (unlikely(found != 1)) {
        return __Pyx_PySet_RemoveNotFound(set, key, found);
    }
    return 0;
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* Print */
#if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static PyObject *__Pyx_GetStdout(void) {
    PyObject *f = PySys_GetObject((char *)"stdout");
    if (!f) {
        PyErr_SetString(PyExc_RuntimeError, "lost sys.stdout");
    }
    return f;
}
static int __Pyx_Print(PyObject* f, PyObject *arg_tuple, int newline) {
    int i;
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    for (i=0; i < PyTuple_GET_SIZE(arg_tuple); i++) {
        PyObject* v;
        if (PyFile_SoftSpace(f, 1)) {
            if (PyFile_WriteString(" ", f) < 0)
                goto error;
        }
        v = PyTuple_GET_ITEM(arg_tuple, i);
        if (PyFile_WriteObject(v, f, Py_PRINT_RAW) < 0)
            goto error;
        if (PyString_Check(v)) {
            char *s = PyString_AsString(v);
            Py_ssize_t len = PyString_Size(v);
            if (len > 0) {
                switch (s[len-1]) {
                    case ' ': break;
                    case '\f': case '\r': case '\n': case '\t': case '\v':
                        PyFile_SoftSpace(f, 0);
                        break;
                    default:  break;
                }
            }
        }
    }
    if (newline) {
        if (PyFile_WriteString("\n", f) < 0)
            goto error;
        PyFile_SoftSpace(f, 0);
    }
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
}
#else
static int __Pyx_Print(PyObject* stream, PyObject *arg_tuple, int newline) {
    PyObject* kwargs = 0;
    PyObject* result = 0;
    PyObject* end_string;
    if (unlikely(!__pyx_print)) {
        __pyx_print = PyObject_GetAttr(__pyx_b, __pyx_n_s_print);
        if (!__pyx_print)
            return -1;
    }
    if (stream) {
        kwargs = PyDict_New();
        if (unlikely(!kwargs))
            return -1;
        if (unlikely(PyDict_SetItem(kwargs, __pyx_n_s_file, stream) < 0))
            goto bad;
        if (!newline) {
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                goto bad;
            if (PyDict_SetItem(kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                goto bad;
            }
            Py_DECREF(end_string);
        }
    } else if (!newline) {
        if (unlikely(!__pyx_print_kwargs)) {
            __pyx_print_kwargs = PyDict_New();
            if (unlikely(!__pyx_print_kwargs))
                return -1;
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                return -1;
            if (PyDict_SetItem(__pyx_print_kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                return -1;
            }
            Py_DECREF(end_string);
        }
        kwargs = __pyx_print_kwargs;
    }
    result = PyObject_Call(__pyx_print, arg_tuple, kwargs);
    if (unlikely(kwargs) && (kwargs != __pyx_print_kwargs))
        Py_DECREF(kwargs);
    if (!result)
        return -1;
    Py_DECREF(result);
    return 0;
bad:
    if (kwargs != __pyx_print_kwargs)
        Py_XDECREF(kwargs);
    return -1;
}
#endif

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) ((size_t) 0 - (size_t) 1), const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* PrintOne */
#if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static int __Pyx_PrintOne(PyObject* f, PyObject *o) {
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    if (PyFile_SoftSpace(f, 0)) {
        if (PyFile_WriteString(" ", f) < 0)
            goto error;
    }
    if (PyFile_WriteObject(o, f, Py_PRINT_RAW) < 0)
        goto error;
    if (PyFile_WriteString("\n", f) < 0)
        goto error;
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
    /* the line below is just to avoid C compiler
     * warnings about unused functions */
    return __Pyx_Print(f, NULL, 0);
}
#else
static int __Pyx_PrintOne(PyObject* stream, PyObject *o) {
    int res;
    PyObject* arg_tuple = PyTuple_Pack(1, o);
    if (unlikely(!arg_tuple))
        return -1;
    res = __Pyx_Print(stream, arg_tuple, 1);
    Py_DECREF(arg_tuple);
    return res;
}
#endif

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* CoroutineBase */
#include <structmember.h>
#include <frameobject.h>
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = tstate->frame;
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (self->ob_refcnt > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    self->ob_refcnt = 1;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = self->ob_refcnt;
        _Py_NewReference(self);
        self->ob_refcnt = refcnt;
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(self->ob_type) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* VoidPtrImport */
#ifndef __PYX_HAVE_RT_ImportVoidPtr
#define __PYX_HAVE_RT_ImportVoidPtr
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, desc);
        goto bad;
    }
    *p = PyCObject_AsVoidPtr(cobj);}
#endif
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
